//    _____ _                       _  _  ______                     ______      _  _
//   / ____) |                     | || ||  ____)                   |  ____)    | || |
//  | |    | |  _ _ ___ ___  ___   |_||_|| |__   _ __ _ __ ___  _ __| |__  _   ___||_|
//  | |    | |/ _` | __) __)/ _ \        |  __) | `__) `__) _ \| `__)  __) \ \/ /
//  | |____| | (_| |__ \__ \  __/        | |____| |  | | | (_) | |  | |____ )  (
//   \_____)_|\__,_|___/___/\___|        |______)_|  |_|  \___/|_|  |______)_/\_\
//
//


package cat.lib.errors;

import cat.lib.files.LogsFile;
import cat.lib.utils.StringUtils;
import java.util.Vector;


/**
 *
 * LA CLASE "ErrorEx":
 * --------------------
 *
 *   La clase "ErrorEx" la utilizan muchas de las clases implementadas en
 * "JavaUtils" y permite transmitir errores a la aplicación sin necesidad de
 * utilizar excepciones.
 *
 *   El uso de esta clase, permite que las aplicaciones, en caso de que se
 * produzca un error, puedan continuar ejecutandose y no se produzcan roturas
 * del flujo de ejecución; permitiendo una programación más estructurada y
 * diáctica.
 *
 *   Funcionalmente, la clase "ErrorEx" dispone de métodos para que la aplicación
 * almacene los errores que se van produciendo dando una información muy interesante
 * al programador.
 *
 *   Cada error, a parte de un identificador, ta,bién puede tener una lista de
 * parejas variable/valor con información extra sobre los detalles que han
 * provocado el error, o el estado de algunas variables significativas.
 *
 *   El "inconveniente" de utilizar la clase "ErrorEx" en lugar de utilizar las
 * excepciones, es que el programador debe asumir el control del flujo de
 * ejecución cuando se produce un error.
 *
 *   A algunos programadores, seran reacios a utilizar esta herramienta, porque
 * exige un pequeño esfuerzo, aunque a la larga, resulta extremadamente útil y
 * al evitar las roturas de flujo, los programas creados son más estructurados.
 *
 *
 * @author Miquel Abadal i Cervià
 */
public class ErrorEx {

    //     __   _        _ _           _
    //   / __ \| |      (_) |         | |
    //  | |__| | |_ _ __ _| |_   _   _| |_ ___
    //  |  __  | __) `__) | `_ \| | | | __) __)
    //  | |  | | |_| |  | | |_) | |_| | |_\__ \
    //  |_|  |_|\__)_|  |_|_,__/ \__,_|\__)___/
    //
    //

    private Vector errorList = null;    // Lista de errores producida
    private LogsFile logs = null;


    //   __  __      _            _                              _                    _
    //  |  \/  |    | |          | |                            | |                  | |
    //  | \  / | ___| |_ ___   __| | ___     ____ ___  _ __  ___| |_ _ __ _   _  ____| |_ ___  _ __
    //  | |\/| |/ _ \ __) _ \ / _` |/ _ \   / ___) _ \| `_  \ __) __) `__) | | |/ ___) __) _ \| `__)
    //  | |  | |  __/ |_ (_) | (_| |  __/  | (___ (_) | | | |__ \ |_| |  | |_| | (___| |_ (_) | |
    //  |_|  |_|\___|\__)___/ \__,_|\___|   \____)___/|_| |_|___/\__)_|   \__,_|\____)\__)___/|_|
    //
    //


    public ErrorEx () { errorList = new Vector(); }

    public ErrorEx (LogsFile logs) {
        errorList = new Vector();
        this.logs = logs;
    }

    @Override
    public String toString () { return getXmlErrorMessage(); }


    //   __  __      _            _                        _             _
    //  |  \/  |    | |          | |                      (_)           | |
    //  | \  / | ___| |_ ___   __| | ___ ___    _ __  _ __ _ _    _  _ _| |_ ___
    //  | |\/| |/ _ \ __) _ \ / _` |/ _ \ __)  | `_ \| `__) | |  | | _` | __) __)
    //  | |  | |  __/ |_ (_) | (_| |  __/__ \  | |_) | |  | |\ \/ / (_| | |_\__ \
    //  |_|  |_|\___|\__)___/ \__,_|\___|___/  |  __/|_|  |_| \__/ \__,_|\__)___/
    //                                         | |
    //                                         |_|


    //   __  __      _            _             _  _   _____ ______ _______ _  _
    //  |  \/  |    | |          | |           | || | / ____)  ____)__   __) || |
    //  | \  / | ___| |_ ___   __| | ___ ___   |_||_|| (___ | |__     | |  |_||_|
    //  | |\/| |/ _ \ __) _ \ / _` |/ _ \ __)         \___ \|  __)    | |
    //  | |  | |  __/ |_ (_) | (_| |  __/__ \         ____) | |____   | |
    //  |_|  |_|\___|\__)___/ \__,_|\___|___/        (_____/|______)  |_|
    //
    //

    /**
     * Añade un error a la lista de errores.
     *
     * El parámetro "errorCode" identifica el error
     * Los parámetros "errorMsg" coresponden a parejas "variable/valor" que
     * permiten añadir información significativa de las causas del error o su
     * hubicación (clase y nombre de método) o el estado de algunas variables.
     *
     * Las parejas "variable/valor" deben separarse con el caracter ":"
     *
     * Por ejemplo, si se produce un error al grabar un fichero a disco,
     * podemos generar el siguiente error:
     *
     * error.setError ("SAVE_ERROR", "EXCEPTION: IO-EXCEPTION",
     *                               "TRACE:c lase y método donde se ha producido el error",
     *                               "FILE-NAME: nombreFichero");
     *
     */
    public void setError (String errorCode, String... errorMsg) {

        if (errorCode!=null) {
            ErrorItem error = new ErrorItem ();
            error.setErrorCode (errorCode.trim());

            if (errorMsg!=null) {
                for (int cont1=0; cont1<errorMsg.length; cont1++) { error.addErrorMsg (errorMsg[cont1].trim()); }
            }
            errorList.add (error);

            if (logs!=null) { logs.error (error.getErrorMessage());}
            else { System.out.println (error.getErrorMessage()); }
        }
    }

    public void setLogsFile (LogsFile logs) { this.logs = logs; }

    public void setErrorFromText (String text) {
        
        text = StringUtils.copy(text, 13, -1);
        
        if (StringUtils.isNotEmpty(text)) {
            
            String[] errors = text.split("\n");
            setError (errors[0], errors);
        }
        //ERROR:ERROR: NICK_ERROR DESCRIPTION: El nick indicado ya esta en uso
    }
        
    //   __  __      _            _             _  _   _____ ______ _______ _  _
    //  |  \/  |    | |          | |           | || | / ____)  ____)__   __) || |
    //  | \  / | ___| |_ ___   __| | ___ ___   |_||_|| |  __| |__     | |  |_||_|
    //  | |\/| |/ _ \ __) _ \ / _` |/ _ \ __)        | | (_ |  __)    | |
    //  | |  | |  __/ |_ (_) | (_| |  __/__ \        | |__| | |____   | |
    //  |_|  |_|\___|\__)___/ \__,_|\___|___/         \_____|______)  |_|
    //
    //




    /**
     * Obtiene el código del último error producido, o "null" si no se ha
     * producido ningun error.
     */
    public String getError () { 
        ErrorItem error = (ErrorItem) (errorList.isEmpty() ? null : errorList.lastElement());
        if (error!=null) { return error.getErrorCode(); }
        return null;
    }

    /**
     * Obtiene la descripción de los errors con la siguiente estructura:
     *
     * ERROR: errorCode\n
     * MENSAJE1\n
     * MENSAJE2\n
     * ...
     *
     */
    public String getErrorMessage (boolean allErrors) {

        StringBuilder res = new StringBuilder();

        if (errorList.size()>0) {

            int cont1 = 0;
            if (!allErrors) { cont1 = errorList.size()-1; }
            while (cont1<errorList.size()) {

                ErrorItem error = (ErrorItem) errorList.get(cont1);
                res.append (error.getErrorMessage());
                cont1++;
            }
        }

        return res.toString();
    }

    public String getErrorMessage () { return getErrorMessage(false); }

    /**
     * Obtiene la descripción de los errores con formato HTML:
     *
     * <p>[ERROR: errorCode]<br>\n
     * MENSAJE1<br>\n
     * MENSAJE2<br>\n
     * ...
     *
     */
    public String getHtmlErrorMessage (boolean allErrors) {

        StringBuilder res = new StringBuilder();

        if (errorList.size()>0) {

            boolean ok = false;
            int cont1 = 0;            
            while (cont1<errorList.size() && !ok) {

                ErrorItem error = (ErrorItem) errorList.get(cont1);
                res.append (error.getHtmlErrorMessage());
                if (!allErrors) { ok = true; }
                cont1++;
            }            
        }

        return res.toString();
    }
    public String getHtmlErrorMessage () { return getHtmlErrorMessage(false); }

    /**
     * Obtiene la descripción de los errores con formato XML:
     *
     * <result>ERROR</result>
     * <code>errorCode</code>
     * <messageKey1>messageValue1</messageKey1>
     * <messageKey2>messageValue2</messageKey2>
     * ...
     *
     */
    public String getXmlErrorMessage (boolean allErrors) {

        StringBuilder res = new StringBuilder();

        if (errorList.size()>0) {

            int cont1 = 0;
            if (!allErrors) { cont1 = errorList.size()-1; }
            while (cont1<errorList.size()) {

                ErrorItem error = (ErrorItem) errorList.get(cont1);
                res.append (error.getXmlErrorMessage());
                cont1++;
            }
        }
        else { res.append ("<result>OK</result>"); }

        return res.toString();
    }
    public String getXmlErrorMessage () { return getXmlErrorMessage(false); }

    /**
     * Obtiene el valor del mensaje con clave "paramName" del último error
     * producido.
     *
     * Si no se ha producido ningún error, o este no tiene ninguna pareja
     * "variable/valor" con la clave indicada devuelve un valor "null".
     */
    public String getErrorParam (String paramName) {
        
        String res = null;

        ErrorItem error = (ErrorItem) (errorList.isEmpty() ? null : errorList.firstElement());
        if (error!=null) { res = error.getErrorParam (paramName); }

        return StringUtils.notNull (res).trim();
    }


    //   __  __      _            _                     _     _ _
    //  |  \/  |    | |          | |                   | |   | |_)
    //  | \  / | ___| |_ ___   __| | ___    _ __  _   _| |_  | |_  ____ ___
    //  | |\/| |/ _ \ __) _ \ / _` |/ _ \  | `_ \| | | | `_ \| | |/ ___) __)
    //  | |  | |  __/ |_ (_) | (_| |  __/  | |_) | |_| | |_) | | | (___\__ \
    //  |_|  |_|\___|\__)___/ \__,_|\___|  |  __/ \__,_|_,__/|_|_|\____)___/
    //                                     | |
    //                                     |_|

    public void clearErrors() { errorList.clear(); }


    /**
     * Añade un error a la lista de errores.
     *
     * Ver "setError(String, String...)";
     */
    public void addError (String errorCode, String... errorMsg) { setError (errorCode, errorMsg); }

    
    /**
     * Añade una pareja "variable/valor" al último error producido.
     */
    public void addParam (String errorMsg) {

        ErrorItem error = (ErrorItem) (errorList.isEmpty() ? null : errorList.lastElement());
        if (error!=null) { error.addErrorMsg (errorMsg.trim()); }
    }

}


