//    _____ _                       _  _  ______                   _____ _             _  _
//   / ____) |                     | || ||  ____)                 (_   _) |           | || |
//  | |    | |  _ _ ___ ___  ___   |_||_|| |__   _ __ _ __ ___  _ __| | | |_ ___ _ _ ___||_|
//  | |    | |/ _` | __) __)/ _ \        |  __) | `__) `__) _ \| `__) | | __) _ \ ` `  \
//  | |____| | (_| |__ \__ \  __/        | |____| |  | | | (_) | | _| |_| |_  __/ |\ |\ |
//   \_____)_|\__,_|___/___/\___|        |______)_|  |_|  \___/|_|(_____)\__)___|_||_||_|
//
//


package cat.lib.errors;

import cat.lib.utils.StringUtils;
import cat.lib.xml.XmlUtils;
import java.util.Enumeration;
import java.util.Vector;

/**
 *
 * LA CLASE "ErrorItem":
 * ---------------------
 *
 *   La clase "ErrorItem" almacena la información relativa a un error.
 *
 *   Esta clase es de caracter interno y la utiliza la clase "ErrorEx" para
 * almacenar la lista de errores que se van produciendo.
 *
 *   Cada error se identifica por un código "erroCode" y permite tener una lista
 * de mensajes relativos al mismo compuestos por un nombre de mensaje (o clave) y
 * su contenido, separados por el caracter ":".
 *
 *   El separador ":" es importante para cuando se desea obtener un mensaje
 * a partir de su clave.
 *
 *
 *
 * @author Miquel Abadal i Cervià
 */
class ErrorItem {

    //     __   _        _ _           _
    //   / __ \| |      (_) |         | |
    //  | |__| | |_ _ __ _| |_   _   _| |_ ___
    //  |  __  | __) `__) | `_ \| | | | __) __)
    //  | |  | | |_| |  | | |_) | |_| | |_\__ \
    //  |_|  |_|\__)_|  |_|_,__/ \__,_|\__)___/
    //
    //

    private String errorCode = "";
    private Vector errorMsg = null;

    //   __  __      _            _                              _                    _
    //  |  \/  |    | |          | |                            | |                  | |
    //  | \  / | ___| |_ ___   __| | ___     ____ ___  _ __  ___| |_ _ __ _   _  ____| |_ ___  _ __
    //  | |\/| |/ _ \ __) _ \ / _` |/ _ \   / ___) _ \| `_  \ __) __) `__) | | |/ ___) __) _ \| `__)
    //  | |  | |  __/ |_ (_) | (_| |  __/  | (___ (_) | | | |__ \ |_| |  | |_| | (___| |_ (_) | |
    //  |_|  |_|\___|\__)___/ \__,_|\___|   \____)___/|_| |_|___/\__)_|   \__,_|\____)\__)___/|_|
    //
    //


    protected ErrorItem () { errorMsg = new Vector(); }


    //   __  __      _            _             _  _   _____ ______ _______ _  _
    //  |  \/  |    | |          | |           | || | / ____)  ____)__   __) || |
    //  | \  / | ___| |_ ___   __| | ___ ___   |_||_|| (___ | |__     | |  |_||_|
    //  | |\/| |/ _ \ __) _ \ / _` |/ _ \ __)         \___ \|  __)    | |
    //  | |  | |  __/ |_ (_) | (_| |  __/__ \         ____) | |____   | |
    //  |_|  |_|\___|\__)___/ \__,_|\___|___/        (_____/|______)  |_|
    //
    //

    protected void setErrorCode (String errorCode) { this.errorCode = errorCode; }
    protected void addErrorMsg (String errorMsg) { this.errorMsg.add (errorMsg); }


    //   __  __      _            _             _  _   _____ ______ _______ _  _
    //  |  \/  |    | |          | |           | || | / ____)  ____)__   __) || |
    //  | \  / | ___| |_ ___   __| | ___ ___   |_||_|| |  __| |__     | |  |_||_|
    //  | |\/| |/ _ \ __) _ \ / _` |/ _ \ __)        | | (_ |  __)    | |
    //  | |  | |  __/ |_ (_) | (_| |  __/__ \        | |__| | |____   | |
    //  |_|  |_|\___|\__)___/ \__,_|\___|___/         \_____|______)  |_|
    //
    //

    protected String getErrorCode () { return errorCode; }

    /**
     * Obtiene la descripción del error con la siguiente estructura:
     *
     * ERROR: errorCode\n
     * MENSAJE1\n
     * MENSAJE2\n
     * ...
     *
     */
    protected String getErrorMessage () {

        StringBuilder res = new StringBuilder();

        res.append ("ERROR: ");
        res.append (errorCode);

        Enumeration e = errorMsg.elements();
        while (e.hasMoreElements()) {
            res.append ("\n");
            res.append ((String) e.nextElement());
        }

        return res.toString();
    }

    /**
     * Obtiene la descripción del error con formato HTML:
     *
     * <p>[ERROR: errorCode]<br>\n
     * MENSAJE1<br>\n
     * MENSAJE2<br>\n
     * ...
     *
     */
    protected String getHtmlErrorMessage () {

        StringBuilder res = new StringBuilder();
        res.append ("<p>[ERROR: ");
        res.append (errorCode);
        res.append ("]<BR>");

        Enumeration e = errorMsg.elements();
        while (e.hasMoreElements()) {
            res.append("\n[");
            res.append((String) e.nextElement());
            res.append("]<BR>");
        }
        res.append ("</p>");

        return res.toString();
    }

    /**
     * Obtiene la descripción del error con formato XML:
     *
     * <result>ERROR</result>
     * <code>errorCode</code>
     * <messageKey1>messageValue1</messageKey1>
     * <messageKey2>messageValue2</messageKey2>
     * ...
     *
     */
    protected String getXmlErrorMessage () {

        StringBuilder res = new StringBuilder();
        res.append ("<result>ERROR</result><code>");
        res.append (errorCode);
        res.append ("</code>");

        Enumeration e = errorMsg.elements();
        while (e.hasMoreElements()) {
            String message = (String) e.nextElement();

            int pos = StringUtils.charPos (message,':');
            if (pos==-1) { pos = message.length(); }

            String key = StringUtils.copy (message, 0, pos).trim().toLowerCase();
            String value = StringUtils.copy (message, pos+1, message.length()-pos-1).trim();
            res.append ('<');
            res.append (key);
            res.append ('>');
            res.append (XmlUtils.encodeXml (value));
            res.append ("</");
            res.append (key);
            res.append ('>');
        }

        return res.toString();
    }

    /**
     * Obtiene el valor del mensaje con clave "paramName".
     *
     * Si no lo encuentra ningun mensaje con la clave indicada devuelveun valor
     * "null".
     */
    public String getErrorParam (String paramName) {

        String res = null;

        Enumeration e = errorMsg.elements();
        while (e.hasMoreElements()) {
            String message = (String) e.nextElement();

            int pos = StringUtils.charPos (message, ':');
            if (pos>=0) {
                String param = StringUtils.copy (message, pos);
                if (param.equalsIgnoreCase (paramName)) { res = StringUtils.copy (message, pos+1, message.length()-pos); }
            }
        }


        return StringUtils.notNull (res);
    }
}
