//    _____ _                       _  _  _                     _      _    _ _   _ _    _  _
//   / ____) |                     | || || |                   | |    | |  | | | (_) |  | || |
//  | |    | |  _ _ ___ ___  ___   |_||_|| |    ___   ____  _ _| | ___| |  | | |_ _| |____||_|
//  | |    | |/ _` | __) __)/ _ \        | |   / _ \ / ___) _` | |/ _ \ |  | | __) | | __)
//  | |____| | (_| |__ \__ \  __/        | |___ (_) | (___ (_| | |  __/ |__| | |_| | |__ \
//   \_____)_|\__,_|___/___/\___|        |_____)___/ \____)__,_|_|\___|\____/ \__)_|_|___/
//
//

package cat.lib.locale;

import cat.lib.strings.ParamsList;
import cat.lib.utils.SetUtils;
import cat.lib.utils.StringUtils;
import java.util.Locale;
import java.util.MissingResourceException;

/**
 *
 * @author Miquel Abadal i Cervià
 */
public class LocaleUtils {


    // Dos taulas vinculades amb els codis ISO2 y ISO3 dels idiomes que els tenen diferents.
    private static String[] ISO2 = new String[] {"alb", "arm", "baq", "bur", "chi", "cze", "dut", "fre", "geo", "ger", "ice",
                                                 "mac", "may", "mao", "per", "rum", "slo", "tib", "wel"};

    private static String[] ISO3 = new String[] {"sqi", "hye", "eus", "mya", "zho", "ces", "nld", "fra", "kat", "deu", "isl",
                                                 "mkd", "msa", "mri", "fas", "ron", "slk", "bod", "cym"};

    public static boolean isValidLanguageCode (Locale locale, String language) {

        boolean res = false;        
        if (locale!=null) {
            if (language!=null) {
                language = language.toLowerCase();
                if (getISO1Language (locale).equals(language)) { res = true; }
                else if (getISO3Language (locale).equals(language)) { res = true; }
                else if (getISO2Language (locale).equals(language)) { res = true; }
            }
            else { res = true; }
        }
        return res;
    }

    public static boolean isValidCountryCode (Locale locale, String contry) {

        boolean res = false;        
        if (locale!=null) {
            if (contry!=null) {
                contry = contry.toUpperCase();
                if (getISO1Country (locale).equals(contry)) { res = true; }
                else if (getISO3Country (locale).equals(contry)) { res = true; }
            }
            else { res = true; }
        }
        return res;
    }

    public static boolean isValidVariantCode (Locale locale, String variant) {

        boolean res = false;        
        if (locale!=null) {
            if (variant!=null) {
                variant = variant.toLowerCase();
                if (getVariant (locale).equals(variant)) { res = true; }
            }
            else { res = true; }
        }
        return res;
    }

    public static boolean isValidLocaleCode (Locale locale, String code) {

        boolean res = false;

        ParamsList list = new ParamsList(code, '-');
        String language = list.getParam(0);
        String country = list.getParam(1);
        String variant = list.getParam(2);
        if (isValidLanguageCode (locale, language)) {
            if (StringUtils.isEmpty (country) || isValidCountryCode (locale, country)) {
                if (StringUtils.isEmpty (variant) || isValidVariantCode (locale, variant)) { res = true; }
            }
        }
        return res;
    }

    /**
     * Obtiene el objeto Locale correspondiente al idioma "language" y al pais
     * "country"
     *
     * Detecta si el código de idioma indicado, corresponde a la codificación ISO3,
     * devoelviendo el Locale adecuado.
     *
     * Si no existe ningun Locale para el idioma indicado, lo crea.
     */
    public static Locale getLocale (String language, String country, String variant, boolean createIfNoExist) {

        Locale locale = null;

        if (!StringUtils.isEmpty (language)) {
            Locale[] locales = Locale.getAvailableLocales();
            int cont1 = 0;
            while (locale==null && cont1<locales.length) {

                if (isValidLanguageCode (locales[cont1], language)) {
                    language = getISO1Language (locales[cont1]);

                    if (isValidCountryCode (locales[cont1], country)) {
                        country = getISO1Country (locales[cont1]);
                        locale = new Locale (language, StringUtils.notNull (country), StringUtils.notNull (variant));
                    }
                }
                cont1++;
            }
        }

        if (locale==null && createIfNoExist) { locale = new Locale (language, StringUtils.notNull (country), StringUtils.notNull (variant)); }
        
        return locale;
    }

    public static Locale getLocale (String localeCode, boolean createIfNoExist) {
        ParamsList list = new ParamsList(localeCode, '_');
        String language = list.getParam(0);
        String country = list.getParam(1);
        String variant = list.getParam(2);
        return getLocale (language, country, variant, createIfNoExist);
    }


    /**
     * Obtiene la lista de idiomas aceptados por la petición HTTP.
     *
     * NOTA: Las peticiones HTTP incorporan en la cabecera la etiqueta
     * "accept-language" con la lista de idiomas y sus variantes geográficas
     * que acepta el usuario.
     *
     * Esta etiqueta tiene una lista de entradas sepadas por el caracter ","
     * con el siguiente formato:
     *
     * |------------------------------------|
     * | idioma-localidad;q=rango , .....   |
     * |------------------------------------|
     *
     * Sólo el campo idioma es obligatorio, los otros son opcionales.
     *
     * Pero como hay distintos standares de codificación de idiomas y cada navegador
     * eligue el standart a utilizar, puede ser complejo identificar algunos
     * idiomas, por lo tanto, este método, detecta el standart de codificación
     * y lo convierte en ISO 639-3, el cual codifica cada idioma con una secuencia
     * de 3 caracteres.
     *
     * Este método, pues, nos devuelve la lista de idiomas (en standart ISO 639-3)
     * que el usuario solicita, siendo el primero el preferido.
     *
     * NOTA: Recuerde, que el formato ISO 639-3 utiliza tres caracteres para
     * codificar los idiomas, por lo tanto no debemos esperar recibir los
     * códigos "en", "es", "it", "fr", "de"... utilizados habitualmente.
     *
     * En su lugar, obtendremos los códigos "eng", "spa", "ita", "fra" o "deu",
     * respectivamente.
     *
     * Puede consultar la tabla de códigos en:
     * http://en.wikipedia.org/wiki/List_of_ISO_639-1_codes
     *
     */
    public static Locale[] getHttpLocales (String httpAcceptLanguages) {

        ParamsList list = new ParamsList (httpAcceptLanguages, ',');
        Locale[] res = new Locale[list.paramCount()];

        for (int cont1=0; cont1<list.paramCount(); cont1++) {

            String language = StringUtils.copy (list.getParam(cont1), '-', StringUtils.CASE_INSENSITIVE, false);
            res[cont1] = getLocale (language, true);
        }

        return res;
    }

    public static Locale getHttpFavoriteLocale (String httpAcceptLanguages) {

        ParamsList list = new ParamsList (httpAcceptLanguages, ',');
        Locale locale = null;

        if (list.paramCount()>0) {
            String languageCountry = StringUtils.copy (list.getParam(0), ';', StringUtils.CASE_INSENSITIVE, false);

            String language = StringUtils.getParam (languageCountry, '-', 0);
            String country = StringUtils.getParam (languageCountry, '-', 1);

            if (language.equalsIgnoreCase("es") && country.isEmpty()) { language = language + "_ES"; }

            locale = getLocale (language, true);
        }

        return locale;
    }

    public static String getISO1Language (Locale locale) {
        String res = null;
        if (locale!=null) { res = locale.getLanguage(); }
        return res;
    }

    public static String getISO2Language (Locale locale) {

        String iso2 = null;
        String iso3 = getISO3Language (locale);

        int index = SetUtils.searchStringInArray (iso3, StringUtils.CASE_INSENSITIVE, ISO3);
        if (index!=-1) { iso2 = ISO2[index]; }
        else { iso2 = iso3; }

        return iso2;        
    }

    /**
     * Obtiene el código de idioma en ISO3 del Locale indicado.
     *
     * Si este Locale no dispone de dicha información, devuelve el código de IDIOMA
     * en ISO1 sin generar ninguna excepción.
     */
    public static String getISO3Language (Locale locale) {
        String res = null;
        if (locale!=null) {
            try { res = locale.getISO3Language(); }
            catch (MissingResourceException e) { res = locale.getLanguage(); }
        }
        return res;
    }

    public static String getISO1Country (Locale locale) {
        String res = null;
        if (locale!=null) { res = locale.getCountry(); }
        return res;
    }

    public static String getISO3Country (Locale locale) {
        String res = null;
        if (locale!=null) {
            try { res = locale.getISO3Country(); }
            catch (MissingResourceException e) { res = locale.getLanguage(); }
        }
        return res;
    }

    public static String getVariant (Locale locale) {
        String res = null;
        if (locale!=null) { res = locale.getVariant(); }
        return res;
    }

    public static String getLanguageName (Locale locale, Locale inLocale) {

       String res = null;
        if (locale!=null && inLocale!=null) { res = StringUtils.firstCharToUpperCase (locale.getDisplayLanguage (inLocale));  }
        return res;
    }

    public static String getLanguageName (String language, Locale inLocale) {

        String res = null;

        Locale locale = LocaleUtils.getLocale (language, false);
        if (locale!=null && inLocale!=null) { res = StringUtils.firstCharToUpperCase (locale.getDisplayLanguage (inLocale));  }
        else { res = StringUtils.firstCharToUpperCase (language);  }

        return res;
    }

}
