//    _____ _                       _  _  ____                          _  _
//   / ____) |                     | || ||  __ \                       | || |
//  | |    | |  _ _ ___ ___  ___   |_||_|| |__) |___  _ _ __    _ _ _ ___||_|
//  | |    | |/ _` | __) __)/ _ \        |  _  // _ \| ` `  \ / _` | `_  \
//  | |____| | (_| |__ \__ \  __/        | | \ \ (_) | |\ |\ | (_| | | | |
//   \_____)_|\__,_|___/___/\___|        |_|  \_\___/|_||_||_|\__,_|_| |_|
//
//

package cat.lib.math;



/**
 * La classe <code>Roman</code> encapsula (classe wrapper) un valor codificat
 * utilitzant el sistema numéric <code>Romà</code> en un objecte. Un objecte
 * de la classe <code>Roman</code> conté un atribut que enmagatzema el valor
 * codificat en <code>Romà</code> utilitzant una instància del tipus
 * <code>String</code> i un altre atribut que enmagatzema el valor convertit
 * a codificació decimal.
 *
 *  <p>
 *
 * A més a més, aquesta classe aporta diversos métodes per a convertir a
 * <code>Romà</code> una cadena de text <code>String</code> o un valor sencer
 * positiu superior a zero; així com els métodes per a convertir un valor
 * <code>Romà</code> a <code>int</code>..
 *
 *  <p>
 *
 * <code>Nota:</code>Cal recordar que els números romans es codifiquen seguint
 * la següent taula:
 *
 * |------------------------------------------------|
 * |  M   |  D   |  C   |  L   |  X   |   V  |   I  |
 * |------------------------------------------------|
 * | 1000 | 500  | 100  |  50  |  10  |   5  |   1  |
 * |------------------------------------------------|
 *
 * Amb la restricció de que els caràcters han d'estar ordenats de més valor a
 * meny. Tanmateix, es permeten certes excepcions en les que es permet altera
 * el ordre de manera que el valor inferior es resta del valor superior.
 *
 * Aquestes excepcions són limitades i es mostren en la següent taula:
 *
 * |-----------------------------------------|
 * | CM   | CD   |  XC  |  XL  |  IX  |  IV  |
 * |-----------------------------------------|
 * | 900  | 400  |  90  |  40  |  9   |  4   |
 * |-----------------------------------------|
 *
 * Finalment, els caracters amb un valor que comença per 1 (M, C, X, I) es poden
 * repetir un màxim de tres cops.
 *
 *
 * @author  Miquel Abadal i Cervià
 *          Col.laboració: Marc Viñas
 * @version 1.0, 06/12/09
 * @since JDK1.0
 */
public final class Roman extends Number implements Comparable<Roman> {

    //     __   _        _ _           _
    //   / __ \| |      (_) |         | |
    //  | |__| | |_ _ __ _| |_   _   _| |_ ___
    //  |  __  | __) `__) | `_ \| | | | __) __)
    //  | |  | | |_| |  | | |_) | |_| | |_\__ \
    //  |_|  |_|\__)_|  |_|_,__/ \__,_|\__)___/
    //
    //

    static String [][] lletres_MARC ={{"","I","II","III","IV","V","VI","VII","VIII","IX"},
				      {"","X","XX","XXX","XL","L","LX","LXX","LXXX","XC"},
				      {"","C","CC","CCC","CD","D","DC","DCC","DCCC","CM"},
				      {"","M","MM","MMM"}};

    private static String[] lletres = {"Z",  "MZ", "R", "MR", "M",  "CM", "D", "CD", "C", "XC", "L", "XL", "X", "IX", "V", "IV", "I"};
    private static int[] valors =     {10000, 9000, 5000, 4000, 1000, 900,  500, 400,  100, 90,   50,  40,   10,  9,    5,   4,    1};

    private final String romanValue;
    private final int value;

    /**
     * <code>MIN_VALUE</code> és una constant que enmagatzema el menor número
     * expressable utilitzant codificació romana.
     */
    public static final int MIN_VALUE = 1;
    
    /**
     * <code>MIN_VALUE</code> és una constant que enmagatzema el major número 
     * expressable utilitzant codificació romana.
     */
    public static final int MAX_VALUE = 3999;

    /**
     * La constant <code>Class</code> representa la clase primitiva.
     */
    public static final Class<Integer>TYPE = Integer.TYPE;


    //   __  __      _            _                        _             _
    //  |  \/  |    | |          | |                      (_)           | |
    //  | \  / | ___| |_ ___   __| | ___ ___    _ __  _ __ _ _    _  _ _| |_ ___
    //  | |\/| |/ _ \ __) _ \ / _` |/ _ \ __)  | `_ \| `__) | |  | | _` | __) __)
    //  | |  | |  __/ |_ (_) | (_| |  __/__ \  | |_) | |  | |\ \/ / (_| | |_\__ \
    //  |_|  |_|\___|\__)___/ \__,_|\___|___/  |  __/|_|  |_| \__/ \__,_|\__)___/
    //                                         | |
    //                                         |_|

    /**
     * Busca a la taula de lletres alguna que contingui la cadena de text
     * indicada, i en retorna la seva posició o -1 si no en troba cap.
     *
     * @param   lletra Cadena de text a cercer.
     * @return  Index de la posició on ha trobat la lletra o -1 si no l'ha
     * trobada.
     */
    private static int searchLletra (String lletra) {

        int trobat = -1;
        int cont1 = 0;
        lletra = lletra.toUpperCase();
        while (trobat==-1 && cont1<lletres.length) {
            if (lletres[cont1].equals (lletra)) trobat = cont1;
            cont1++;
        }

        return trobat;
    }

    /**
     * Busca a la taula de lletres alguna que contingui el cartacter indicat,
     * i en retorna la seva posició o -1 si no en troba cap.
     *
     * @param   lletra Caràcter a cercer.
     * @return  Index de la posició on ha trobat el caràcter o -1 si no l'ha
     * trobada.
     */
    private static int searchLletra (char lletra) {

        int trobat = -1;
        int cont1 = 0;
        lletra = Character.toUpperCase(lletra);

        while (trobat==-1 && cont1<lletres.length) {
            if (lletres[cont1].length()==1 && lletres[cont1].charAt(0)==lletra) trobat = cont1;
            cont1++;
        }

        return trobat;
    }


    /**
     * Converteix una valor enter a <code>Romà</coda> utilitzant l'algoritme del
     * Mac Viñas. És més eficient que el meu, però no es pot configurar per que
     * pugui acceptar (opcionalment) números superiors a 3999.
     *
     * @param   value Valor a convertir en números romans.
     * @return  Cadena de text amb el número codificat en números romans.
     * @exception  RomanFormatException si el valor indicat no es pot convertir
     * 		   a números romans.
     */
    private static String intToRoman_Marc (int value) throws NumberFormatException {

        String res = "";
        if (value>=MIN_VALUE && value<=MAX_VALUE) {

            res = ((lletres_MARC[3][value/1000])+
                    lletres_MARC[2][(value%1000)/100]+
                    lletres_MARC[1][(value%100)/10]+
                    lletres_MARC[0][(value%10)]);
        }
        else {
            if (value==0) { throw new NumberFormatException ("Número fuera de rango, los números romanos no pueden codificar el número cero"); }
            else if (value<MIN_VALUE) { throw new NumberFormatException ("Número fuera de rango, los números romanos no pueden codificar números negativos"); }
            else if (value>MAX_VALUE) { throw new NumberFormatException ("Número fuera de rango, los números romanos no pueden codificar núeros superiores a \""+MAX_VALUE+"\""); }
        }

       return res;

    }

    /**
     * Converteix una valor enter a <code>Romà</coda>.
     *
     * @param   value Valor a convertir en números romans.
     * @return  Cadena de text amb el número codificat en números romans.
     * @exception  RomanFormatException si el valor indicat no es pot convertir
     * 		   a números romans.
     */
    private static String intToRoman (int value) throws NumberFormatException {

        StringBuffer res = new StringBuffer (10);

        if (value>=MIN_VALUE && value<=MAX_VALUE) {

            int index = 0;
            while (value>0) {
                
                boolean trobat = false;
                while (!trobat && index<lletres.length)  {
                    if (value>=valors[index]) {
                        res.append (lletres[index]);
                        value = value-valors[index];
                        trobat = true;
                    }
                    else index++;
                }
            }
        }
        else {
            if (value==0) { throw new NumberFormatException ("Número fuera de rango, los números romanos no pueden codificar el número cero"); }
            else if (value<MIN_VALUE) { throw new NumberFormatException ("Número fuera de rango, los números romanos no pueden codificar números negativos"); }
            else if (value>MAX_VALUE) { throw new NumberFormatException ("Número fuera de rango, los números romanos no pueden codificar núeros superiores a \""+MAX_VALUE+"\""); }
        }

        return res.toString();
    }

    /**
     * Converteix un cadena de text que conté un valor codificat en números
     * romans al seu valor enter.
     *
     * @param   text Cadena de text amb un valor codificat en números romans.
     * @return  El valor enter del número indicat.
     * @exception  RomanFormatException si el valor indicat no es un número
     * Romà vàlid.
     */
    private static int romanToInt (String text) throws NumberFormatException {

        int res = 0;
        int error = 0;

        if (text!=null) {
            text = text.trim().toUpperCase();

            if (text.length()>0) {

                int cont1 = 0;
                int incr = 1;
                int index = -1;
                int lastValor = valors[0]+1;
                int lastIndex = -1;
                int valor;

                while (error==0 && cont1<text.length()) {

                    // Busca una convinacio doble de valors i si no el troba, en busca una de simple.
                    // A "index" guarda la combinació vàlida que ha trobat.
                    // Si val -1 vol dir que la combinacio no és vàlida.
                    // A "incr" guarda el increment a aplicar al "cont1" del bucle, i val 1 si la combinació
                    // correspon a un únic caràcter o 2 si és una combinació de dos caràcters.
                    incr = 1;
                    if (text.length()>cont1+1) { index = searchLletra (text.substring (cont1, cont1+2)); }
                    else index = -1;

                    if (index==-1) { index = searchLletra (text.charAt(cont1)); }
                    else { incr = 2; }


                    // Si ha trobat una parella de caràcters o un caràcter individual vàlid/es...
                    // Sino error.
                    if (index!=-1) {

                        valor = valors[index];

                        // Si el valor obtingut més el anterior valor obtingut són més grans que
                        // el valor de la combinació inmediatament superior al anterior valor,
                        // llavors es tracta d'una combinació erronea:
                        //
                        // CMC = 900 + 100 = 1000 = M
                        // CMD = 900 + 500 = 1400 = MCD
                        // IXI = 9 + 1 = 10 = X
                        if (lastIndex>0 && valor+lastValor>=valors[lastIndex-1]) { error = 3; }

                        // Si es repeteix el últim valor, llavors s'acumula per a evitar que es pugui repetir
                        // fins a superar o igualar el valor de la combinacio inmediatament superior.
                        //
                        // VV = 10 = X
                        // IIII = 4 = IV
                        if (index==lastIndex) { lastValor = lastValor + valor; }
                        else { lastValor = valor; }
                        lastIndex = index;

                        // Suma el valor de la combinacio trobada al total.
                        res = res + valor;
                    }
                    else error = 4;

                    cont1 = cont1 + incr;
                }
            }
            else error = 2;
        }
        else error = 1;

        switch (error) {
            case 1 : throw new NumberFormatException ("Número fuera de rango, un puntero nulo no és un valor codificable");
            case 2 : throw new NumberFormatException ("Número fuera de rango, la cadena en blanco no és un valor codificable");
            case 3 : throw new NumberFormatException ("Combinación incorrecta \""+text+"\"");
            case 4 : throw new NumberFormatException ("Caracter no reconocido \""+text+"\"");
        }

        return res;
    }

    //   __  __      _            _                      _        _   _
    //  |  \/  |    | |          | |                    | |      | | (_)
    //  | \  / | ___| |_ ___   __| | ___ ___     ___ ___| |_  _ _| |_ _  ____ ___
    //  | |\/| |/ _ \ __) _ \ / _` |/ _ \ __)   / _ \ __) __) _` | __) |/ ___) __)
    //  | |  | |  __/ |_ (_) | (_| |  __/__ \  |  __/__ \ |_ (_| | |_| | (___\__ \
    //  |_|  |_|\___|\__)___/ \__,_|\___|___/   \___|___/\__)__,_|\__)_|\____)___/
    //
    //

    /**
     * Converteix un cadena de text que conté un valor codificat en números
     * romans i en retorna el seu valor amb <code>String</code>.
     *
     * @param   text Cadena de text amb un valor codificat en números romans.
     * @return  Cadena de text amb el valor codificat en números romans.
     * @exception  RomanFormatException si el valor indicat no es un número
     * Romà vàlid.
     */
    public static String toString (String  value) throws NumberFormatException {
        romanToInt (value);  // Valida que el numero entrat sigui vàlid.
        return value;
    }


    /**
     * Converteix un cadena de text que conté un valor codificat en números
     * romans a una cadena de text que conté aquest valor amb codificació decimal.
     *
     * @param   text Cadena de text amb un valor codificat en números romans.
     * @return  El valor convertir a números decimals.
     * @exception  RomanFormatException si el valor indicat no es un número
     * Romà vàlid.
     */
    public static String toDecimanString (String  value) throws NumberFormatException {
        return Integer.toString (romanToInt (value));
    }

    /**
     * Converteix un cadena de text que conté un valor codificat en números
     * romans a una cadena de text que conté aquest valor amb codificació
     * hexadecimal.
     *
     * @param   text Cadena de text amb un valor codificat en números romans.
     * @return  El valor convertir a números hexadecimals.
     * @exception  RomanFormatException si el valor indicat no es un número
     * Romà vàlid.
     */
    public static String toHexString (String  value) throws NumberFormatException {
	return Integer.toHexString (romanToInt (value));
    }

    /**
     * Converteix un cadena de text que conté un valor codificat en números
     * romans a una cadena de text que conté aquest valor amb codificació octal.
     *
     * @param   text Cadena de text amb un valor codificat en números romans.
     * @return  El valor convertir a números octals.
     * @exception  RomanFormatException si el valor indicat no es un número
     * Romà vàlid.
     */
    public static String toOctalString (String  value) throws NumberFormatException {
	return Integer.toOctalString (romanToInt (value));
    }

    /**
     * Converteix un cadena de text que conté un valor codificat en números
     * romans a una cadena de text que conté aquest valor amb codificació
     * binaria.
     *
     * @param   text Cadena de text amb un valor codificat en números romans.
     * @return  El valor convertir a números binaris.
     * @exception  RomanFormatException si el valor indicat no es un número
     * Romà vàlid.
     */
    public static String toBinaryString  (String  value) throws NumberFormatException {
	return Integer.toBinaryString (romanToInt (value));
    }

    /**
     * Converteix un cadena de text que conté un valor codificat en números
     * decimals i el converteix a una cadena que conté el mateix nombre però
     * codificat amb números romans.
     *
     * @param   text Cadena de text amb un valor codificat en números decimals.
     * @return  El valor convertir a números romans.
     * @exception  RomanFormatException si el valor indicat no es un número
     * decimal vàlid o no es pot codificar en números romans.
     */
    public static String parseRoman (String  value) throws NumberFormatException {
        int res = 0;
        try { res = Integer.parseInt (value); }
        catch (NumberFormatException e) {throw new NumberFormatException ("Numero invalido \""+value+"\""); }

	return intToRoman (res);
    }

    /**
     * Converteix un valor enter a una cadena que conté el mateix nombre però
     * codificat amb números romans.
     *
     * @param   text Cadena de text amb un valor codificat en números decimals.
     * @return  El valor convertir a números romans.
     * @exception  RomanFormatException si el valor indicat no es pot codificar
     * en números romans.
     */
    public static String parseRoman  (int value) throws NumberFormatException {
	return intToRoman (value);
    }

    /**
     * Converteix el valor enmagatzemat a la cadena <code>value</code>, codificat
     * amb números romans, al seu corresponent valor enter.
     *
     * @param   text Cadena de text amb un valor codificat en números romans.
     * @return  El valor convertir a números enters.
     * @exception  RomanFormatException si el valor indicat no es un número
     * romà vàlid.
     */
    public static int parseInt (String value) throws NumberFormatException {
	return romanToInt (value);
    }

    /**
     * Crea una instància de tipus <code>Roma</code> assignant-li el valor
     * <code>value</code>, tenint en comte que aquest valor ha de venir
     * previament codificat en números romans.
     *
     * @param   value Cadena de text amb un valor codificat en números romans.
     * @return  Una instància de la classe <code>Roma</code> que encapsula el
     * valor indicat.
     * @exception  RomanFormatException si el valor indicat no es un número romà
     * vàlid.
     */
    public static Roman valueOf  (String  value) throws NumberFormatException {
	return new Roman (value);
    }

    /**
     * Crea una instància de tipus <code>Roma</code> assignant-li el valor
     * <code>value</code>.
     *
     * @param   value Valor enter.
     * @return  Una instància de la classe <code>Roma</code> que encapsula el
     * valor indicat.
     * @exception  RomanFormatException si el valor indicat no es pot codificar
     * a núeros romans.
     */
    public static Roman valueOf  (int value) throws NumberFormatException {
	return new Roman (value);
    }


    //   __  __      _            _                              _                    _
    //  |  \/  |    | |          | |                            | |                  | |
    //  | \  / | ___| |_ ___   __| | ___     ____ ___  _ __  ___| |_ _ __ _   _  ____| |_ ___  _ __
    //  | |\/| |/ _ \ __) _ \ / _` |/ _ \   / ___) _ \| `_  \ __) __) `__) | | |/ ___) __) _ \| `__)
    //  | |  | |  __/ |_ (_) | (_| |  __/  | (___ (_) | | | |__ \ |_| |  | |_| | (___| |_ (_) | |
    //  |_|  |_|\___|\__)___/ \__,_|\___|   \____)___/|_| |_|___/\__)_|   \__,_|\____)\__)___/|_|
    //
    //

    /**
     * Métode constructor que crea una instància de tipus <code>Roma</code> i
     * li assigna el valor <code>value</code>.
     *
     * @param   value Valor enter.
     * @return  Una instància de la classe <code>Roma</code> que encapsula el
     * valor indicat.
     * @exception  RomanFormatException si el valor indicat no es pot codificar
     * a núeros romans.
     */
    public Roman (int value) throws NumberFormatException {
        this.romanValue = intToRoman (value);
        this.value = value;
    }

    /**
     * Métode constructor que crea una instància de tipus <code>Roma</code> i
     * li assigna el valor que conté la cadena de text <code>value</code>.
     *
     * @param   value Cadena de text que codifica un número utilitzant números
     * romans.
     * @return  Una instància de la classe <code>Roma</code> que encapsula el
     * valor indicat.
     * @exception  RomanFormatException si el valor indicat no es un número
     * Romà vàlid.
     */
    public Roman (String value) throws NumberFormatException {
        this.romanValue = value;
        this.value = romanToInt (value);
    }


    //   __  __      _            _                     _     _ _
    //  |  \/  |    | |          | |                   | |   | |_)
    //  | \  / | ___| |_ ___   __| | ___    _ __  _   _| |_  | |_  ____ ___
    //  | |\/| |/ _ \ __) _ \ / _` |/ _ \  | `_ \| | | | `_ \| | |/ ___) __)
    //  | |  | |  __/ |_ (_) | (_| |  __/  | |_) | |_| | |_) | | | (___\__ \
    //  |_|  |_|\___|\__)___/ \__,_|\___|  |  __/ \__,_|_,__/|_|_|\____)___/
    //                                     | |
    //                                     |_|


    /**
     * Retorna el valor que encapsula la instància utilitzant una variable de
     * tipus <code>byte</code>.
     */
    public byte byteValue()  {
        return (byte) value;
    }

    /**
     * Retorna el valor que encapsula la instància utilitzant una variable de
     * tipus <code>short</code>.
     */
    public short shortValue() {
	return (short) value;
    }

    /**
     * Retorna el valor que encapsula la instància utilitzant una variable de
     * tipus <code>int</code>.
     */
    public int intValue() {
	return value;
    }

    /**
     * Retorna el valor que encapsula la instància utilitzant una variable de
     * tipus <code>long</code>.
     */
    public long longValue() {
	return (long) value;
    }

    /**
     * Retorna el valor que encapsula la instància utilitzant una variable de
     * tipus <code>float</code>.
     */
    public float floatValue() {
	return (float) value;
    }

    /**
     * Retorna el valor que encapsula la instància utilitzant una variable de
     * tipus <code>double</code>.
     */
    public double doubleValue() {
	return (double) value;
    }

    /**
     * Retorna el valor que encapsula la instància utilitzant una variable de
     * tipus <code>String</code>.
     */
    public String toString() {
	return romanValue;
    }

    public int hashCode() {
	return value;
    }

    /**
     * Compara el valor que encapsula la instància de classe <code>Roma</code>
     * amb el que encapsula l'objecte indicat.
     *
     * Si contenen exactament el mateix valor retorna <code>true</code> i
     * <code>false</code> en cas contrari.
     *
     * @param   obj   El objecte amb el qual cal realitzar la comparació.
     * @return  <code>true</code> Si el objecte conté la mateixa informació;
     * <code>false</code> en cas contrari.
     */
    public boolean equals (Object obj) {
        boolean res = false;
	if (obj instanceof Roman) { res = (intValue()==((Roman)obj).intValue()); }
	return res;
    }

    /**
     * Compara el valor que encapsula la instància actual amb el que encapsula
     * la instància <code>anotherRoman</code>.
     *
     * Si el valor de la instància <code>anotherRoman</code> és més gran retorna
     * un valor a +1; si és més petita retorna un valor a -1; i si es igual,
     * retorna un zero.
     *
     * @param   anotherRoman   Instància amb la qual cal realitzar la comparació.
     * @return  <code>-1</code> Si anotherRoman és més gran; <code>+1</code> si
     * és més petit; o <code>0</code> si són iguals.
     */
    public int compareTo (Roman anotherRoman) {
	int anotherValue = anotherRoman.intValue();
	int res = 0;
        if (value<anotherValue) res = -1;
        else if (value>anotherValue) res = -1;
        return res;
    }


}
