//    _____ _                       _  _  ____                            _   _          ______      _  _
//   / ____) |                     | || ||  __ \                         | | (_)        |  ____)    | || |
//  | |    | |  _ _ ___ ___  ___   |_||_|| |__) | __ ___  _ __   ___ _ __| |_ _  ___ ___| |__  _   ___||_|
//  | |    | |/ _` | __) __)/ _ \        |  ___/ `__) _ \| `_ \ / _ \ `__) __) |/ _ \ __)  __) \ \/ /
//  | |____| | (_| |__ \__ \  __/        | |   | | | (_) | |_) |  __/ |  | |_| |  __/__ \ |____ )  (
//   \_____)_|\__,_|___/___/\___|        |_|   |_|  \___/|  __/ \___|_|   \__)_|\___|___/______)_/\_\
//                                                       | |
//                                                       |_|

package cat.lib.properties;

import cat.lib.errors.ErrorEx;
import cat.lib.files.FileUtils;
import cat.lib.utils.StringToNumber;
import cat.lib.utils.StringUtils;
import java.io.BufferedWriter;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Properties;


/**
 *
 *   _____ _                       _  _  ____                            _   _          ______      _  _
 *  / ____) |                     | || ||  __ \                         | | (_)        |  ____)    | || |
 * | |    | |  _ _ ___ ___  ___   |_||_|| |__) | __ ___  _ __   ___ _ __| |_ _  ___ ___| |__  _   ___||_|
 * | |    | |/ _` | __) __)/ _ \        |  ___/ `__) _ \| `_ \ / _ \ `__) __) |/ _ \ __)  __) \ \/ /
 * | |____| | (_| |__ \__ \  __/        | |   | | | (_) | |_) |  __/ |  | |_| |  __/__ \ |____ )  (
 *  \_____)_|\__,_|___/___/\___|        |_|   |_|  \___/|  __/ \___|_|   \__)_|\___|___/______)_/\_\
 *                                                      | |
 *                                                      |_|
 *
 * La classe "PropertiesEx" és una ampliació de la classe "Properties" que permet cercar les propietats
 * sense tenir en compte la seva tipografia.
 *
 * Recordem que la classe "Properties" quan carrega un fitxer de propietats amb el métode "load"
 * l'interpreta i genera un HashMap amb cadascun dels elements. Però aquest "HashMap" diferencia la
 * tipografia de les entrades de manera que la classe "Properties" no pot recuperar una entrada sino
 * està escrita EXACTAMENT amb la mateixa tipografia.
 *
 * Per a evitar-ho, s'ha creat aquesta classe, com a ampliació de la classe "Properties" però se li han
 * afeguit métodes capaços de localitzar les entrades sense tenir en compte la tipografia.
 *
 * El métode "getProperty" permet recuperar el valor associat amb una propietat.
 * El métode "containsKey" permet saber si la llista conté una propietat determinada.
 * El métode "containsValue" permet saber si existeix alguna propietat amb el valor indicat.
 *
 * Naturalment aquest métodes no són tan eficients com el métode "getProperty" original, però en canvi
 * permeten recuperar una propietat sense necessitat de conèixer-ne la tipografia.
 *
 * ATENCIÓ: Aquesta classe, pot plantejar problemes en cas de que el fitxer properties contingui dues
 * entrades iguals però amb tipografia diferent, ja que llavors els métodes que incorpora no seran
 * capaços de treballar amb una de les dues entrades.
 *
 * @author Miquel Abadal i Cervià
 */
public class PropertiesEx {

    //     __   _        _ _           _
    //   / __ \| |      (_) |         | |
    //  | |__| | |_ _ __ _| |_   _   _| |_ ___
    //  |  __  | __) `__) | `_ \| | | | __) __)
    //  | |  | | |_| |  | | |_) | |_| | |_\__ \
    //  |_|  |_|\__)_|  |_|_,__/ \__,_|\__)___/
    //
    //

    private boolean caseSensitive;
    private Properties properties = null;
    private String fileName = null;

    
    //   __  __      _            _                              _                    _
    //  |  \/  |    | |          | |                            | |                  | |
    //  | \  / | ___| |_ ___   __| | ___     ____ ___  _ __  ___| |_ _ __ _   _  ____| |_ ___  _ __
    //  | |\/| |/ _ \ __) _ \ / _` |/ _ \   / ___) _ \| `_  \ __) __) `__) | | |/ ___) __) _ \| `__)
    //  | |  | |  __/ |_ (_) | (_| |  __/  | (___ (_) | | | |__ \ |_| |  | |_| | (___| |_ (_) | |
    //  |_|  |_|\___|\__)___/ \__,_|\___|   \____)___/|_| |_|___/\__)_|   \__,_|\____)\__)___/|_|
    //
    //

    public PropertiesEx (InputStream inputStream, boolean caseSensitive, ErrorEx error) {

        this.caseSensitive = caseSensitive;
        this.properties = new Properties ();       
        this.fileName = null;

        if (inputStream!=null) {
            try { properties.load (inputStream); }
            catch (IOException e) { error.setError ("PROPERTYS_ERROR", "DESCRIPTION: No s'ha pogut carregar el fitxer de propietats", "FILENAME: "+fileName, "EXCEPTION: "+e.getMessage()); }
        }
        else {error.setError ("PROPERTYS_ERROR", "DESCRIPTION: No s'ha pogut carregar el fitxer de propietats", "STREAM: Null"); }
    }

    public PropertiesEx  (String fileName, boolean caseSensitive, ErrorEx error) {

        this.caseSensitive = caseSensitive;
        this.properties = new Properties ();
        this.fileName = fileName;

        try { properties.load (new FileInputStream (fileName)); }
        catch (IOException e) { error.setError ("PROPERTYS_ERROR", "DESCRIPTION: No s'ha pogut carregar el fitxer de propietats", "FILENAME: "+fileName, "EXCEPTION: "+e.getMessage()); }

    }


    //   __  __      _            _                        _             _
    //  |  \/  |    | |          | |                      (_)           | |
    //  | \  / | ___| |_ ___   __| | ___ ___    _ __  _ __ _ _    _  _ _| |_ ___
    //  | |\/| |/ _ \ __) _ \ / _` |/ _ \ __)  | `_ \| `__) | |  | | _` | __) __)
    //  | |  | |  __/ |_ (_) | (_| |  __/__ \  | |_) | |  | |\ \/ / (_| | |_\__ \
    //  |_|  |_|\___|\__)___/ \__,_|\___|___/  |  __/|_|  |_| \__/ \__,_|\__)___/
    //                                         | |
    //                                         |_|

    /**
     * Busca a la llista de propietats una entrada amb la clau "key". Rretorna
     * la clau i en cas de que no existeixi retorna un valor a "null".
     *
     * Si el atribut "caseSentitive" està desactivat, busca qualsevol clau amb
     * la cadena "key", encara que estigui escrita amb diferent tipografia, i en
     * retorna la clau amb la tipografia real.
     *
     * @param key
     * @return
     */
    protected String searchKey (String key) {

        String trobat = null;

        if (!caseSensitive) {
            Iterator i = getPropertiesList();
            while (trobat==null && i.hasNext()) {
                Map.Entry e = (Entry) i.next();
                String realkey = (String) e.getKey();
                if (StringUtils.equals (realkey, key, false)) { trobat = realkey; }
            }
        }
        else if (properties.containsKey (key)) { trobat = key; }


        return trobat;
    }

    /**
     * Busca a la llista de propietats la primera clau que té com a valor
     * la cadena "value" (tenint o no en compta la tipografia segons el valor
     * del paràmetre "caseSensitive") i en retorna la clau.
     *
     * Si no existeix cap entrada amb el valor indicat, retorna "null".
     *
     * @param key
     * @return
     */
    protected Map searchValueKeys (String value, boolean caseSensitive) {

        Map trobat = new HashMap ();

        Iterator i = getPropertiesList();
        while (i.hasNext()) {
            Map.Entry e = (Entry) i.next();
            String key = (String) e.getKey();
            String keyValue = (String) e.getValue();
            if (StringUtils.equals (keyValue, value, caseSensitive)) { trobat.put (key, value); }
        }

        return trobat;
    }

    //   __  __      _            _                         _     _ _
    //  |  \/  |    | |          | |                       | |   | |_)
    //  | \  / | ___| |_ ___   __| | ___ ___    _ __  _   _| |_  | |_  ____ ___
    //  | |\/| |/ _ \ __) _ \ / _` |/ _ \ __)  | `_ \| | | | `_ \| | |/ ___) __)
    //  | |  | |  __/ |_ (_) | (_| |  __/__ \  | |_) | |_| | |_) | | | (___\__ \
    //  |_|  |_|\___|\__)___/ \__,_|\___|___/  |  __/ \__,_|_,__/|_|_|\____)___/
    //                                         | |
    //                                         |_|


    public void save (String fileName, String encode, ErrorEx error) {
        try {
            String folder = FileUtils.extractFilePath (fileName);
            FileUtils.createFolder (folder, true, false, error);

            Writer out = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(fileName), encode));

            properties.store (out, null);
            this.fileName = fileName;
        }
        catch (FileNotFoundException ex) { error.setError ("IOEXCEPTION", "ACTION: SAVE", "TRACE:PropertiesEx.save", "DESCRIPTION: "+ex.getMessage(), "FILENAME: "+fileName); }
        catch (IOException ex) { error.setError ("IOEXCEPTION", "ACTION: SAVE", "TRACE:PropertiesEx.save", "DESCRIPTION: "+ex.getMessage(), "FILENAME: "+fileName); }
    }

    public void save (String fileName, ErrorEx error) { save (fileName, "ISO-8859-1", error); }
    public void save (ErrorEx error) { save (fileName, "ISO-8859-1", error); }

    /**
     * Obté un iterador per a recìrrer la llista de propietats.
     *
     * @return
     */
    public Iterator getPropertiesList () { 
        return properties.entrySet().iterator();
    }

    /**
     * Obté el valor enmagatzemat a la propietat amb la clau "key".
     *
     * Si la clau "key" té el valor "null" o no existeix cap entrada amb la clau
     * indicada, retorna el valor per defecte indicat "defaultValue".
     *
     * Si existeix una entrada amb la clau "key", en retorna el seu valor o una
     * cadena en blanc (si el valor contingut és "null").
     *
     */
    public String getProperty (String key, String defaultValue) {

        String res = null;
        key = searchKey (key);
        if (key!=null) { res = StringUtils.notNull (properties.getProperty (key));  }
        else { res = defaultValue; }
        
        return res;
    }

    /**
     * Obté el valor enmagatzemat a la propietat amb la clau "key".
     *
     * Si la clau "key" té el valor "null" o no existeix cap entrada amb la clau
     * indicada, retorna un valor a "Null".
     *
     * Si existeix una entrada amb la clau "key", en retorna el seu valor o una
     * cadena en blanc (si el valor contingut és "null").
     *
     */
    public String getProperty (String key) { return getProperty (key, null); }

    public byte getPropertyAsByte (String key) { return StringToNumber.byteValue (getProperty (key, null)); }
    public short getPropertyAsShort (String key) { return StringToNumber.shortValue (getProperty (key, null)); }
    public int getPropertyAsInteger (String key) { return StringToNumber.intValue (getProperty (key, null)); }
    public float getPropertyAsFloat (String key) { return StringToNumber.floatValue (getProperty (key, null)); }
    public double getPropertyAsDouble (String key) { return StringToNumber.doubleValue (getProperty (key, null)); }
    public Date getPropertyAsDate (String key) { return StringToNumber.dateValue (getProperty (key, null)); }
    public boolean getPropertyAsBoolean (String key) { return StringToNumber.booleanValue (getProperty (key, null)); }



    /**
     * Assigna el valor "value" a la propietat amb la clau "key". Si no existia
     * cap entrada amb aquesta clau, s'afegueix a la llista, i en cas contrari
     * se'n modifica el valor i es perd el valor original.
     *
     * NOTA: Si el paràmetre "key" conté el valor "null" no es realitza CAP
     * acció.
     *
     */
    public void setProperty (String key, String value) { 
        if (key!=null) { properties.setProperty (key, StringUtils.notNull(value)); }
    }

    /**
     * Retorna un valor booleà que indica si la llista de propietats conté alguna
     * clau amb el valor "key".
     *
     * Si està desactivat el atribut "caseSensitive" busca la primera clau
     * amb el valor "key" sigui quina sigui la seva tipografia.
     *
     * @param key
     * @return
     */
    public boolean containsKey (String key) {

        key = searchKey (key);
        return  (key!=null);
     }

    /**
     * Retorna un valor booleà que indica si dintre de la llista de propietats
     * n'hi ha alguna que contingui el valor "value".
     *
     * Amb el paràmetre "caseSensitive" podem indicar si volem fer la comprovació
     * tenint en compte o ignorant la tipografia.
     * 
     * @param value
     * @return
     */
     public boolean containsValue (String value, boolean caseSensitive) {

         boolean res = false;
         if (!caseSensitive) {
             Map keys = searchValueKeys (value, caseSensitive);
             res = (keys.size()!=0);
         }
         else { res = properties.containsValue (value); }

         return res;
     }

     public boolean containsValue (String value) { return containsValue (value, false); }

     /**
      * Obté les claus de totes les entrades que contenen el valor "value".
      * 
      * @param value
      * @param caseSensitive
      * @return
      */
     public Map getPropertyKeys (String value, boolean caseSensitive) { return searchValueKeys (value, caseSensitive); }

     public Map getPropertyKeys (String value) { return getPropertyKeys (value, false); }

     /**
      * Obté la clau de la primera entrada que conté el valor "value".
      *
      * @param value
      * @param caseSensitive
      * @return
      */
     public String getPropertyKey (String value, boolean caseSensitive) {

        String res = null;
        Map keys = searchValueKeys (value, caseSensitive);

        Iterator iterator = keys.entrySet().iterator();
        if (iterator.hasNext()) {
            Map.Entry e = (Entry) iterator.next();
            res = (String) e.getKey();
        }
        return res;
     }

     public String getPropertyKey (String value) { return getPropertyKey (value, false); }
}
