//    _____ _                       _  _  ____        ___ _            _   ______      _  _
//   / ____) |                     | || ||  __ \     / __) |          | | |  ____)    | || |
//  | |    | |  _ _ ___ ___  ___   |_||_|| |__) |___| |_ | | ___  ____| |_| |__  _   ___||_|
//  | |    | |/ _` | __) __)/ _ \        |  _  // _ \  _|| |/ _ \/ ___) __)  __) \ \/ /
//  | |____| | (_| |__ \__ \  __/        | | \ \  __/ |  | |  __/ (___| |_| |____ )  (
//   \_____)_|\__,_|___/___/\___|        |_|  \_\___|_|  |_|\___|\____)\__)______)_/\_\
//
//


package cat.lib.reflect;

import cat.lib.errors.ErrorEx;
import cat.lib.math.IntegerEx;
import cat.lib.utils.SetUtils;
import cat.lib.utils.StringToNumber;
import cat.lib.utils.StringUtils;
import java.lang.reflect.Array;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Stack;
import java.util.Vector;

/**
 * CLASSE "ReflectEx":
 * -------------------
 *
 *   La classe "ReflectEx" és una classe instanciable que permet executar métodes
 * i obtenir el valor d'atributs dinámicament.
 *
 *   Tambe permet executar "expressions" Java que impliquin la execució d'un métode
 * o l'accés a un atribut.
 *
 *
 * NOTA:
 *
 *   Per motius d'eficiència, s'ha hagut de dissenyar de manera instanciable perque
 * la recerca de métodes requereix força temps d'execució.
 *
 *   Per aplicacions normals que realitzen pocs accessos, no seria cap problema,
 * però el framework "Java Object Faces", l'utilitza de manera massiva
 * en algoritmes crítics en rendiment d'execució.
 *
 *   Per això, a mesura que es van cercant els métodes de diferents classes, per
 * a facilitar-ne la posterior consulta s'enmagatzemen en un mapa que no es
 * recomana que sigui "static" per evitar problemes amb el Tomcat.
 *
 *
 * -----------------------------------------------------------------------------
 * @author Miquel Abadal (Interface Safe Access)
 * @date 02-02-2011
 *
 */
public class ReflectEx {

    //    _____                _              _          _          _        _ _           _                   _        _   _
    //   / ____)              | |            | |        (_)        | |      (_) |         | |                 | |      | | (_)
    //  | |     ___  _ __  ___| |_  _ _ _ __ | |_ ___    _      _ _| |_ _ __ _| |_   _   _| |_ ___     ___ ___| |_  _ _| |_ _  ____ ___
    //  | |    / _ \| `_  \ __) __) _` | `_  \ __) __)  | |   / _` | __) `__) | `_ \| | | | __) __)   / _ \ __) __) _` | __) |/ ___) __)
    //  | |____ (_) | | | |__ \ |_ (_| | | | | |_\__ \  | |  | (_| | |_| |  | | |_) | |_| | |_\__ \  |  __/__ \ |_ (_| | |_| | (___\__ \
    //   \_____)___/|_| |_|___/\__)__,_|_| |_|\__)___/  |_|   \__,_|\__)_|  |_|_,__/ \__,_|\__)___/   \___|___/\__)__,_|\__)_|\____)___/
    //
    //

    protected static final int IS_UNKNOW = 0;
    protected static final int IS_ATTRIBUTE = 1;
    protected static final int IS_METHOD = 2;
    protected static final int IS_STRING = 3;
    protected static final int IS_CHARACTER = 4;
    protected static final int IS_INTEGER = 5;
    protected static final int IS_DOUBLE = 6;
    protected static final int IS_BOOLEAN = 7;
    protected static final int IS_ARRAY = 8;

    
    //     __   _        _ _           _
    //   / __ \| |      (_) |         | |
    //  | |__| | |_ _ __ _| |_   _   _| |_ ___
    //  |  __  | __) `__) | `_ \| | | | __) __)
    //  | |  | | |_| |  | | |_) | |_| | |_\__ \
    //  |_|  |_|\__)_|  |_|_,__/ \__,_|\__)___/
    //
    //

    private Map classesList = null;

    //    _____ _                                      _                 _
    //   / ____) |                                    (_)               | |
    //  | |    | |  _ _ ___ ___  ___ ___    _ __  _ __ _ _    _  _ _  __| | ___ ___
    //  | |    | |/ _` | __) __)/ _ \ __)  | `_ \| `__) | |  | | _` |/ _` |/ _ \ __)
    //  | |____| | (_| |__ \__ \  __/__ \  | |_) | |  | |\ \/ / (_| | (_| |  __/__ \
    //   \_____)_|\__,_|___/___/\___|___/  |  __/|_|  |_| \__/ \__,_|\__,_|\___|___/
    //                                     | |
    //                                     |_|

    final class ExpressionVoid {
        protected ExpressionVoid() { }
        @Override
        public String toString() { return "void"; }
    }

    final class ExpressionData {

        protected Object dataSource = null;
        protected boolean onCast = false;
        protected boolean onBody = false;
        protected boolean onParams = false;
        protected StringBuilder body = new StringBuilder();
        protected StringBuilder cast = new StringBuilder();
        protected Vector paramsList = new Vector();
        protected int bodyType = IS_UNKNOW;

        protected ExpressionData (Object dataSource) { initData (dataSource); }

        protected void initData (Object dataSource) {

            this.dataSource = dataSource;
            this.onCast = false;
            this.onBody = false;
            this.onParams = false;
            this.body = new StringBuilder();
            this.cast = new StringBuilder();
            this.paramsList = new Vector();
            this.bodyType = IS_UNKNOW;
        }

        protected boolean isUnknow() { return (bodyType==IS_UNKNOW); }
        protected boolean isMethod() { return (bodyType==IS_METHOD); }
        protected boolean isAttribute() { return (bodyType==IS_ATTRIBUTE); }
        protected boolean isArray() { return (bodyType==IS_ARRAY); }
        protected boolean isString() { return (bodyType==IS_STRING); }
        protected boolean isCharacter() { return (bodyType==IS_CHARACTER); }
        protected boolean isInteger() { return (bodyType==IS_INTEGER); }
        protected boolean isDouble() { return (bodyType==IS_DOUBLE); }
        protected boolean isBoolean() { return (bodyType==IS_BOOLEAN); }
    }
    

    //   __  __      _            _                              _                    _
    //  |  \/  |    | |          | |                            | |                  | |
    //  | \  / | ___| |_ ___   __| | ___     ____ ___  _ __  ___| |_ _ __ _   _  ____| |_ ___  _ __
    //  | |\/| |/ _ \ __) _ \ / _` |/ _ \   / ___) _ \| `_  \ __) __) `__) | | |/ ___) __) _ \| `__)
    //  | |  | |  __/ |_ (_) | (_| |  __/  | (___ (_) | | | |__ \ |_| |  | |_| | (___| |_ (_) | |
    //  |_|  |_|\___|\__)___/ \__,_|\___|   \____)___/|_| |_|___/\__)_|   \__,_|\____)\__)___/|_|
    //
    //

    public ReflectEx () {}


    //   __  __      _            _                             _             _ _
    //  |  \/  |    | |          | |                           | |           (_) |
    //  | \  / | ___| |_ ___   __| | ___ ___    _ __  _ __ ___ | |_ ___  __ _ _| |_ ___
    //  | |\/| |/ _ \ __) _ \ / _` |/ _ \ __)  | `_ \| `__) _ \| __) _ \/ _` | | __) __)
    //  | |  | |  __/ |_ (_) | (_| |  __/__ \  | |_) | | | (_) | |_  __/ (_| | | |_\__ \
    //  |_|  |_|\___|\__)___/ \__,_|\___|___/  |  __/|_|  \___/ \__)___|\__, |_|\__)___/
    //                                         | |                       __| |
    //                                         |_|                      (___/


    //   __  __      _            _                        _             _
    //  |  \/  |    | |          | |                      (_)           | |
    //  | \  / | ___| |_ ___   __| | ___ ___    _ __  _ __ _ _    _  _ _| |_ ___
    //  | |\/| |/ _ \ __) _ \ / _` |/ _ \ __)  | `_ \| `__) | |  | | _` | __) __)
    //  | |  | |  __/ |_ (_) | (_| |  __/__ \  | |_) | |  | |\ \/ / (_| | |_\__ \
    //  |_|  |_|\___|\__)___/ \__,_|\___|___/  |  __/|_|  |_| \__/ \__,_|\__)___/
    //                                         | |
    //                                         |_|

    /**
     * Obtiene de una tabla de objetos, otra tabla formada por las clases de los
     * objetos de la primera.
     */
    private Class[] getObjectClass (Object[] list) {

        Class[] res = null;
        if (list!=null) {
            res = new Class[list.length];
            for (int cont1=0; cont1<list.length; cont1++) {
                if (list[cont1]!=null) { res[cont1] = list[cont1].getClass(); }
                else { res[cont1] = Object.class; }
            }
        }
        else { res = new Class[0]; }

        return res;

    }

    /**
     * Afegueix al mapa "methodList" la llista de métodes de la classe "clase"
     * indexades amb una clau que conté la "signatura" del métode.
     *
     * NOTA: La signatura està formada pel nom del métode i els tipus dels
     * paràmetres.
     */
    private void addMethodsAndFields (Map list, Class clase) {

        // Obting la llista de metodes de la classe i les seves superclasses
        // i els enmagatzemo a la llista "methodList".
        Class superClase = clase;
        while (superClase!=null) {

            // Per cada métode afegueixo una entrada com la següent:
            // --> nomClase.nomMethode(claseParam1, claseParam2, ...)
            Method[] methods = superClase.getDeclaredMethods();
            for (int cont1=0; cont1<methods.length; cont1++) {
                String signature = getSignature (methods[cont1].getName(), methods[cont1].getParameterTypes());
                list.put (signature, methods[cont1]);
            }

            Field[] fields = superClase.getDeclaredFields();
            for (int cont1=0; cont1<fields.length; cont1++) {
                String signature = getSignature (fields[cont1].getName(), null);
                list.put (signature, fields[cont1]);
            }
            superClase = superClase.getSuperclass();
        }
    }

    private Map addClass (Class clase) {

        if (classesList==null) { classesList = new HashMap(); }

        Map list = (Map) classesList.get (clase.getName());
        if (list==null) {
            list = new HashMap();
            classesList.put (clase.getName(), list);
            addMethodsAndFields (list, clase);
        }

        return list;

    }

    /**
     * Executa una expressió enmagatzemada amb un a instància de classe ExpressionData.
     *
     * En aquesta classe, hi ha els següents camps:
     *
     * - body --> Nom del métode, atribut o constant.
     * - bodyType --> Tipus de cos (0 - UNKNOW, 1 - IS_METHOD, 2 - IS_ATTIBUTE,
     *                              3 - IS_LITERAL (String), 4 - IS_CHARACTER,
     *                              5 - IS_INTEGER, 6 - IS_DOUBLE)
     * - cast --> Tipus al que s'ha de convertir el resultat ("byte", "short",
     *                              "int", "integer", "long", "float", "double",
     *                              "boolean", "char", "character" i "string")
     * - data.paramsList --> Vector amb la llista de paràmetres
     * - data.dataSource --> Objecte del qual s'ha d'executar el métode o obtenir
     * el atribut.
     *
     *
     * Si la font de dades "data.dataSource" es "null", lalvors s'ha d'obtenir
     * una font de dades del "DataSourceProvider" que disposa de dos métodes
     * una en la que se li passa el nom de la font de dades, i l'altre que simplemement
     * retorna una font de dades per defecte.
     *
     */
    private Object execute (DataSourceProvider dataSourceProvider, ExpressionData data, ErrorEx error) {

        Object res = null;

        if (data.body.length()>0 || data.bodyType==IS_STRING) {
            String body = data.body.toString();
            if (body.equalsIgnoreCase("true") || body.equalsIgnoreCase("false")) { data.bodyType = IS_BOOLEAN; }

            Object[] valueParams = SetUtils.VectorToArray (data.paramsList);

            if (data.isMethod()) {
                if (data.dataSource==null && dataSourceProvider!=null) {
                    data.dataSource = dataSourceProvider.getDataSource (body);
                    if (data.dataSource==null) {
                        data.dataSource = dataSourceProvider.getDataSource();
                        res = executeMethod (data.dataSource, body, valueParams, error, true);
                    }
                    else { res = data.dataSource; }
                }
                else { res = executeMethod (data.dataSource, body, valueParams, error, true); }
            }
            else if (data.isAttribute()) {
                if (data.dataSource==null && dataSourceProvider!=null) {
                    data.dataSource = dataSourceProvider.getDataSource (body);
                    if (data.dataSource==null) {
                        data.dataSource = dataSourceProvider.getDataSource();
                        res = getAttribute (data.dataSource, body, error, true);
                    }
                    else { res = data.dataSource; }
                }
                else { res = getAttribute (data.dataSource, body, error, true); }
            }
            else if (data.isArray()) {
                if (data.dataSource==null && dataSourceProvider!=null) {
                    data.dataSource = dataSourceProvider.getDataSource (body);
                    if (data.dataSource==null) {
                        data.dataSource = dataSourceProvider.getDataSource();
                        res = getTable (data.dataSource, body, valueParams, error, true);
                    }
                    else { res = data.dataSource; }
                }
                else { res = getTable (data.dataSource, body, valueParams, error, true); }
            }
            else if (data.isString()) { res = body; }
            else if (data.isCharacter()) { res = body.charAt(0); }
            else if (data.isInteger()) { res = StringToNumber.intValue (body, 0, error); }
            else if (data.isDouble()) { res = StringToNumber.doubleValue (body, 0, error); }
            else if (data.isBoolean()) { res = StringToNumber.booleanValue (body, false, error); }

            if (data.cast.length()>0 && res!=null) {

                String valueStr = null;
                if (res instanceof String) { valueStr = (String) res; }
                else { valueStr = res.toString(); }

                String cast = data.cast.toString().toLowerCase();
                if (cast.equals("byte")) { res = StringToNumber.byteValue (valueStr, (byte) 0); }
                else if (cast.equals("short")) { res = StringToNumber.shortValue(valueStr, (short) 0); }
                else if (cast.equals("int")) { res = StringToNumber.intValue (valueStr, 0); }
                else if (cast.equals("integer")) { res = StringToNumber.intValue (valueStr, 0); }
                else if (cast.equals("long")) { res = StringToNumber.longValue (valueStr, 0); }
                else if (cast.equals("float")) { res = StringToNumber.floatValue (valueStr, 0); }
                else if (cast.equals("double")) { res = StringToNumber.doubleValue (valueStr, 0); }
                else if (cast.equals("char") || cast.equals("character")) {
                    res = '\0';
                    if (valueStr.length()>0) { res = valueStr.charAt(0); }
                }
                else if (cast.equals("boolean")) { res = StringToNumber.booleanValue (valueStr, false); }

                else if (cast.equals("string")) {  }
                else { error.setError ("CAST_ERROR", "DESCRIPTION: Tipo de variable no reconocido.", "VALUE: "+res.toString(), "CAST_TO: "+data.cast); }
                
            }
        }

        return res;
    }

    private ExpressionData closedExpression (Stack stack, ExpressionData data, DataSourceProvider dataSourceProvider, String expression, char caract, ErrorEx error) {

        Object expressionResult = execute (dataSourceProvider, data, error);

        if (!stack.empty()) { stack.pop(); }
        else { error.setError ("EXPRESSION_ERROR", "DESCRIPTION: Sentencia incompleta", "EXPRESSION: "+expression); }

        if (!stack.empty()) {
            data = (ExpressionData) stack.peek();
            data.paramsList.add (expressionResult);
            if (!data.onParams) { error.setError ("EXPRESSION_ERROR", "DESCRIPTION: No se esperaba el caracter '"+caract+"'", "EXPRESSION: "+expression); }
        }
        else {
            data = new ExpressionData (expressionResult);
            stack.push (data);
        }

        return data;

    }

    /**
     * Algoritme que interpreta una expressió Java i l'executa utilitzant una o
     * varies fonts de dades.
     *
     * Si s'indica una font de dades primigènia "object" aquesta s'utilitzarà en
     * totes les expressions, i en cas contrari, cada expressió cercarà del
     * "DataSourceProvider" la font de dades que ha d'utiliztar.
     *
     * Veure la descripció "executeExpression" per a consultar la sintaxis
     * que reconeix aquest intérpret.
     *
     */
    private Object interpretExpression (Object object, DataSourceProvider dataSourceProvider, String expression, IntegerEx index, ErrorEx error) {

        ExpressionData data = new ExpressionData (object);

        if (expression!=null) {
            Stack stack = new Stack();
            stack.push (data);

            boolean onEscapeChar = false;
            boolean onLiteral = false;
            boolean addCaract = false;
            boolean endOfExpression = false;

            char caract = '\0';
            char nextCaract = '\0';
            char literalChar = '\0';
            int openArray = 0;

            while (index.getValue()<expression.length() && error.getError()==null && !endOfExpression) {

                caract = expression.charAt(index.getValue());
                if (index.getValue()+1<expression.length()) { nextCaract = expression.charAt(index.getValue()+1); }
                else { nextCaract = '\0'; }

                addCaract = false;

                if (!onLiteral) {
                    if (caract==' ') { }
                    else if (caract=='"') {
                        onLiteral = true; literalChar = caract; data.bodyType = IS_STRING;
                    }
                    else if (caract=='\'') { onLiteral = true; literalChar = caract; data.bodyType = IS_CHARACTER; }
                    else if(caract == '(') {
                        if (!data.onCast && !data.onBody && data.dataSource==null) { data.onCast = true; }
                        else if (data.onBody) {
                            data.onParams = true;
                            data.bodyType = IS_METHOD;
                            data = new ExpressionData (object);
                            stack.push (data);
                        }
                        else { error.setError ("EXPRESSION_ERROR", "DESCRIPTION: No se esperaba el caracter '"+caract+"'", "EXPRESSION: "+expression); }
                    }
                    else if (caract==')') {
                        if (data.onCast) {
                            if (data.cast.length()>0) { data.onCast = false; }
                            else { error.setError ("EXPRESSION_ERROR", "No se ha indicado el tipo al que debe encapsularse el resultado')'", "EXPRESSION: "+expression); }
                        }
                        else  {
                            if (data.body.length()==0) { error.setError ("EXPRESSION_ERROR", "DESCRIPTION: No se esperaba el caracter '"+caract+"'", "EXPRESSION: "+expression); }

                            data = closedExpression (stack, data, dataSourceProvider, expression, caract, error);
                        }
                    }
                    else if (caract=='[') {
                        openArray++;
                        if (data.onBody) {
                            data.onParams = true;
                            data.bodyType = IS_ARRAY;
                            data = new ExpressionData (object);
                            stack.push (data);
                        }
                        else { error.setError ("EXPRESSION_ERROR", "DESCRIPTION: No se esperaba el caracter '"+caract+"'", "EXPRESSION: "+expression); }
                    }
                    else if (caract==']') {
                        if (data.body.length()==0) { error.setError ("EXPRESSION_ERROR", "DESCRIPTION: No se esperaba el caracter '"+caract+"'", "EXPRESSION: "+expression); }

                        data = closedExpression (stack, data, dataSourceProvider, expression, caract, error);
                        openArray--;
                    }
                    else if (caract==',') {
                        if (data.onParams) {
                            data = new ExpressionData (object);
                            stack.push (data);
                        }
                        else { error.setError("EXPRESSION_ERROR", "DESCRIPTION: No se esperaba el caracter '" + caract + "'", "EXPRESSION: " + expression); }
                    }
                    else if (caract=='.') {
                        if (data.onBody && data.bodyType==IS_INTEGER) { data.bodyType = IS_DOUBLE; addCaract = true; }
                        else {
                            if (data.body.length()>0) {
                                Object expressionResult = expressionResult = execute (dataSourceProvider, data, error);
                                data.initData (expressionResult);
                            }
                            if (data.dataSource==null) { error.setError ("EXECUTE_ERROR", "DESCRIPTION: Null pointer exception", "EXPRESSION: "+expression); }
                        }
                    }
                    else {
                        if (!data.onCast) {
                            if (!data.onBody) {
                                data.onBody = true;
                                if ((caract>='a' && caract<'z') || (caract>='A' && caract<='Z') || caract=='$' || caract=='_') { data.bodyType = IS_ATTRIBUTE; }
                                else if((caract>='0' && caract<='9') || caract=='+' || caract=='-') { data.bodyType = IS_INTEGER; }
                                else { error.setError ("EXPRESSION_ERROR", "DESCRIPTION: El caracter '"+caract+"' no es válido en estas posición", "EXPRESSION: "+expression); }
                            }
                            data.body.append(caract);
                        }
                        else { data.cast.append(caract); }
                    }

                    if (nextCaract==']' && openArray==0) {endOfExpression = true; }
                }
                else {
                    if (caract=='\\' && !onEscapeChar) { onEscapeChar = true; }
                    else if(caract != literalChar || onEscapeChar) {
                        addCaract = true;
                        onEscapeChar = false;
                    }
                    else { onLiteral = false; }
                }

                if (addCaract) {
                    if (data.onBody || onLiteral) { data.body.append (caract); }
                    else if (data.onCast)  { data.cast.append (StringUtils.toLowerCase (caract)); }
                    else { error.setError ("EXPRESSION_ERROR", "DESCRIPTION: El caracter '"+caract+"' no es válido en estas posición", "EXPRESSION: "+expression); }
                }

                if (!data.onCast && (nextCaract==')' || nextCaract==',')) {
                    if (data.isUnknow()) {
                        if (!stack.empty()) { stack.pop(); }
                        else { error.setError ("EXPRESSION_ERROR", "DESCRIPTION: Sentencia incompleta", "EXPRESSION: "+expression); }

                        if (!stack.empty()) { data = (ExpressionData) stack.peek(); }
                        else { error.setError("EXPRESSION_ERROR", "DESCRIPTION: No se esperaba el caracter '" + nextCaract + "'", "EXPRESSION: " + expression); }
                    }
                    if (!data.isMethod()) { data = closedExpression (stack, data, dataSourceProvider, expression, nextCaract, error); }
                }
                index.inc();
            }

            if (data.onBody) { data.dataSource = execute (dataSourceProvider, data, error); }
            stack.pop();

            //if (index.deep<0 && index.value<expression.length()) { error.setError ("EXPRESSION_ERROR", "DESCRIPTION: No se esperaba el caracter ')'", "EXPRESSION: "+expression); }
            //if (onParams) { error.setError ("EXPRESSION_ERROR", "DESCRIPTION: Se esperaba el caracter ')'", "EXPRESSION: "+expression); }
            if (onLiteral) { error.setError ("EXPRESSION_ERROR", "DESCRIPTION: Literal no terminado", "EXPRESSION: "+expression); }
            if (!stack.isEmpty()) { error.setError ("EXPRESSION_ERROR", "DESCRIPTION: Sentencia incompleta", "EXPRESSION: "+expression); }
        }

        return data.dataSource;
    }

    //   __  __      _            _             _  _   _____ ______ _______ _  _
    //  |  \/  |    | |          | |           | || | / ____)  ____)__   __) || |
    //  | \  / | ___| |_ ___   __| | ___ ___   |_||_|| |  __| |__     | |  |_||_|
    //  | |\/| |/ _ \ __) _ \ / _` |/ _ \ __)        | | (_ |  __)    | |
    //  | |  | |  __/ |_ (_) | (_| |  __/__ \        | |__| | |____   | |
    //  |_|  |_|\___|\__)___/ \__,_|\___|___/         \_____|______)  |_|
    //
    //

    /**
     * Obté el nom de la classe wrapper de la clase "clase"
     */
    public String getWrapperName (Class clase) {

        if (clase==boolean.class) { clase = Boolean.class; }
        else if (clase == byte.class) { clase = Byte.class; }
        else if (clase == short.class) { clase = Short.class; }
        else if (clase == int.class) { clase = Integer.class; }
        else if (clase == long.class) { clase = Long.class; }
        else if (clase == float.class) { clase = Float.class; }
        else if (clase == double.class) { clase = Double.class; }
        else if (clase == char.class) { clase = Character.class; }

        return clase.getName();
    }

    /**
     * Obté la signatura d'un métode o atribut.
     * Si params conté "null" s'interpreta com a atribut.
     *
     * NOTA: La signatura està formada pel nom del métode i els tipus dels
     * paràmetres.
     */
    public String getSignature (String name, Class[] params) {

        String res = name;
        String paramsStr = "";

        if (params!=null) {
            for (int cont1=0; cont1<params.length; cont1++) {
                Class param = params[cont1];
                String paramName = param.getName();
                if (param.isPrimitive()) { paramName = getWrapperName (param); }
                paramsStr = StringUtils.concat(paramsStr, ",", paramName);
            }
            res = res + "("+ paramsStr +")";
        }

        return res;
    }

    /**
     * Obté la signatura d'un métode.
     *
     * NOTA: La signatura està formada pel nom del métode i els tipus dels
     * paràmetres.
     */
    public String getSignature (String name, Object[] params) {

        String res = name;
        String paramsStr = "";

        if (params!=null) {
            for (int cont1=0; cont1<params.length; cont1++) {

                Class clase = null;
                if (params[cont1]!=null) { clase = params[cont1].getClass(); }
                else { clase = Object.class; }

                paramsStr = StringUtils.concat(paramsStr, ",", clase.getName());
            }
            res = res + "("+ paramsStr +")";
        }
        return res;
    }



    /**
     * Obté el métode amb el nom "methodName" de la classe "clase" i amb els paràmetres indicats.
     *
     * NOTA: Amb el paràmetre "caseSensitive" es pot indicar si volem ser estrictes al
     * localitzar el métode o no.
     *
     * NOTA: Si es permet cercar métodes sense tenir en compte la tipografia, si
     * la classe conté dos o més métodes amb el mateix nom (i diferent tipografia) i els
     * materixos paràmetres, no podem saber quin dels métodes ha trobat.
     *
     */
    public Field getField (Class clase, String fieldName, boolean caseSensitive) {

        Map list = addClass (clase);
        return (Field) SetUtils.getMapValue (getSignature (fieldName, null), list, caseSensitive);
    }

    /**
     * Obté el métode amb el nom "methodName" de la classe "clase" i amb els paràmetres indicats.
     *
     * NOTA: Amb el paràmetre "caseSensitive" es pot indicar si volem ser estrictes al
     * localitzar el métode o no.
     *
     * NOTA: Si es permet cercar métodes sense tenir en compte la tipografia, si
     * la classe conté dos o més métodes amb el mateix nom (i diferent tipografia) i els
     * materixos paràmetres, no podem saber quin dels métodes ha trobat.
     *
     */
    public Method getMethod (Class clase, String methodName, Class[] params, boolean caseSensitive) {

        Map list = addClass (clase);
        return (Method) SetUtils.getMapValue (getSignature (methodName, params), list, caseSensitive);
    }

    /**
     * Obté la llista de métodes de la classe "clase" i la retorna en un HashMap on
     * cada element és un objecte de la classe "Method"
     */
    public Map getMethods (Class clase) {

        Map methodsAndFields = addClass (clase);
        Map res = new HashMap();

        Iterator i = methodsAndFields.entrySet().iterator();
        while (i.hasNext()) {
            Map.Entry e = (Entry) i.next();
            String key = (String) e.getKey();
            Object value = e.getValue();
            if (value instanceof Method) { res.put (key, (Method) value); }
        }
        return res;
    }

    /**
     * Obté la llista d'atributs de la classe "clase" i la retorna en un HashMap on
     * cada element és un objecte de la classe "Field"
     */
    public Map getFields (Class clase) {

        Map methodsAndFields = addClass (clase);
        Map res = new HashMap();

        Iterator i = methodsAndFields.entrySet().iterator();
        while (i.hasNext()) {
            Map.Entry e = (Entry) i.next();
            String key = (String) e.getKey();
            Object value = e.getValue();
            if (value instanceof Field) { res.put (key, (Field) value); }
        }
        return res;
    }

    /**
     * Obté el métode amb el nom "methodName" de la classe "clase" i amb els paràmetres indicats.
     *
     * NOTA: El métode cal indicarlo amb la tipografia exacte o no el trobarà.
     *
     */
    public Method getMethod (Class clase, String methodName, Class[] params) { return getMethod (clase, methodName, params, StringUtils.CASE_SENSITIVE); }

    /**
     * Obté el métode constructor de la classe "clase" amb els paràmetres indicats.
     */
    public Constructor getConstructor (Class clase, Class[] params) {

        Constructor res = null;
        Constructor[] constructors = clase.getConstructors();

        int cont1 = 0;
        while (res==null && cont1<constructors.length) {

            Class[] constructorParams = constructors[cont1].getParameterTypes();

            if (constructorParams.length == params.length) {

                int cont2 = 0;
                boolean trobat = true;
                while (trobat==true && cont2<constructorParams.length) {
                    if (constructorParams[cont2]!=params[cont2]) { trobat = false; }
                    cont2++;
                }
                if (trobat) { res = constructors[cont1]; }
            }

            cont1++;
        }
        return res;
    }

    /**
     * Obté el contingut d'un atribut implementat amb técnica "CamelCase", o sigui
     * el atribut definit com a privat i un métode públic amb el prefixe "get".
     *
     * Retorna un objecte amb el valor del atribut.
     *
     * NOTA: Només pot accedir al métode si aquest està definit com a "public".
     *
     */
    public Object getCamelCaseAttribute (Object object, String attribute, Object[] params, ErrorEx error)  {

        // Obté el nom del métode a invocar
        String methodName = attribute;
        if (!StringUtils.copy (attribute, 0, 3).equalsIgnoreCase("get")) {
            methodName= "get"+StringUtils.firstCharToUpperCase (attribute, StringUtils.SUBST_ONE);
        }

        return executeMethod (object, methodName, params, error, false);
    }


    //   __  __      _            _             _  _   _____ ______ _______ _  _
    //  |  \/  |    | |          | |           | || | / ____)  ____)__   __) || |
    //  | \  / | ___| |_ ___   __| | ___ ___   |_||_|| (___ | |__     | |  |_||_|
    //  | |\/| |/ _ \ __) _ \ / _` |/ _ \ __)         \___ \|  __)    | |
    //  | |  | |  __/ |_ (_) | (_| |  __/__ \         ____) | |____   | |
    //  |_|  |_|\___|\__)___/ \__,_|\___|___/        (_____/|______)  |_|
    //
    //

    /**
     * Modifica el contingut d'un atribut implementat amb técnica "CamelCase", o sigui
     * el atribut definit com a privat i un métode públic amb el prefixe "set".     *
     *
     * NOTA: Només pot accedir al métode si aquest està definit com a "public".
     */
    public void setCamelCaseAttribute (Object object, String attribute, Object[] params, ErrorEx error)  {

        // Obté el nom del métode a invocar
        String methodName = attribute;
        if (!StringUtils.copy (attribute, 0, 3).equalsIgnoreCase("set")) {
            methodName= "set"+StringUtils.firstCharToUpperCase (attribute, StringUtils.SUBST_ONE);
        }

        executeMethod (object, methodName, params, error, false);
    }


    //   __  __      _            _                     _     _ _
    //  |  \/  |    | |          | |                   | |   | |_)
    //  | \  / | ___| |_ ___   __| | ___    _ __  _   _| |_  | |_  ____ ___
    //  | |\/| |/ _ \ __) _ \ / _` |/ _ \  | `_ \| | | | `_ \| | |/ ___) __)
    //  | |  | |  __/ |_ (_) | (_| |  __/  | |_) | |_| | |_) | | | (___\__ \
    //  |_|  |_|\___|\__)___/ \__,_|\___|  |  __/ \__,_|_,__/|_|_|\____)___/
    //                                     | |
    //                                     |_|

    /**
     *      Métode "executeExpression" és un procediment capaç d'executar sobre el
     * objecte determinat la expressió Java indicada.
     *
     *      Les expressions que pot reconèixer han de respectar la següent
     * sintaxis:
     *
     * |---------------------------------------------------------------------------|
     * | (cast) [dataSource.]methodName (expression, expression....).expressio...  |
     * | (cast) [dataSource.]attributeName.expressio                               |
     * | (cast) constant                                                           |
     * |---------------------------------------------------------------------------|
     *
     *      Com es pot veure, tan es pot invocar un métode com obtenir el valor
     * d'un atribut o definir una constant.
     *
     *
     * Casting de variables:
     * ---------------------
     *
     *      En tots tres casos es pot definir el tipus de variable al qual es vol
     * convertir el resultat obtingut. Els possibles tipus disponibles són els
     * següents:
     *
     * |------------------|--------------------------|
     * | byte             | -128 .. +127             |
     * | short            | -32768.. +32767          |
     * | int o integer    | -2147483648 a 2147483647 |
     * | long             | -2e63 a (2e63)-1         |
     * | float            | 10e-46 a 10e38           |
     * | double           | 10e-324 a 10e308         |
     * | boolean          | false ... true           |
     * | char o character | 0 .. 65535               |
     * | string           | cadenes de text          |
     * |------------------|--------------------------|
     *
     *      Naturalment, si el tipus de variable obtingut no es pot convertir al
     * tipus indicat, aquest queda saturat a "0" (si és numeric), "false" si és
     * boolean, '\0' si és un caràcter o a una cadena buida si és una cadena de
     * text.
     *
     *
     * Les constants:
     * --------------
     *
     *      Les constants, només poden ser numétiques, i si són alfanumériques
     * han d'estar delimitades per les cometes simples si són de tipus caràcter
     * o per les cometes dobles si son cadenes de text.
     *
     *      Si són numériques, per defecte el intérpret les encapsula amb una
     * variable de tipus "integer", excepte en el cas de que continguin caràcters
     * decimals, ja que en aquest cas les encapsula amb una variable de tipus
     * "double".
     *
     *      Si es vol definir una variable amb un tipus concret, cal indicar-ho
     * mitjançant el casting.
     *
     * |-----------------------------------------------------|
     * |  "Hello word";     --> Constant de tipus "String"   |
     * |  'a'               --> Constant de tipus "char"     |
     * |  12                --> Constant de tipus "int"      |
     * |  12.0              --> Constant de tipus "double"   |
     * |  true              --> Constant de tipus "boolean"  |
     * |-----------------------------------------------------|
     *
     *
     * Els atributs i els métodes:
     * ---------------------------
     *
     *      Els atributs i els métodes han de començar amb un caràcter alfabétic
     * entre la 'a' o la 'z' tot i que també s'accepta que comencin amb els
     * caràcter '$' i '_'.
     *
     *      En ambdos casos, per a poder-se executar, cal que es defineixi un
     * objecte o font de dades sobre el qual s'executarà el métode o el qual ha
     * de contenir el atribut indicat.
     *
     *      Existeixen dos sobre-escriptures d'aquest métode d'execució d'expressions:
     *
     * |-----------------------------------------------------------------------------------------------------------|
     * | public Object executeExpression (Object dataSource, String expression, ErrorEx error)                     |
     * | public Object executeExpression (DataSourceProvider dataSourceProvider, String expression, ErrorEx error) |
     * |-----------------------------------------------------------------------------------------------------------|
     *
     *      La primera utilitza el objecte "dataSource" com a font de dades, i TOTS
     * els métodes o atributs, obtindran la informació d'aquest objecte.
     *
     *      La segona, permet utilitzar un "proveïdor" de fonts de dades al qual
     * li indiquem el "nom" de la font de dades i ell ens en retorna l'objecte.
     *
     *      Per a que aquesta segona funcionalitat pugui ser efectiva el proveïdor
     * de fonts de dades, ha d'implementar la classe "dataSourceProvider" que
     * obliga a implementar els métodes següents:
     *
     * |-------------------------------------------------------|
     * |  public Object getDataSource(String dataSourceName);  |
     * |  public Object getDefaultDataSource();                |
     * |-------------------------------------------------------|
     *
     *      Aquest métodes s'invoquen automàticament quan el intérpret vol executar
     * un métode o obtenir un atribut i encara no té l'objecte sobre el qual realitzar
     * la operació.
     *
     *      Si en algun moment, s'intenta executar un métode o atribut i no es pot
     * determinar la font de dades, llavors es produeix un error "null pointer".
     *
     *      S'ha implementat aquesta funcionalitat per a permetre que en una mateixa
     * senténcia es puguin incloure diverses expressions Java, cadascuna de les
     * quals utilitzi una font de dades pròpia.
     *
     *
     * Concatenació d'expressions:
     * ---------------------------
     *
     *      Les expressions es poden concatenar utilitzant el separador '.' de manera
     * que el objecte obtingut com a resultat d'una expressió, passa a ser la font
     * de dades de la següent expresió.
     *
     *      Si una expressió retorna un valor null, i el resultat està concatenat
     * amb una altre expressió, llavors es produeix un error "null pointer".
     *
     * 
     * Paràmetres:
     * -----------
     * 
     *      Els métodes, poden necessitar paràmetres, els quals s'implementen
     * com qualsevol altre expression i per tant, poden definir el seu propi
     * casting, incloure métodes amb els seus corresponents paràmetres, atributs
     * o constants, així com a contatenació d'expressions.
     * 
     * 
     * Ejemples amb una font de dades UNICA:
     * -------------------------------------
     *
     *      Per ejemple si tenim un objecte "product" (que NO implementa la
     * interfície "DataSourceProvider") amb els següents métodes:
     *
     * |-----------------------------------------------------------------------|
     * |  public String getName() { return name; }                             |
     * |  public double getPrice() { return price; }                           |
     * |  public int calcIva (double price, double iva) { return a*b/100; }    |
     * |-----------------------------------------------------------------------|
     *
     *      I invoquem el métode passant-li com a font de dades el objecte
     * "product":
     *
     * |----------------------------------------------------------------------------------------|
     * | public Object executeExpression (Object dataSource, String expression, ErrorEx error)  |
     * |----------------------------------------------------------------------------------------|
     *
     *      Llavors, podem executar les següents expressions:
     *     *
     * |-----------------------------------------------------------------------|
     * |  getName() --> Retorna un String                                      |
     * |  getName().charAt(2) --> Retorna un caràcter.                         |
     * |                      --> Si getName() retorna un null, es retorna un  |
     * |                          error "null pointer".                        |
     * |                      --> Si getName() retorna un string de menys de 2 |
     * |                          caracter retorna un error.                   |
     * |  calcIva (getPrice(), 16.0)  --> Retorna un double                    |
     * |-----------------------------------------------------------------------|
     *
     *      Cal observar que TOTS els métodes (i atributs si fos el cas) s'obtenen
     * de la font de dades "product"; en realitat el algoritme executa les
     * següents sentències:
     *
     * |-----------------------------------------------|
     * |  product.getName();                           |
     * |  product.getName().charAt(2);                 |
     * |  product.calcIva (product.getPrice(), 18.0)   |
     * |-----------------------------------------------|
     *
     *
     * Exemples amb fonts de dades MÚLTIPLES:
     * --------------------------------------
     *
     *      Si volem utilitzar una font de dades diferent per a cadascuna de
     * les expressions implicades, llavors hem d'utilitzar un "proveïdor" de
     * fons de dades.
     *
     *      I això es realitza invocant la segona sobre-escriptura del métode
     * passant-li com a font de dades, un objecte que implementi la interfície
     * "DataSourceProvider".
     *
     *      Aquest proveïdor de dades, ha d'implementar forçosament els métodes
     * següents:
     *
     * |-------------------------------------------------------|
     * |  public Object getDataSource(String dataSourceName);  |
     * |  public Object getDefaultDataSource();                |
     * |-------------------------------------------------------|
     *
     *      Els quals han de retornar el objecte que fara la funció de font de dades,
     * tal i com mostra el codi següent:
     *
     * |-------------------------------------------------------------------------|
     * |  class OwnerClass implements OwnerInterface {                           |
     * |                                                                         |
     * |    public Object getDataSource (String dataSourceName) {                |
     * |      if (dataSource.equalsIgnoreCase ("dataBean")) { return dataBean; } |
     * |      if (dataSource.equalsIgnoreCase ("product")) { return product; }   |
     * |      return null;                                                       |
     * |    }                                                                    |
     * |                                                                         |
     * |    public Object getDataSource () { return defaultDataSource; }         |
     * |  }                                                                      |
     * |-------------------------------------------------------------------------|
     *
     * NOTA: Per a fer funcional aquest ej. anterior, cal definir les instàncies
     * "dataBean", "product" i "defaultDataSource" així com altres possibles fonts de
     * dades i si cal, els seus corresponents métodes "get/set".
     *
     *      Un cop creada aquesta classe, al invocar el métode d'execució d'expressions
     * per cadascuna d'elles en cercarà en primera instància la font de dades.
     *
     * |-----------------------|
     * |  product.getName()    |
     * |  dataBean.getName()   |
     * |  getName()            |
     * |-----------------------|
     *
     *      Així, al intentar executar les expressions anteriors, en el primer cas
     * solicitarà al proveidor de fonts de dades el objecte vinculat amb el nom
     * "dataBean" que en aquest exemple retorna la instància "dataBean".
     *
     *      En el segon cas, solicitarà la instància vinculada amb el nom "product".
     *
     *      Finalment, en el últim cas, en primera instància es demana al proveïdor
     * de fonts de dades el objecte vinculat amb el nom "getName" i com que
     * no el troba i retorna "null", a continuació se li demana la font de dades
     * per defecte (invocant el métode getDataSource" sense paràmetres), el qual
     * retorna la instància "defaultDataSource". Així que en realitat, la última
     * expressió es resolt executant la sentència:
     *
     * |--------------------------------|
     * |  defaultDataSource.getName()   |
     * |--------------------------------|
     *
     *      Aquest procés, que a primera vista pot semblar una mica complicat,
     * permet executar sentències amb fonts de dades molt diferent com per exemple:
     *
     * |------------------------------------------------|
     * |  dataBean.calcIva (product.getPrice(), 18.0)   |
     * |------------------------------------------------|
     *
     * NOTA: El parametre "index" és opcional i serveix per a indicarli al interpret
     * el caràcter on comença la expressió. Aquest "index" té la propietat de que
     * en completarse el procés retorna la posició on s'ha acabat la expressió.
     */
    public Object executeExpression (Object dataSource, String expression, IntegerEx index, ErrorEx error) {
        return interpretExpression (dataSource, null, expression, index, error);
    }

    public Object executeExpression (Object dataSource, String expression, ErrorEx error) {
        return executeExpression (dataSource, expression, new IntegerEx (0), error);
    }

    /**
     * Veure explicació del métode:
     *
     * executeExpression (Object dataSource, String expression, ErrorEx error)
     */
    public Object executeExpression (DataSourceProvider dataSourceProvider, String expression, IntegerEx index, ErrorEx error) {
        return interpretExpression (null, dataSourceProvider, expression, index, error);
    }

    public Object executeExpression (DataSourceProvider dataSourceProvider, String expression, ErrorEx error) {
        return executeExpression (dataSourceProvider, expression, new IntegerEx (0), error);
    }

    /**
     * Obté el valor del atribut "attributeName".
     *
     * Si s'activa el paràmetre "accessToPrivateFields" podem accedir a tots els atributs,
     * inclus els definits com a "protected", "package" i "private".
     *
     * Sinó, només podem accedir als atributs "publics".
     *
     */
    public Object getAttribute (Object object, String attributeName, ErrorEx error, boolean accessToPrivateFields) {

        Object res = null;
        try {
            Field field = getField (object.getClass(), attributeName, false);
            //Field field = object.getClass().getField (attributeName);
             if (field!=null) { 
                 if (accessToPrivateFields) { field.setAccessible(true); }
                 res = field.get (object);
             }
             else { error.setError ("INVOKE_ERROR", "DESCRIPTION: No se ha encontrado el atributo", "ATTRIBUTE_NAME: "+attributeName); }
        }
        //catch (NoSuchFieldException e) { error.setError ("INVOKE_ERROR", "DESCRIPTION: Se ha producido un error al intentar acceder al atributo", "ATTIBUTE_NAME: "+attributeName, "EXCEPTION: "+e.getMessage()); }
        catch (IllegalArgumentException e) { error.setError ("INVOKE_ERROR", "DESCRIPTION: Se ha producido un error al intentar acceder al atributo", "ATTIBUTE_NAME: "+attributeName, "EXCEPTION: "+e.getMessage()); }
        catch (IllegalAccessException e) { error.setError ("INVOKE_ERROR", "DESCRIPTION: Se ha producido un error al intentar acceder al atributo", "ATTIBUTE_NAME: "+attributeName, "EXCEPTION: "+e.getMessage()); }
        catch (SecurityException e) { error.setError ("INVOKE_ERROR", "DESCRIPTION: Se ha producido un error al intentar acceder al atributo", "ATTIBUTE_NAME: "+attributeName, "EXCEPTION: "+e.getMessage()); }

        return res;
    }

    /**
     * Obté el valor del atribut "attributeName".
     *
     * NOTA: Només pot accedir als atributs definits com a "publics".
     *
     */
    public Object getAttribute (Object object, String attributeName, ErrorEx error) { return getAttribute (object, attributeName, error, false); }

    public Object getTable (Object object, String tableName, Object[] values, ErrorEx error, boolean accessToPrivateFields)   {

        Object res = null;

        if (object!=null) {

            int dims = values.length;

            // La taula "values" conté el índex del element a obtenir. Cal passarlo a "int".
            int[] indexTable = new int[dims];
            for (int cont1=0; cont1<dims; cont1++) {
                if (!(values[cont1] instanceof Integer)) { error.setError ("INVALID_ARRAY_INDEX", "DESCRIPTION: EL índice del elemento de la tabla debe ser entero", "TABLE_NAME: "+tableName); }
                else { indexTable[cont1] = (Integer) values[cont1]; }
            }

            if (error.getError()==null) {
                Object table = getAttribute (object, tableName, error, accessToPrivateFields);

                if (table!=null) {

                    int cont1 = 0;
                    int index;
                    while (cont1<dims-1 && error.getError()==null) {
                        index = indexTable[cont1];
                        try { table = Array.get (table, index); }
                        catch (ArrayIndexOutOfBoundsException e) { error.setError ("EXECUTE_ERROR", "DESCRIPTION: The table don't have this element"); }
                        catch (IllegalArgumentException e) { error.setError ("EXECUTE_ERROR", "DESCRIPTION: El objeto no es una tabla", "EXCEPTION: "+e.getMessage()); }

                        cont1++;
                    }

                    if (error.getError()==null) {
                        try {
                            index = indexTable[dims-1];
                            res = Array.get (table, index);
                        }
                        catch (ArrayIndexOutOfBoundsException e) { error.setError ("EXECUTE_ERROR", "DESCRIPTION: The table don't have this element"); }
                        catch (IllegalArgumentException e) { error.setError ("EXECUTE_ERROR", "DESCRIPTION: El objeto no es una tabla", "EXCEPTION: "+e.getMessage()); }
                    }
                }
                else { error.setError ("INVALID_ATTRIBUTE", "DESCRIPTION: No se ha encontrado la tabla", "TABLE_NAME: "+getSignature (tableName, null)); }
            }
        }
        else { error.setError ("EXECUTE_ERROR", "DESCRIPTION: Null pointer exception", "TABLE_NAME: "+tableName); }

        return res;
        //return "";
    }
    
    /**
     * Executa el métode "method" del objecte "object" i li passa els paràmtres "params".
     *
     * Retorna el resultat d'executar el métode i si es produeix algun error en retorna
     * la descripció amb la instància "error".
     */
    public Object executeMethod (Object object, Method method, Object[] params, ErrorEx error, boolean accessToPrivateFields) {

        Object res = null;

        if (object!=null) {
            if (method!=null) {

                try {
                    if (accessToPrivateFields) { method.setAccessible(true); }
                    res = method.invoke (object, params);
                    Class clase = method.getReturnType();
                    if (clase==void.class) { res = new ExpressionVoid(); }
                }
                catch (IllegalAccessException e) { error.setError ("INVALID_ATTRIBUTE", "DESCRIPTION: Se ha producido un error al ejecutar el método '"+method.getName()+"' de la clase'"+object.getClass().getSimpleName()+"'", "MESSAGE: "+e.getMessage()); }
                catch (IllegalArgumentException e) { error.setError ("INVALID_ATTRIBUTE", "DESCRIPTION: Se ha producido un error al ejecutar el método '"+method.getName()+"' de la clase'"+object.getClass().getSimpleName()+"'", "MESSAGE: "+e.getMessage()); }
                catch (InvocationTargetException e) { error.setError ("INVALID_ATTRIBUTE", "DESCRIPTION: Se ha producido un error al ejecutar el método '"+method.getName()+"' de la clase'"+object.getClass().getSimpleName()+"'", "MESSAGE: "+e.getMessage()); }
             }
        }
        else { error.setError ("EXECUTE_ERROR", "DESCRIPTION: Null pointer exception", "METHOD_NAME: "+method.getName()); }

        return res;
    }

    /**
     * Executa el métode "method" del objecte "object" i li passa els paràmtres "params".
     *
     * Retorna el resultat d'executar el métode i si es produeix algun error en retorna
     * la descripció amb la instància "error".
     *
     * NOTA: Només pot accedir al métode si aquest està definit com a "public".
     */

    public Object executeMethod (Object object, Method method, Object[] params, ErrorEx error) {
        return executeMethod (object, method, params, error, false);
    }
    
    /**
     * Executa el métode amb nom "methodName" del objecte "object" i li passa els paràmtres "params".
     *
     * Retorna el resultat d'executar el métode i si es produeix algun error en retorna
     * la descripció amb la instància "error".
     */
    public Object executeMethod (Object object, String methodName, Object[] values, ErrorEx error, boolean accessToPrivateFields) {

        Object res = null;

        if (object!=null) {
            Class[] params = getObjectClass (values);

            Method method = getMethod (object.getClass(), methodName, params);
            if (method!=null) {res = executeMethod (object, method, values, error, accessToPrivateFields); }
            else { error.setError ("INVALID_ATTRIBUTE", "DESCRIPTION: No se ha encontrado el método", "METHOD_NAME: "+getSignature (methodName, values)); }
        }
        else { error.setError ("EXECUTE_ERROR", "DESCRIPTION: Null pointer exception", "METHOD_NAME: "+getSignature (methodName, values)); }

        return res;
    }


    /**
     * Executa el métode amb nom "methodName" del objecte "object" i li passa els paràmtres "params".
     *
     * Retorna el resultat d'executar el métode i si es produeix algun error en retorna
     * la descripció amb la instància "error".
     *
     * NOTA: Només pot accedir al métode si aquest està definit com a "public".
     */
    public Object executeMethod (Object object, String methodName, Object[] values, ErrorEx error) {
        return executeMethod (object, methodName, values,  error, false);
    }
}
