//    _____ _                       _  _  ______ _      _     _  _____                                    _            _  _
//   / ____) |                     | || ||  ____)_)    | |   | |/ ____)                                  | |          | || |
//  | |    | |  _ _ ___ ___  ___   |_||_|| |___  _  ___| | __| | |     ___  _ _ __  _ __    _ _ _ __  _ _| |_ ___  _ ___||_|
//  | |    | |/ _` | __) __)/ _ \        |  ___)| |/ _ \ |/ _` | |    / _ \| ` `  \| `_ \ / _` | `__) _` | __) _ \| `__)
//  | |____| | (_| |__ \__ \  __/        | |    | |  __/ | (_| | |____ (_) | |\ |\ | |_) | (_| | | | (_| | |_ (_) | |
//   \_____)_|\__,_|___/___/\___|        |_|    |_|\___|_|\__,_|\_____)___/|_||_||_|  __/ \__,_|_|  \__,_|\__)___/|_|
//                                                                                 | |
//                                                                                 |_|


package cat.lib.sort;

import cat.lib.utils.StringUtils;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Comparator;

/**
 * La classe "FieldComparator" permet mitjançant el métode "sort" de la classe "Collection"
 * ordenar una llista d'elements per qualsevol dels seus atributs (sempre hi quan implementi
 * el seu corresponent métode "get" utilitzant l'standart "CamelCase").
 * 
 * Així doncs, per a ordenar una llista cal fer el següent:
 * 
 * |------------------------------------------------------------|
 * | Collections.sort (llista, new FieldComparator (atribut));  |
 * |------------------------------------------------------------|
 * 
 * on la "llista" és una implementació de la classe "Collections" com ara "Vector" i "atribut"
 * és el nom del atribut pel qual volem ordenar.
 * 
 * El métode "sort", necessita que el comparador sigui una classe que extengui la 
 * interfície "Comparator" i implementi el métode abstracte
 * "public int compare (Object o1, Object o2)".
 * 
 * Aquest métode és el que defineix el ordre en que es colocaran els diferents objectes.
 * 
 * La comparació entre els diferents elements de la llista és realitza de la següent manera:
 * 
 * - Si els objectes comparats NO són de la mateixa classe; no es comparen i es
 * consideren iguals.
 * 
 * - Si algun dels objectes comparats NO disposa del métode "get" corresponent al 
 * "atribut" o conté un valor "null"; no es comparen i es consideren iguals.
 * 
 * - Si els valors obtinguts són de caracter numéric (byte, short, int, long, float o double)
 * s'ordenen de menor a major.
 * 
 * - Si els valors obtinguts no són de caracter numéric, s'ordenen de menor a major
 * mitjançant la cadena obtinguda amb el métode "toString()" sense tenir en compte la
 * tipografia.
 *
 * @author Miquel Abadal Cervià
 */
public class FieldComparator implements Comparator {

    //     __   _        _ _           _
    //   / __ \| |      (_) |         | |
    //  | |__| | |_ _ __ _| |_   _   _| |_ ___
    //  |  __  | __) `__) | `_ \| | | | __) __)
    //  | |  | | |_| |  | | |_) | |_| | |_\__ \
    //  |_|  |_|\__)_|  |_|_,__/ \__,_|\__)___/
    //
    //

    private String fieldName = null;    // Camp utilitzat per a realitzar la comparació.
    private String methodName = null;

    
    //   __  __      _            _                              _                    _
    //  |  \/  |    | |          | |                            | |                  | |
    //  | \  / | ___| |_ ___   __| | ___     ____ ___  _ __  ___| |_ _ __ _   _  ____| |_ ___  _ __
    //  | |\/| |/ _ \ __) _ \ / _` |/ _ \   / ___) _ \| `_  \ __) __) `__) | | |/ ___) __) _ \| `__)
    //  | |  | |  __/ |_ (_) | (_| |  __/  | (___ (_) | | | |__ \ |_| |  | |_| | (___| |_ (_) | |
    //  |_|  |_|\___|\__)___/ \__,_|\___|   \____)___/|_| |_|___/\__)_|   \__,_|\____)\__)___/|_|
    //
    //

    public FieldComparator (String fieldName, String prefix) { setFieldName (fieldName, prefix); }


    //   __  __      _            _             _  _   _____ ______ _______ _  _
    //  |  \/  |    | |          | |           | || | / ____)  ____)__   __) || |
    //  | \  / | ___| |_ ___   __| | ___ ___   |_||_|| |  __| |__     | |  |_||_|
    //  | |\/| |/ _ \ __) _ \ / _` |/ _ \ __)        | | (_ |  __)    | |
    //  | |  | |  __/ |_ (_) | (_| |  __/__ \        | |__| | |____   | |
    //  |_|  |_|\___|\__)___/ \__,_|\___|___/         \_____|______)  |_|
    //
    //


    public String getFieldName () { return fieldName; }

    //   __  __      _            _             _  _   _____ ______ _______ _  _
    //  |  \/  |    | |          | |           | || | / ____)  ____)__   __) || |
    //  | \  / | ___| |_ ___   __| | ___ ___   |_||_|| (___ | |__     | |  |_||_|
    //  | |\/| |/ _ \ __) _ \ / _` |/ _ \ __)         \___ \|  __)    | |
    //  | |  | |  __/ |_ (_) | (_| |  __/__ \         ____) | |____   | |
    //  |_|  |_|\___|\__)___/ \__,_|\___|___/        (_____/|______)  |_|
    //
    //

    public void setFieldName (String fieldName, String prefix) {
        this.fieldName = fieldName;
        this.methodName = prefix+StringUtils.firstCharToUpperCase (fieldName, StringUtils.SUBST_ONE);
    }

    //   __  __      _            _                        _             _
    //  |  \/  |    | |          | |                      (_)           | |
    //  | \  / | ___| |_ ___   __| | ___ ___    _ __  _ __ _ _    _  _ _| |_ ___
    //  | |\/| |/ _ \ __) _ \ / _` |/ _ \ __)  | `_ \| `__) | |  | | _` | __) __)
    //  | |  | |  __/ |_ (_) | (_| |  __/__ \  | |_) | |  | |\ \/ / (_| | |_\__ \
    //  |_|  |_|\___|\__)___/ \__,_|\___|___/  |  __/|_|  |_| \__/ \__,_|\__)___/
    //                                         | |
    //                                         |_|

    private int compareValues (byte valor1, byte valor2) {

        if (valor1>valor2) { return 1; }
        else if(valor1 < valor2) { return -1; }
        else { return 0; }
    }

    private int compareValues (short valor1, short valor2) {

        if (valor1>valor2) { return 1; }
        else if(valor1 < valor2) { return -1; }
        else { return 0; }
    }

    private int compareValues (int valor1, int valor2) {

        if (valor1>valor2) { return 1; }
        else if(valor1 < valor2) { return -1; }
        else { return 0; }
    }

    private int compareValues (long valor1, long valor2) {

        if (valor1>valor2) { return 1; }
        else if(valor1 < valor2) { return -1; }
        else { return 0; }
    }

    private int compareValues (float valor1, float valor2) {

        if (valor1>valor2) { return 1; }
        else if(valor1 < valor2) { return -1; }
        else { return 0; }
    }

    private int compareValues (double valor1, double valor2) {

        if (valor1>valor2) { return 1; }
        else if(valor1 < valor2) { return -1; }
        else { return 0; }
    }

    private int compareValues (String valor1, String valor2) { 
    
        return StringUtils.compare(valor1, valor2, false);
    }
    
    private Object getFieldObject (Object object) {

        Object res = null;

        Class classe = object.getClass();

        try {
            Method method = classe.getMethod (methodName, new Class[] {});
            res = method.invoke  (object, new Object[] {});
        }
        catch (IllegalAccessException ex) {}
        catch (InvocationTargetException ex) {}
        catch (NoSuchMethodException ex) {}
        catch (NullPointerException ex) {}
        catch (SecurityException ex) {}

        return res;
    }

    //   __  __      _            _                     _     _ _
    //  |  \/  |    | |          | |                   | |   | |_)
    //  | \  / | ___| |_ ___   __| | ___    _ __  _   _| |_  | |_  ____ ___
    //  | |\/| |/ _ \ __) _ \ / _` |/ _ \  | `_ \| | | | `_ \| | |/ ___) __)
    //  | |  | |  __/ |_ (_) | (_| |  __/  | |_) | |_| | |_) | | | (___\__ \
    //  |_|  |_|\___|\__)___/ \__,_|\___|  |  __/ \__,_|_,__/|_|_|\____)___/
    //                                     | |
    //                                     |_|


    public int compare (Object o1, Object o2) {

        int res = 0;

        Class classeObject1 = o1.getClass();
        Class classeObject2 = o2.getClass();

        if (classeObject1==classeObject2) {
        
            Object valueObject1 = getFieldObject (o1);
            Object valueObject2 = getFieldObject (o2);

            if (valueObject1!=null && valueObject2!=null) {
                if (valueObject1 instanceof Byte) { res = compareValues ((Byte) valueObject1, (Byte) valueObject2); }
                else if(valueObject1 instanceof Short) { res = compareValues((Short) valueObject1, (Short) valueObject2); }
                else if (valueObject1 instanceof Integer) { res = compareValues ((Integer) valueObject1, (Integer) valueObject2); }
                else if (valueObject1 instanceof Long) { res = compareValues ((Long) valueObject1, (Long) valueObject2); }
                else if (valueObject1 instanceof Double) { res = compareValues ((Double) valueObject1, (Double) valueObject2); }
                else { res = compareValues (valueObject1.toString(), valueObject2.toString()); }
            }
        }

        return res;

    }

    public boolean equals (Object o) { return this == o; }

}
