//    _____ _                       _  _   _____ _        _              _______   _    _           _            _  _
//   / ____) |                     | || | / ____) |      (_)            (__   __) | |  | |         | |          | || |
//  | |    | |  _ _ ___ ___  ___   |_||_|| (___ | |_ _ __ _ _ __   __ _ ___| | ___| |  | |___  ____| |_ ___  _ ___||_|
//  | |    | |/ _` | __) __)/ _ \         \___ \| __) `__) | `_  \/ _` | __) |/ _ \ |  | | _ \/ ___) __) _ \| `__)
//  | |____| | (_| |__ \__ \  __/         ____) | |_| |  | | | | | (_| |__ \ | (_) | \/ /  __/ (___| |_ (_) | |
//   \_____)_|\__,_|___/___/\___|        (_____/ \__)_|  |_|_| |_|\__, |___/_|\___/ \__/ \___|\____)\__)___/|_|
//                                                                 __| |
//                                                                (___/

package cat.lib.strings;

import cat.lib.utils.SetUtils;
import cat.lib.utils.StringUtils;
import java.util.Vector;

/**
 * La classe <code>StringsToVector</code> permet segmentar una cadena de text
 * que utilitza un caracter "separador" en diverses cadenes i després
 * recuperarles a mitjançant un índex.
 *
 * Aquesta classe està pensada per a que serveixi com a base per a crear classes
 * més concretes com <code>ParamsList</code> o <code>StringsList</code>.
 *
 * La primera permet segmentar cadenes de text que contenen un nombre indeterminat
 * de paràmetres utilitzant qualsevol tipus de caràcter separador, habitualment
 * el ';' o la ','. A més a més, és capaç d'identificar els literals delimitats
 * entre cometes simples o dobles i ignorar els caràcters separadors que hi trobi.
 *
 * Mentre que la segona permet segmentar cadenes de text que contenen un nombre
 * indeterminat de línies utilitzant com el caràcter salt de línia '\n' com a
 * separador. En aquest cas però, no interessa detectar els literals i per tant
 * no es poden descartar caracters separadors.
 *
 * Per a enmagatzemar la llista de cadenes utilitza un <code>Vector</code> que
 * permet un accés directe als diferents elements.
 *
 * Tanmateix, la classe és plenament funcional (malgrat que està definida com a
 * <code>abstract</code> i els métodes no són publics), de manera que les
 * classes que l'extenen únicament han de re-encapsular els métodes
 * renombrant-los si és necessari o sobreescribint-los.
 *
 *
 * @author Miquel Abadal i Cervià
 */
public abstract class StringsToVector {

    //     __   _        _ _           _
    //   / __ \| |      (_) |         | |
    //  | |__| | |_ _ __ _| |_   _   _| |_ ___
    //  |  __  | __) `__) | `_ \| | | | __) __)
    //  | |  | | |_| |  | | |_) | |_| | |_\__ \
    //  |_|  |_|\__)_|  |_|_,__/ \__,_|\__)___/
    //
    //

    private Vector lines = new Vector ();

    private char separador;
    private int numLines = 0;
    private boolean autoTrim = false;
    private boolean detectLiterals = false;

    public static boolean DONT_DETECT_LITERALS = false;
    public static boolean DETECT_LITERALS = true;
    public static boolean AUTO_TRIM = true;
    public static boolean NO_AUTO_TRIM = false;

    //   __  __      _            _                                  _                    _
    //  |  \/  |    | |          | |                                | |                  | |
    //  | \  / | ___| |_ ___   __| | ___ ___     ____ ___  _ __  ___| |_ _ __ _   _  ____| |_ ___  _ __ ___
    //  | |\/| |/ _ \ __) _ \ / _` |/ _ \ __)   / ___) _ \| `_  \ __) __) `__) | | |/ ___) __) _ \| `__) __)
    //  | |  | |  __/ |_ (_) | (_| |  __/__ \  | (___ (_) | | | |__ \ |_| |  | |_| | (___| |_ (_) | |  \__ \
    //  |_|  |_|\___|\__)___/ \__,_|\___|___/   \____)___/|_| |_|___/\__)_|   \__,_|\____)\__)___/|_|  (___/
    //
    //

    public StringsToVector (String text, char separador, boolean autoTrim, boolean detectLiterals) {
        this.autoTrim = autoTrim;
        this.detectLiterals = detectLiterals;
        setText (text, separador);
    }

    public StringsToVector (String text, char separador, boolean detectLiterals) {
        this.autoTrim = NO_AUTO_TRIM;
        this.detectLiterals = detectLiterals;
        setText (text, separador);
    }

    public StringsToVector (String text, boolean detectLiterals) {
        this.autoTrim = NO_AUTO_TRIM;
        this.detectLiterals = detectLiterals;
        setText (text, '\n');
    }


    //   __  __      _            _                         _     _ _
    //  |  \/  |    | |          | |                       | |   | |_)
    //  | \  / | ___| |_ ___   __| | ___ ___    _ __  _   _| |_  | |_  ____ ___
    //  | |\/| |/ _ \ __) _ \ / _` |/ _ \ __)  | `_ \| | | | `_ \| | |/ ___) __)
    //  | |  | |  __/ |_ (_) | (_| |  __/__ \  | |_) | |_| | |_) | | | (___\__ \
    //  |_|  |_|\___|\__)___/ \__,_|\___|___/  |  __/ \__,_|_,__/|_|_|\____)___/
    //                                         | |
    //                                         |_|

    /** Esborra totes les linies.
     *
     */
    protected void clear () {
        lines.clear();
        numLines=0;
    }

    /**
     * Afegueix una línia a la llista
     *
     * @param text Text a afeguir a la llista
     */
    protected void add (String text, boolean isParam) {
        if (text!=null) {
            lines.add (numLines, text);
            numLines++;
        }
    }

    protected void add (String text) { add (text, false); }

    /**
     * Esborra la línia situada a la posició <code>index</code>.
     *
     * <code>Nota:</code>Si el índex està fora de rang no realitza cap acció.
     *
     * @param index Posició de la línia a eliminar.
     */
    protected void remove (int index) {
        if (index>=0 && index<numLines) {
            String origLine = getLine(index);
            lines.remove (index);
            numLines--;
        }
    }


    //   __  __      _            _             _  _   _____ ______ _______ _  _
    //  |  \/  |    | |          | |           | || | / ____)  ____)__   __) || |
    //  | \  / | ___| |_ ___   __| | ___ ___   |_||_|| (___ | |__     | |  |_||_|
    //  | |\/| |/ _ \ __) _ \ / _` |/ _ \ __)         \___ \|  __)    | |
    //  | |  | |  __/ |_ (_) | (_| |  __/__ \         ____) | |____   | |
    //  |_|  |_|\___|\__)___/ \__,_|\___|___/        (_____/|______)  |_|
    //
    //

    /**
     * Esborra totes les línies i en genera de noves utilitzant el
     * <code>text</code> indicat i el separador que cal utilitzar.
     *
     * <code>Nota:</code>Al construïr la llista de línies si s'ha activat el
     * atribut <code>detectLiterals</code>, no utilitza els caràcters separadors
     * situats a dintre d'un literal.
     *
     * @param text Text original
     * @param separador Caràcter utilitzat per a separar les línies
     */
    protected void setText (String text, char separador) {

        this.separador = separador;
        if (separador==' ') text = StringUtils.blankAdv (text);

        clear();
        if (text!=null) {

            int textSize = text.length();
            StringBuffer buffer = new StringBuffer (textSize);

            char caract;
            char literalChar = '\0';
            boolean onLiteral = false;

            for (int cont1=0; cont1<textSize; cont1++) {
                caract = text.charAt (cont1);
                if (onLiteral==false) {
                    if (detectLiterals && SetUtils.charIn (caract, false, '\'', '"')) {
                        onLiteral=true;
                        literalChar=caract;
                    }

                    if (caract==separador) {
                        lines.add (numLines, buffer.toString());
                        numLines++;
                        buffer.setLength (0);
                    }
                    else { buffer = buffer.append (caract); }
                }
                else {
                    if (caract==literalChar) { onLiteral=false; }
                    buffer = buffer.append (caract);
                }
            }

            if (buffer.length()>0) {
                lines.add (numLines, buffer.toString());
                numLines++;
            }
        }
    }

    /**
     * Modifica el contingut de la línia situada a la posició <code>index</code>
     * i li assigna el valor <code>text</code>.
     *
     * <code>Nota:</code>Si el índex està fora de rang no realitza cap acció.
     *
     * @param index Índex de la línia a modificar
     * @param text Nou text a assignar a la línia
     */
    public void setLine (int index, String text) {
        if (index>=0 && index<numLines && text!=null) { lines.setElementAt (text, index); } }

    public void setLastLine (String text) { setLine (numLines-1, text); }
    

    /**
     * Modifica el atribut <code>autoTrim</code> que seveix per a indicarli als
     * métodes de recuperació de línies (inclos el métode <code>getText</code>
     * que reconstrueix la cadena, que ha d'eliminar tots els caràcters en blanc
     * que apareguin al principi i final de cadascuna de les línies
     *
     * @param autoTrim <code>true</code> si es vol activar la funció <code>
     * autoTrim</code> o <code>false</code> en cas contrari.
     */
    protected void setAutoTrim (boolean autoTrim) { this.autoTrim = autoTrim; }



    //   __  __      _            _             _  _   _____ ______ _______ _  _
    //  |  \/  |    | |          | |           | || | / ____)  ____)__   __) || |
    //  | \  / | ___| |_ ___   __| | ___ ___   |_||_|| |  __| |__     | |  |_||_|
    //  | |\/| |/ _ \ __) _ \ / _` |/ _ \ __)        | | (_ |  __)    | |
    //  | |  | |  __/ |_ (_) | (_| |  __/__ \        | |__| | |____   | |
    //  |_|  |_|\___|\__)___/ \__,_|\___|___/         \_____|______)  |_|
    //
    //

    /**
     * Reconstrueix la cadena de text a partir de les línies actuals i unint-los
     * mitjançant el caràcter separador.
     *
     * @return Retorna una cadena de text amb totes les línies separades amb el
     * caràcter separador.
     */
    @Override
    public String toString  () {

        StringBuilder res = new StringBuilder ();
        for (int cont1=0; cont1<numLines; cont1++) {
            String line = (String) lines.get (cont1);

            if (autoTrim) line = line.trim();

            res.append (line);
            if (cont1<numLines-1) res.append (separador);
        }

        return res.toString();
    }


    protected String getText () { return toString (); }

    /**
     * Obté el contingut de la línia situada a la posició <code>index</code>.
     *
     * <code>Nota:</code>Si el índex està fora de rang retorna amb el valor
     * <code>null</code>
     *
     * @param index Índex de la línia
     * @return Retorna el contingut de la línia
     */
    protected String getLine (int index) {
        String res = null;
        if (index>=0 && index<numLines) {
            res = (String) lines.get (index);
            if (autoTrim) res = res.trim();
        }
        return res;
    }


    /**
     * Retorna el número de línies disponibles.
     *
     * @return Número de línies
     */
    protected int getSize () { return numLines; }

    /**
     * Obté el estat del atribut <code>autoTrim</code>.
     *
     * @return <code>true</code> si està activat o <code>false</code> en cas
     * contrari.
     */
    protected boolean isAutoTrim () { return autoTrim; }

    /** Retorna el vector amb les linies
     *
     */
    protected Vector getLines() { return lines; }

}

