/*
 *   _____ _                       _  _ __    __        _ __   _           _     _  _
 *  / ____) |                     | || |\ \  / /       | |  \ | |         | |   | || |
 * | |    | |  _ _ ___ ___  ___   |_||_| \ \/ / _ _ __ | |   \| | ___   __| | ____||_|
 * | |    | |/ _` | __) __)/ _ \          )  ( | ` `  \| | |\ ` |/ _ \ / _` |/ _ \
 * | |____| | (_| |__ \__ \  __/         / /\ \| |\ |\ | | | \  | (_) | (_| |  __/
 *  \_____)_|\__,_|___/___/\___|        /_/  \_\_||_||_|_|_|  \_|\___/ \__,_|\___|
 *
 */

package cat.lib.xml;

import cat.lib.utils.StringUtils;

/**
 * La classe "XmlNode" encapsula la informació d'un node XML.
 *
 * <p>Cadascun dels nodes s'indentifica per tres atributs (accessibles amb els
 * corresponents métodes "get/set"):
 *
 * <p><b>path</b>: Llista de tots els nodes "pare" concatenats amb el caracter '.'.<br>
 * <b>value</b>: Valor del node.<br>
 * <b>type</b>: Tipus de node.
 *
 * <p>Hi ha els següents tipus de nodes:
 *
 * <p><b>ROOT_NODE</b>: Nodes de tipus "root".<br>
 * <b>COMMENT_NODE</b>: Nodes de tipus comentari.<br>
 * <b>LABEL_NODE</b>: Nodes de tipus etiqueta.<br>
 * <b>ATTRIBUTE_NODE</b>: Nodes de tipus atribut.<br>
 * <b>DATA_NODE</b>: Nodes de tipus dada.<br>
 *
 * <p>Una instrucció XML com la següent:
 *
 * <p>[etiqueta atribut1="valor1" atribut2="valor2"]>valor3[/LABEL]
 *
 * <p>Crea els següents nodes:
 *
 *
 * <code>
 * |----------------|----------------|----------------------| <br>
 * | Type           | value          | path                 | <br>
 * |----------------|----------------|----------------------| <br>
 * | LABEL_NODE     | etiqueta       |                      | <br>
 * | ATTRIBUTE_NODE | atribut1       | etiqueta             | <br>
 * | DATA_NODE      |  valor1        | etiqueta.parametre1  | <br>
 * | ATTRIBUTE_NODE | atribut2       | etiqueta             | <br>
 * | DATA_NODE      | valor2         | etiqueta.parametre2  | <br>
 * | DATA_NODE      | valor3         | etiqueta             | <br>
 * |----------------|----------------|----------------------| <br>
 * </code>
 *
 * @author Miquel Abadal Cervià
 * @date 16-03-2010
 */
public class XmlNode {

    //    _____                _              _           
    //   / ____)              | |            | |          
    //  | |     ___  _ __  ___| |_  _ _ _ __ | |_ ___     
    //  | |    / _ \| `_  \ __) __) _` | `_  \ __) __)    
    //  | |____ (_) | | | |__ \ |_ (_| | | | | |_\__ \    
    //   \_____)___/|_| |_|___/\__)__,_|_| |_|\__)___/    
    //                                                    
    //                                                    

    /** Node de tipus "root". */
    public final static int ROOT_NODE = 0;
    /** Node de tipus comentari. */
    public final static int COMMENT_NODE = 1;
    /** Node de tipus etiqueta. */
    public final static int LABEL_NODE = 2;
    /** Node de tipus paràmetre. */
    public final static int ATTRIBUTE_NODE = 3;
    /** Node de tipus valor. */
    public final static int DATA_NODE = 4;
    /** Node de tipus docType. */
    public final static int DOCTYPE_NODE = 5;


    //     __   _        _ _           _
    //   / __ \| |      (_) |         | |
    //  | |__| | |_ _ __ _| |_   _   _| |_ ___
    //  |  __  | __) `__) | `_ \| | | | __) __)
    //  | |  | | |_| |  | | |_) | |_| | |_\__ \
    //  |_|  |_|\__)_|  |_|_,__/ \__,_|\__)___/
    //
    //
    
    private String value;
    private int type = -1;

    protected XmlNode parent = null;
    protected XmlNode previousBrother = null;
    protected XmlNode nextBrother = null;
    protected XmlNode firstChild = null;

    //   __  __      _            _                                  _                    _
    //  |  \/  |    | |          | |                                | |                  | |
    //  | \  / | ___| |_ ___   __| | ___ ___     ____ ___  _ __  ___| |_ _ __ _   _  ____| |_ ___  _ __ ___
    //  | |\/| |/ _ \ __) _ \ / _` |/ _ \ __)   / ___) _ \| `_  \ __) __) `__) | | |/ ___) __) _ \| `__) __)
    //  | |  | |  __/ |_ (_) | (_| |  __/__ \  | (___ (_) | | | |__ \ |_| |  | |_| | (___| |_ (_) | |  \__ \
    //  |_|  |_|\___|\__)___/ \__,_|\___|___/   \____)___/|_| |_|___/\__)_|   \__,_|\____)\__)___/|_|  (___/
    //
    //

    /**
     * Métode constructor
     *
     * @param value Valor del node
     * @param type Tipus de node
     */
    public XmlNode (String value, int type) {
        this.value = value;
        this.type = type;
        this.parent = null;
        this.previousBrother = null;
        this.nextBrother = null;
        this.firstChild = null;
    }

    //   __  __      _            _             _  _   _____ ______ _______ _  _
    //  |  \/  |    | |          | |           | || | / ____)  ____)__   __) || |
    //  | \  / | ___| |_ ___   __| | ___ ___   |_||_|| |  __| |__     | |  |_||_|
    //  | |\/| |/ _ \ __) _ \ / _` |/ _ \ __)        | | (_ |  __)    | |
    //  | |  | |  __/ |_ (_) | (_| |  __/__ \        | |__| | |____   | |
    //  |_|  |_|\___|\__)___/ \__,_|\___|___/         \_____|______)  |_|
    //
    //

    /**
     * Obté la ruta del node.
     *
     * @return Ruta del node.
     */
    public String getPath () { 
        String path = "";
        XmlNode node = this.parent;
        while (node!=null) {
            path = StringUtils.concat (node.value, ".", path);
            node = node.parent;
        }
        return path;
    }

    /**
     * Obté la profunditat del node.
     * 
     * @return
     */
    public int getDeep () {
        int deep = 0;
        XmlNode node = this.parent;
        while (node!=null) {
            deep++;
            node = node.parent;
        }
        return deep;
    }

    /**
     * Obté el valor del primer fill de node.
     * 
     * Si el node no té fills, returna <b>null</b>.
     * 
     * @return Valor del primer fill del node.
     */
    public String getFirstChildValue() {

        String res = null;
        XmlNode childNode = this.firstChild;
        if (childNode!=null) { res = childNode.getValue(); }

        return res;
    }

    /**
     * Obté el valor del node.
     *
     * @return Valor del node.
     */
    public String getValue () { return value; }

    /**
     * Obté el tipus de node.
     *
     * <p>Pot prendre els següents valors:
     *
     * <p><b>ROOT_NODE</b>: Node de tipus "root".<br>
     * <b>COMMENT_NODE</b>: Node de tipus comentari.<br>
     * <b>LABEL_NODE</b>: Node de tipus etiqueta.<br>
     * <b>ATTRIBUTE_NODE</b>: Node de tipus atribut.<br>
     * <b>DATA_NODE</b>: Node de tipus data.<br>
     * <b>DOCTYPE_NODE</b>: No de de tipus doctypè.<br>
     *
     * @return Tipus de node
     */
    public int getType () { return type; }

    /**
     * Obté la descripció del tipus de node.
     *
     * @return Descripció del tipus de node.
     */
    public String getTypeDescription() {

        String res = "";
        switch (type) {
            case ROOT_NODE : res = "ROOT"; break;
            case COMMENT_NODE : res = "COMMENT"; break;
            case LABEL_NODE   : res = "LABEL"; break;
            case ATTRIBUTE_NODE   : res = "ATTRIBUTE"; break;
            case DATA_NODE    : res = "DATA"; break;
            case DOCTYPE_NODE : res = "DOCTYPE"; break;
            default  : res = "UNKNOW";
        }
        return res;

    }

    /**
     * Obté el node "root".
     * 
     * @return Node "root".
     */
    public XmlNode getRootNode() {
        XmlNode node = this;
        while (node.parent!=null) { node = node.parent; }

        return node;
    }

    /**
     * Obté el "pare" del node actual.
     *
     * @return "Pare" del node actual;
     */
    public XmlNode getParent() { return parent; }

    /**
     * Obté el primer germà. El més "gran".
     *
     * <p><b>NOTA</B>: Els nodes estàn ordenats de més antic a més nou, (o sigui
     * que el primer és el germà més "vell" i el últim el més "jove".
     *
     * @return Primer germà
     */
    public XmlNode getFirstBrother() {
        XmlNode node = this;
        while (node.previousBrother!=null) { node = node.previousBrother; }
        return node;
    }

    /**
     * Obté el germà anterior
     *
     * @return Germà anterior
     */
    public XmlNode getPreviousBrother() { return previousBrother; }

    /**
     * Obté el següent germà.
     *
     * @return Següent germà
     */
    public XmlNode getNextBrother() { return nextBrother; }

    /**
     * Obté el últim germà. El més "petit".
     *
     * <p><b>NOTA</B>: Els nodes estàn ordenats de més antic a més nou, (o sigui
     * que el primer és el germà més "vell" i el últim el més "jove".
     *
     * @return Últim germà
     */
    public XmlNode getLastBrother() {
        XmlNode node = this;
        while (node.nextBrother!=null) { node = node.nextBrother; }
        return node;
    }

    /**
     * Obté el primer fill. El més "gran"
     *
     * <p><b>NOTA</B>: Els nodes estàn ordenats de més antic a més nou, (o sigui
     * que el primer és el germà més "vell" i el últim el més "jove".
     *
     * @return Primer fill
     */
    public XmlNode getFirstChild() { return firstChild; }

    /**
     * Obté el últim fill. El més "petit"
     *
     * <p><b>NOTA</B>: Els nodes estàn ordenats de més antic a més nou, (o sigui
     * que el primer és el germà més "vell" i el últim el més "jove".
     *
     * @return Últim fill
     */
    public XmlNode getLastChild() {
        XmlNode node = firstChild;
        if (node!=null) { node.getLastBrother(); }
        return node;
    }

    /**
     * Obté el primer fill que conté com a "value" el valor indicat.
     * 
     * Si no en troba cap, retorna <b>null</b>.
     * 
     * @param value Valor del fill a cercar.
     * @return
     */
    public XmlNode getChild (String value) {

        XmlNode trobat = null;
        XmlNode node = firstChild;
        while (node!=null && trobat==null) {
            if (node.value!=null && node.value.equalsIgnoreCase(value)) { trobat = node; }
            node = node.getNextBrother();
        }
        return trobat;
    }

    /**
     * Obté el node anterior.
     *
     * <p><b>NOTA:</b> Amb aquest métode és pot recorrer recursivement en ordre
     * invers la llista de nodes. Desplegant-se primer cap els fills "petits",
     * després cap els germans "grans"
     *
     * @return Node anterior
     */
    public XmlNode getPreviousNode() {

        XmlNode node = this;
        XmlNode trobat = null;

        if (firstChild!=null) { trobat = firstChild.getLastBrother(); }
        else if (previousBrother!=null) { trobat = previousBrother; }
        else {
            while (trobat==null && node!=null) {
                if (node.parent!=null && node.parent.previousBrother!=null) { trobat = node.parent.previousBrother; }
                else { node = node.parent; }
            }
        }

        return trobat;
    }

    /**
     * Obté el node següent.
     *
     * <p><b>NOTA:</b> Amb aquest métode és pot recorrer recursivement la llista
     * de nodes. Desplegant-se primer cap els fills "grans", després cap els germans
     * "petits"
     *
     * @return Node següent
     */
    public XmlNode getNextNode() {

        XmlNode node = this;
        XmlNode trobat = null;

        if (firstChild!=null) { trobat = firstChild; }
        else if (nextBrother!=null) { trobat = nextBrother; }
        else {
            while (trobat==null && node!=null) {
                if (node.parent!=null && node.parent.nextBrother!=null) { trobat = node.parent.nextBrother; }
                else { node = node.parent; }
            }
        }

        return trobat;
    }


    //   __  __      _            _             _  _   _____ ______ _______ _  _
    //  |  \/  |    | |          | |           | || | / ____)  ____)__   __) || |
    //  | \  / | ___| |_ ___   __| | ___ ___   |_||_|| (___ | |__     | |  |_||_|
    //  | |\/| |/ _ \ __) _ \ / _` |/ _ \ __)         \___ \|  __)    | |
    //  | |  | |  __/ |_ (_) | (_| |  __/__ \         ____) | |____   | |
    //  |_|  |_|\___|\__)___/ \__,_|\___|___/        (_____/|______)  |_|
    //
    //

    /**
     * Permet assignar el valor del node
     *
     * @param value Valor del node
     */
    public void setValue (String value) { this.value = value; }

    /**
     * Permet assignar el tipus de node.
     *
     * <p>Pot prendre els següents valors:
     *
     * <p><b>ROOT_NODE</b>: Node de tipus comentari.<br>
     * <b>COMMENT_NODE</b>: Node de tipus comentari.<br>
     * <b>LABEL_NODE</b>: Node de tipus etiqueta.<br>
     * <b>PARAM_NODE</b>: Node de tipus paràmetre.<br>
     * <b>DATA_NODE</b>: Node de tipus data.<br>
     * <b>DOCTYPE_NODE</b>: No de de tipus doctypè.<br>
     *
     * @param nodeType Tipus de node.
     */
    public void setType (int type) { this.type = type; }

    /**
     * Extreu el node actual del arbre en el que estava, i l'inserta com a últim
     * fill del node "<b>newParent</b>".
     * 
     * <b>NOTA</b>: Si el nou pare és <b>null</b>, llavors extreu el node del arbre
     * que es trobava i el deixa com a node "root".
     * 
     * @param newParent "Nou pare" del node.
     */
    public void setParent (XmlNode newParent) {

        // Extreu el node del arbre en que està.
        if (previousBrother!=null) { previousBrother.nextBrother = nextBrother; }
        if (nextBrother!=null) { nextBrother.previousBrother = previousBrother; }
        if (parent!=null && parent.firstChild==this) {parent.firstChild = nextBrother; }

        previousBrother = null;
        nextBrother = null;
        parent = newParent;

        if (newParent!=null) {

            // Afegueix el node com a fill de "newParent" i l'enllaça amb els "nous" germans.
            if (newParent.firstChild!=null) {
                XmlNode node = newParent.firstChild.getLastBrother();
                node.nextBrother = this;
                previousBrother = node;
            }
            else { newParent.firstChild = this; }
        }
    }

    /**
     * Retorna una cadena de text amb el node en formnat <b>XML</b>.
     *
     * @return Node en format <b>XML</b>.
     */
    @Override
    public String toString() { return value; }
}
