//    _____ _                       _  _ __    __        _ ____                     ______      _  _
//   / ____) |                     | || |\ \  / /       | |  __ \                  |  ____)    | || |
//  | |    | |  _ _ ___ ___  ___   |_||_| \ \/ / _ _ __ | | |__) |_ _ _ __ ___  ___| |__  _   ___||_|
//  | |    | |/ _` | __) __)/ _ \          )  ( | ` `  \| |  ___/ _` | `__) __)/ _ \  __) \ \/ /
//  | |____| | (_| |__ \__ \  __/         / /\ \| |\ |\ | | |  | (_| | |  \__ \  __/ |____ )  (
//   \_____)_|\__,_|___/___/\___|        /_/  \_\_||_||_|_|_|   \__,_|_|  (___/\___|______)_/\_\
//
//


package cat.lib.xml;

import cat.lib.errors.ErrorEx;
import cat.lib.files.DataFile;
import cat.lib.strings.ParamsList;
import cat.lib.utils.StringUtils;
import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.io.UnsupportedEncodingException;


/**
 *
 * La classe "<B>XmlParserEx</B>" permet interpretar un fitxer XML i processar-ne, seqüencialment
 * tots els nodes, o recuperar-los en forma de "<b>Vector</b>".
 *
 * <p>Cadascuna de les instàncies de la classe "XmlParserEx" té els següents atributs (als
 * quals es pot accedit mitjançant els corresponents métodes "get/set":
 *
 * <p>
 * - <b>caseSensitive</b> : Indica si el interpret de Xml és sensible a la tipografia o no.<br>
 * - <b>createXmlTree</b> : Indica si s'ha de crear l'arbre de nodes o no (especialment
 * interessant per a tractar documents molt grossos que consumeixen molta memòria).<br>
 * - <b>xmlVersion</b> : Conté la versió del document Xml. Aquesta informació sobté al interpretar un
 * document Xml.<br>
 * - <b>xmlEncoding</b> : Conté la codificació del document Xml. Aquesta informació sobté al interpretar un
 * document Xml.<br>
 * "<b>loadFromFile()</b>".<br>
 * 
 * <p>Per a interpretar un document XML, cal utilitzar el métode "<b>parse()</b>" i indicarli la font de dades
 * que pot ser un String, un Stream o un fitxer de disc.
 * 
 * <p>Aquest métode interpreta el document Xml, i per cada node executa un event "<b>onParse</b>".
 * 
 * <p>Per a interceptar-lo cal crear una classe que implementi la interfície "<b>Parseable</b>" i
 * haber construït l'objecte "<b>XmlParserEx</b>" indicant-li en el constructor l'esmentada
 * classe.
 * 
 * <p><code>
 * <font color=navy>public class</font> <b>MyClass</b> <font color=navy>implements Parseable</font> {<br><br>
 *   <font color=navy>public boolean</font> <b>onParse</b> (XmlNode node) {<br><br>
 *       System.out.println (node);<br>
 *       <font color=navy>return false;</font></br>
 *   }<br><br>
 *
 *   <font color=navy>public void</font> <b>process</b> () {<br><br>
 *       XmlParserEx xml = <font color=navy>new</font> XmlParserEx (this);<br>
 *       xml.setCaseSensitive (false);<br><br>
*
*        xml.loadFromFile (filename);<br><br>
 *
 *       xml.parse();<br>
 *   }<br>
 * }<br>
 * </code>
 *
 * <p>En el exemple anterior, s'intercepta el métode "<B>onParse</b>" en la mateixa
 * classe que la que executa el procés, pero poden estar en classes separades.
 *
 * <p>El métode "<b>onParse</b>" ha de retornar una variable <b>booleana</b> que indica si ja s'ha
 * trobat el node cercat, o no. En el cas d'haber-lo trobat cal retornar un <b>true</b>
 * i el interpret dóna el procés per acabat.
 * 
 * <p>Si en canvi no ens interessa interceptar el event "<b>onParse</b>" podem crear la
 * instància "<b>XmlParserEx</b>" passant-li un <b>null</b> en el métode constructor i després
 * de completar-se la interpretació del document amb el métode "<b>parse()</b>" podem obtenir
 * la llista de nodes amb el métode "<b>getNodes()</b>" o cercar-ne un de concret amb
 * el métode "<b>getNode()</b>".
 * 
 * <p><code>
 * <font color=navy>public class</font> <b>MyClass</b> {<br><br>
 *
 *   <font color=navy>public void</font> <b>process</b> () {<br><br>
 *       XmlParserEx xml = <font color=navy>new</font> XmlParserEx (null);<br>
 *       xml.setCaseSensitive (false);<br><br>
*
*        xml.parseFromFile (filename);<br><br>
 *
 *       XmlNode = xml.getRootNode();<br>
 *   }<br>
 * }<br>
 * </code>
 *
 * <p>Un cop s'ha generat la llista de nodes, la podem utilitzar per a recorre-la
 * recursivament amb els métodes "<b>getNextNode()</b>" i "<b>getPreviousNode</b>", per a cercar
 * un node concret amb els métodes "<b>searchFirstNode()</b>" i "<b>searchNextNode</b>" o per a
 * "reconstruïr" el document Xml original, mitjançant el métode "<b>generateXml</b>".
 *
 * <p>També, si es desitja, es pot grabar el document generat a un arxiu, utilitzant el métode
 * "<b>saveToFile</b>".
 *
 * <p><b>NOTA</b>: Si es genera el arbre XML, cal tenir en compte que tots els arbres comencen
 * amb un node de nom "root" de tipus "ROOT_NODE" que fa de pare únic de la resta de nodes.
 * Tanmateix, al fer la recerca per path, si no s'indica, el cercador l'inclou implícitament.
 *
 * Els dos documents Xml següents, són equivalents i generen una mateixa estructura de dades:
 *
 * |------------------|   |------------------------------------------|
 * | [label1]         |   | [label1 param1="valor1" param2=valor2"]  |
 * |     [param1]     |   |     valor3                               |
 * |         valor1   |   | [/label1]                                |
 * |     [/param1]    |   | [label2/]                                |
 * |     [param2]     |   |------------------------------------------|
 * |         valor2   |
 * |     [/param2]    |
 * |     valor3       |       |========(0)=|
 * | [/label1]        |       | name: root |
 * | [label2/]        |       | type: ROOT |
 * |------------------|       |============|
 *                                  |
 *                          /-------|-------\
 *                          |               |
 *                    |-----|-(1)-|   |-----|-(7)-|
 *                    | label1    |   | label2    |
 *                    | LABEL     |   | LABEL     |
 *                    |-----|-----|   |-----------|
 *                          |
 *         /----------------|---------------\
 *         |                |               |
 *   |-----|-(2)-|    |-----|-(4)-|   |-----|-(6)-|
 *   | param1    |    | param2    |   | valor3    |
 *   | ATTRIBUTE |    | ATTRIBUTE |   | DATA      |
 *   |-----|-----|    |-----|-----|   |-----------|
 *         |                |
 *         |                |
 *         |                |
 *   |-----|-(3)-|    |-----|-(5)-|
 *   | valor1    |    | valor2    |
 *   | DATA      |    | DATA      |
 *   |-----------|    |-----------|
 *
 *  <p><b>POSSIBLES MILLORES</b>: Actualment, per a cercar un node és recorre la
 * estructura recursivament (seguint la numeració dels nodes indicada en el gràfic),
 * però es podria fer una recerca dicotòmica entrant únicament a la branca que ens
 * interessa
 *
 * <P>Aixi cercar el element amb path: "root.label1.param2", només li caldrien 3 consultes.
 * Ara n'ha de fer cinc:
 *
 *
 *
 * @author Miquel Abadal Cervià
 * @date 16-03-2010
 */
public class XmlParserEx {

    //    _____                _              _
    //   / ____)              | |            | |
    //  | |     ___  _ __  ___| |_  _ _ _ __ | |_ ___
    //  | |    / _ \| `_  \ __) __) _` | `_  \ __) __)
    //  | |____ (_) | | | |__ \ |_ (_| | | | | |_\__ \
    //   \_____)___/|_| |_|___/\__)__,_|_| |_|\__)___/
    //
    //

    private final static String XML_ERROR = "XML_ERROR";
    private final char CR = 13;
    private final char LF = 10;
    private final String ROOT_NODE_NAME = "xmlParserExRootNode";
    private final String DEFAULT_ENCODING ="utf-8";
    private final int HEADER_SIZE = 100;



    //     __   _        _ _           _
    //   / __ \| |      (_) |         | |
    //  | |__| | |_ _ __ _| |_   _   _| |_ ___
    //  |  __  | __) `__) | `_ \| | | | __) __)
    //  | |  | | |_| |  | | |_) | |_| | |_\__ \
    //  |_|  |_|\__)_|  |_|_,__/ \__,_|\__)___/
    //
    //

    private Parseable parseableObject = null;           // Instància de la classe que implementa el métode "onParse()".
    private ParamsList path = new ParamsList ('.');     // Guarda el path del últim tag.
    private XmlNode rootNode = null;                    // Node "root"
    private XmlNode currentNode = null;                 // Últim node creat.
    private String topLabel = null;                     // Primera etiqueta del document Xml.

    private int line = 0;
    private int col = 0;
    private boolean complited;
    private String searchPath;

    private String xmlVersion;
    private String xmlEncoding;
    private boolean caseSensitive;
    private boolean createXmlTree;


    private boolean closeTag = false;
    private boolean autoCloseTag = false;
    private boolean added = false;
    private boolean onLabel = false;    // Indica que està dintre dels corxets <...>
    private boolean onOrder = false;
    private boolean onAttribute = false;
    private boolean onValue = false;
    private boolean onLiteral = false;
    private boolean onComment = false;
    private boolean onDocType = false;
    private boolean onXml = false;

    private boolean foundEqual = false;

    private char caract = '\0';
    private char nextCaract = '\0';
    private char literalChar = '\0';

    private StringBuffer order = null;
    private StringBuffer param = null;
    private StringBuffer value = null;
    private StringBuffer literal = null;


    //   __  __      _            _                                  _                    _
    //  |  \/  |    | |          | |                                | |                  | |
    //  | \  / | ___| |_ ___   __| | ___ ___     ____ ___  _ __  ___| |_ _ __ _   _  ____| |_ ___  _ __ ___
    //  | |\/| |/ _ \ __) _ \ / _` |/ _ \ __)   / ___) _ \| `_  \ __) __) `__) | | |/ ___) __) _ \| `__) __)
    //  | |  | |  __/ |_ (_) | (_| |  __/__ \  | (___ (_) | | | |__ \ |_| |  | |_| | (___| |_ (_) | |  \__ \
    //  |_|  |_|\___|\__)___/ \__,_|\___|___/   \____)___/|_| |_|___/\__)_|   \__,_|\____)\__)___/|_|  (___/
    //
    //

    /**
     * Métode constructor
     *
     * @param parseableObject Instància de la classe que implementa el métode "<b>onParse</b>".
     * Si no es vol interceptar, es pot assignar un valor "<b>null</b>".
     *
     * <p>Per defecte assigna "<B>true</B>" als atributs "<b>caseSensitive</b>" i
     * "<b>saveNodesList</b>",
     */
    public XmlParserEx (Parseable parseableObject) {
        this.parseableObject = parseableObject;
        this.caseSensitive = true;
        if (parseableObject==null) { this.createXmlTree = true; }
        else { this.createXmlTree = false; }
        this.xmlVersion = "1.0";
        this.xmlEncoding = "utf-8";
    }


    //   __  __      _            _                        _             _
    //  |  \/  |    | |          | |                      (_)           | |
    //  | \  / | ___| |_ ___   __| | ___ ___    _ __  _ __ _ _    _  _ _| |_ ___
    //  | |\/| |/ _ \ __) _ \ / _` |/ _ \ __)  | `_ \| `__) | |  | | _` | __) __)
    //  | |  | |  __/ |_ (_) | (_| |  __/__ \  | |_) | |  | |\ \/ / (_| | |_\__ \
    //  |_|  |_|\___|\__)___/ \__,_|\___|___/  |  __/|_|  |_| \__/ \__,_|\__)___/
    //                                         | |
    //                                         |_|


    private String processEncodingLine (String text) {

        String res = null;

        if (text!=null) {
            text = text.trim();
            if (text.length()>0) {
                if (StringUtils.copy (text, 5).equals("<?xml") && text.charAt(text.length()-1)=='>') {
                    res = StringUtils.copy (text, "encoding=\"", "\"", StringUtils.CASE_INSENSITIVE, false);
                }
            }
        }

        return res;
    }

    private String processEncodingLine (byte[] taula) {

        StringBuilder text = new StringBuilder (taula.length);
        int cont1 = 0;
        boolean eol = false;
        char caract = '\0';
        while (cont1<taula.length && eol==false) {
            caract = (char) taula[cont1];
            if (caract==CR || caract==LF) { eol = true; }
            else { text.append(caract); }
            cont1++;
        }

        return processEncodingLine (text.toString());
    }

    private Reader getReader (InputStream in, ErrorEx error) {

        StringBuilder text = new StringBuilder ();
        Reader res = null;
        if (in!=null) {

            String encoding = null;
            try {

                if (!(in instanceof BufferedInputStream)) { in = new BufferedInputStream (in); }

                in.mark (HEADER_SIZE);
                byte[] header = new byte[HEADER_SIZE];
                in.read (header, 0, HEADER_SIZE);
                in.reset();
                encoding = processEncodingLine (header);
                if (StringUtils.isEmpty (encoding)) { encoding = DEFAULT_ENCODING; }
                res = new InputStreamReader (in, encoding);
            }
            catch (UnsupportedEncodingException ex) { error.setError ("IOEXCEPTION", "ACTION:SET_XML_DOCUMENT", "TRACE:XmlParserEx.setXmlDocument", "DESCRIPTION: No se ha podido identificar la codificación", "CODIFICACION:"+encoding, "MESSAGE:"+ex.getMessage());  }
            catch (IOException ex) { error.setError ("IOEXCEPTION", "ACTION:SET_XML_DOCUMENT", "TRACE:XmlParserEx.setXmlDocument", "DESCRIPTION: No se ha podido leer el fichero XML", "MESSAGE:"+ex.getMessage());  }
        }
        return res;
    }

    // Processa una etiqueta
    private void addLabel (String value, int type, boolean closeTag, boolean acceptEmptyValues, ErrorEx error) {

        value = StringUtils.blankAdv (value);

        if (value.length()>0 || acceptEmptyValues) {

            boolean createNode = true;
            String lastLabel = path.getParam (path.paramCount()-1);

            // Comproba si l'etiqueta de tancament correspont amb la etiqueta d'inici
            // En cas afirmatiu, treu el últim node del path.
            if (type==XmlNode.LABEL_NODE || type==XmlNode.ATTRIBUTE_NODE || type==XmlNode.ROOT_NODE) {

                // ATENCIÓ: S'utilitza el nom ROOT_NAME_NODE enlloc del tipus del node pare, perque
                // si no es crea el arbre no disposem d'aquesta informació.
                if (!closeTag && path.toString().equalsIgnoreCase(ROOT_NODE_NAME)) {
                    if (topLabel!=null) { error.setError (XML_ERROR, "ACTION:ADD_LABEL", "TRACE:XmlParserEx.addLabel", "DESCRIPTION: Documento Xml erroneo", "SINTAX_ERROR: SÓLO SE ADMITE UN ELEMENTO DE NIVEL SUPERIOR [LINE: "+line+" COL :"+col+"]"); }
                    topLabel = value;
                }

                if (closeTag) {
                    
                    if (!StringUtils.equals (lastLabel, value, caseSensitive)) {
                        error.setError (XML_ERROR, "ACTION:ADD_LABEL", "TRACE:XmlParserEx.addLabel", "DESCRIPTION: Documento Xml erroneo", "SINTAX_ERROR: LA ETIQUETA DE FIN '"+value+"' NO COINCIDE CON LA ETIQUETA DE INICIO '"+lastLabel+"' [LINE: "+line+" COL :"+col+"]");
                    }
                    createNode = false;
                    value = "/"+value;
                    path.removeParam (path.paramCount()-1);
                }
            }

            if (parseableObject!=null) { complited = parseableObject.onParse (path.toString(), value, type); }

            // Crea el node i l'inserta a l'arbre.
            if (createNode && createXmlTree) {                
                XmlNode newNode = new XmlNode (value, type);
                newNode.setParent (currentNode);
                if (rootNode==null) {rootNode = newNode; }
                currentNode = newNode;
            }

            // Si el node és una etiqueta, un atribut, o un element root, l'afegueix al path.
            if (type==XmlNode.LABEL_NODE || type==XmlNode.ATTRIBUTE_NODE || type==XmlNode.ROOT_NODE) {
                if (!closeTag) { path.addParam (value); }
            }

            if (closeTag && currentNode!=null) { currentNode = currentNode.parent; }
        }
    }

    private void addLabel (String value, int type, boolean closeTag, ErrorEx error) { addLabel (value, type, closeTag, false, error); }

    private void startParseXml (ErrorEx error) {

        topLabel = null;
        rootNode = null;
        currentNode = null;
        addLabel (ROOT_NODE_NAME, XmlNode.ROOT_NODE, false, error);

        closeTag = false;
        autoCloseTag = false;
        added = false;
        onLabel = false;    // Indica que està dintre dels corxets <...>
        onOrder = false;
        onAttribute = false;
        onValue = false;
        onLiteral = false;
        onComment = false;
        onDocType = false;
        onXml = false;

        foundEqual = false;

        caract = '\0';
        nextCaract = '\0';
        literalChar = '\0';

        order = new StringBuffer ();
        param = new StringBuffer ();
        value = new StringBuffer ();
        literal = new StringBuffer ();

    }

    private void endParseXml (ErrorEx error) {

        if (error.getError()==null) { addLabel (ROOT_NODE_NAME, XmlNode.ROOT_NODE, true, error); }

        String lastTag = path.toString();
        if (error.getError()==null && lastTag.length()>0) { error.setError (XML_ERROR, "ACTION:PARSE_XML", "TRACE:XmlParserEx.parseXml", "DESCRIPTION: Documento Xml erroneo", "SINTAX_ERROR: LA ETIQUETA \""+lastTag+"\" NO SE HA CERRADO [LINE: "+line+" COL :"+col+"]"); }

        currentNode = null;
    }

    // Interpreta el document Xml
    private void parseXml (String text, ErrorEx error) {


        line = 0;
        col = 0;
        complited = false;
        int cont1 = 0;

        while (cont1<text.length() & error.getError()==null & !complited) {

            caract = text.charAt(cont1);
            if (cont1+1<text.length()) { nextCaract = text.charAt(cont1+1); }
            else { nextCaract = '\0'; }

            if (caract==CR) { line++; col = 0; }

            if (!onLiteral && !onComment && !onDocType) {
                if (caract<' ') { caract = ' '; }

                if (onLabel) {

                    switch (caract) {
                        case 0    : break;
                        case '='  : if (onValue || onAttribute) {
                                        onValue = true;
                                        onAttribute = false;
                                        foundEqual = true;
                                    }
                                    else { error.setError (XML_ERROR, "ACTION:PARSE_XML", "TRACE:XmlParserEx.parseXml", "DESCRIPTION: Documento Xml erroneo", "SINTAX_ERROR: UN NOMBRE EMPIEZA CON UN CARACTER NO VALIDO \""+caract+"\" [LINE: "+line+" COL :"+col+"]"); }
                                    break;

                        case '"'  :
                        case '\'' : if (foundEqual) {
                                        onLiteral = true;
                                        literalChar = caract;
                                    }
                                    else { error.setError (XML_ERROR, "ACTION:PARSE_XML", "TRACE:XmlParserEx.parseXml", "DESCRIPTION: Documento Xml erroneo", "SINTAX_ERROR: FALTA EL SIGNO IGUAL ENTRE EL ATRIBUTO Y EL VALOR [LINE: "+line+" COL :"+col+"]"); }
                                    break;

                        case '<'  : error.setError (XML_ERROR, "ACTION:PARSE_XML", "TRACE:XmlParserEx.parseXml", "DESCRIPTION: Documento Xml erroneo", "SINTAX_ERROR: UN NOMBRE EMPIEZA CON UN CARACTER NO VALIDO \""+caract+"\" [LINE: "+line+" COL :"+col+"]");
                                    break;

                        case '>'  : if (order.length()==0 || onValue) { error.setError (XML_ERROR, "ACTION:PARSE_XML", "TRACE:XmlParserEx.parseXml", "DESCRIPTION: Documento Xml erroneo", "SINTAX_ERROR: UN NOMBRE EMPIEZA CON UN CARACTER NO VALIDO \""+caract+"\" [LINE: "+line+" COL :"+col+"]"); }
                                    if (onValue) { error.setError (XML_ERROR, "ACTION:PARSE_XML", "TRACE:XmlParserEx.parseXml", "DESCRIPTION: Documento Xml erroneo", "SINTAX_ERROR: FALTA EL SIGNO IGUAL ENTRE EL ATRIBUTO Y EL VALOR [LINE: "+line+" COL :"+col+"]"); }
                                    if (onAttribute && param.length()>0) { error.setError (XML_ERROR, "ACTION:PARSE_XML", "TRACE:XmlParserEx.parseXml", "DESCRIPTION: Documento Xml erroneo", "SINTAX_ERROR: FALTA EL SIGNO IGUAL ENTRE EL ATRIBUTO Y EL VALOR [LINE: "+line+" COL :"+col+"]"); }

                                    // Si és una ordre <?xml..> no s'envia a la llista de tags.
                                    if (!onXml) {
                                        if (autoCloseTag) {
                                            if (!added) { addLabel (order.toString(), XmlNode.LABEL_NODE, false, error); }
                                            addLabel (order.toString(), XmlNode.LABEL_NODE, true, error);
                                        }
                                        else {
                                            if (!added) { addLabel (order.toString(), XmlNode.LABEL_NODE, closeTag, error); }
                                        }
                                    }

                                    onLabel = false;
                                    break;

                        case ' '  : if (onOrder) {
                                        if (order.length()>0) {
                                            onOrder = false;
                                            onAttribute = true;
                                            foundEqual = false;
                                            param.setLength (0);

                                            // Un cop entrada la ordre, valildo si és alguna de les exceptions de la sintaxis Xml
                                            // !-- Per a comentaris
                                            // ?xml Per a definir les característiques del document Xml
                                            // !DOCTYPE Per a identificar certs elements o les definicions DTD
                                            if (order.charAt(0)=='?') {
                                                if (order.toString().equalsIgnoreCase("?xml")) { onXml = true; }
                                                else { error.setError (XML_ERROR, "ACTION:PARSE_XML", "TRACE:XmlParserEx.parseXml", "DESCRIPTION: Documento Xml erroneo", "SINTAX_ERROR: UN NOMBRE CONTENÍA UN CARACTER NO VÁLIDO [LINE: "+line+" COL :"+col+"]"); }
                                            }

                                            if (order.charAt(0)=='!') {
                                                if (StringUtils.copy (order.toString(), 0, 3).equalsIgnoreCase("!--")) {
                                                    onComment = true;
                                                    literalChar = 0xffff;
                                                }
                                                if (order.toString().equalsIgnoreCase("!DOCTYPE")) { 
                                                    onDocType = true;
                                                    literalChar = 0xffff;
                                                }

                                                if (onDocType==false && onComment==false) { error.setError (XML_ERROR, "ACTION:PARSE_XML", "TRACE:XmlParserEx.parseXml", "DESCRIPTION: Documento Xml erroneo", "SINTAX_ERROR: UN NOMBRE CONTENÍA UN CARACTER NO VÁLIDO [LINE: "+line+" COL :"+col+"]"); }
                                            }
                                        }
                                        else { error.setError (XML_ERROR, "ACTION:PARSE_XML", "TRACE:XmlParserEx.parseXml", "DESCRIPTION: Documento Xml erroneo", "SINTAX_ERROR: NO SE ADMITEN ESPACIOS EN BLANCO EN ESTA UBICACIÓN [LINE: "+line+" COL :"+col+"]"); }
                                    }
                                    if (onAttribute & param.length()>0) {
                                        onAttribute = false;
                                        onValue = true;
                                    }
                                    break;
                        case '/'  : if (order.length()!=0 && nextCaract!='>') { error.setError (XML_ERROR, "ACTION:PARSE_XML", "TRACE:XmlParserEx.parseXml", "DESCRIPTION: Documento Xml erroneo", "SINTAX_ERROR: SE ESPERABA EL CARACTER \">\" [LINE: "+line+" COL :"+col+"]"); }

                                    closeTag = true;
                                    if (nextCaract=='>') { autoCloseTag = true; }
                                    break;

                        case '!'  : if (order.length()!=0) { error.setError (XML_ERROR, "ACTION:PARSE_XML", "TRACE:XmlParserEx.parseXml", "DESCRIPTION: Documento Xml erroneo", "SINTAX_ERROR: UN NOMBRE EMPIEZA CON UN CARACTER NO VALIDO \""+caract+"\" [LINE: "+line+" COL :"+col+"]"); }
                                    autoCloseTag = true;
                                    order.append (caract);
                                    break;

                        case '?'  : if (order.length()!=0 && nextCaract!='>') { error.setError (XML_ERROR, "ACTION:PARSE_XML", "TRACE:XmlParserEx.parseXml", "DESCRIPTION: Documento Xml erroneo", "SINTAX_ERROR: SE ESPERABA EL CARACTER \">\" [LINE: "+line+" COL :"+col+"]"); }
                                    autoCloseTag = true;
                                    if (order.length()==0) { order.append (caract); }
                                    break;

                        default   : if (onValue) { 
                                        error.setError (XML_ERROR, "ACTION:PARSE_XML", "TRACE:XmlParserEx.parseXml", "DESCRIPTION: Documento Xml erroneo", "SINTAX_ERROR: SE ESPERABA UN LITERAL PERO NO SE ENCONTRARON LAS COMILLAS DE APERTURA [LINE: "+line+" COL :"+col+"]");
                                    }
                                    if (onOrder) { order.append (caract); }
                                    if (onAttribute) {
                                        param.append (caract);
                                        if (closeTag) { error.setError (XML_ERROR, "ACTION:PARSE_XML", "TRACE:XmlParserEx.parseXml", "DESCRIPTION: Documento Xml erroneo", "SINTAX_ERROR: LAS ETIQUETAS DE CIERRE NO PUEDEN TENER ATRIBUTOS [LINE: "+line+" COL :"+col+"]"); }
                                    }
                                    break;
                    }
                }
                else {
                    switch (caract) {
                        case 0    : break;
                        case '<'  : if (value.length()>0) { addLabel (value.toString(), XmlNode.DATA_NODE, true, error); }
                        
                                    onLabel = true;
                                    onOrder = true;
                                    onAttribute = false;
                                    onValue = false;
                                    closeTag = false;
                                    onComment = false;
                                    onXml = false;
                                    autoCloseTag = false;
                                    added = false;
                                    order.setLength (0);
                                    param.setLength (0);
                                    value.setLength (0);
                                    break;

                        default   : value.append (caract);
                    }
                }
            }
            else {
                if (onComment) {
                    if (StringUtils.copy (text, cont1-2, 3).equals ("-->")) { caract = literalChar; }
                }
                if (onDocType) {
                    if (caract=='>') { caract = literalChar; }
                }

                if (caract==literalChar) {
                    onLiteral = false;

                    value.append (literal);

                    if (onValue) {
                        
                        onValue = false;
                        onAttribute = true;

                        // Si NO és la ordre "?Xml" cal enviar el node de l ordre (si no s'ha enviat anteriorment",
                        // el d'inici del atribut, el valor i el final del atribut.
                        // Si és la ordre "?Xml", cal processar els atributs "version" i "encoding" i afeguir el node
                        if (!onXml) {
                            if (!added) {
                                addLabel (order.toString(), XmlNode.LABEL_NODE, false, error);
                                added = true;
                            }
                            addLabel (param.toString(), XmlNode.ATTRIBUTE_NODE, false, error);
                            addLabel (value.toString(), XmlNode.DATA_NODE, true, true, error);
                            addLabel (param.toString(), XmlNode.ATTRIBUTE_NODE, true, error);
                        }
                        else {
                            if (param.toString().equalsIgnoreCase("version")) { xmlVersion = value.toString(); }
                            if (param.toString().equalsIgnoreCase("encoding")) { xmlEncoding = value.toString(); }
                        }
                        param.setLength (0);
                        value.setLength (0);
                    }
                    else { value.append (caract); }

                    if (onComment) {
                        String comment = value.toString();
                        comment = StringUtils.copy (comment, 0, value.length()-3);
                        addLabel (comment , XmlNode.COMMENT_NODE, true, error);

                        onLabel = false;
                        onComment = false;
                        param.setLength (0);
                        value.setLength (0);
                    }

                    if (onDocType) {
                        String docType = value.toString();
                        addLabel (docType , XmlNode.DOCTYPE_NODE, true, error);

                        onLabel = false;
                        onDocType = false;
                        param.setLength (0);
                        value.setLength (0);
                    }

                    literal.setLength (0);

                }
                else { literal.append (caract); }
            }
            cont1++;
            col++;
        }
    }

    // Genera el document Xml
    private String generateXml (XmlNode node, boolean useTabs) {

        String res = "";
        if (node==rootNode) { res = res + "<?xml version=\""+xmlVersion+"\" encoding=\""+xmlEncoding+"\" ?>"; }

        boolean descent = false;
        boolean closed = true;
        String tab = "";
        int deep = node.getDeep();
        node = node.getNextNode();

        while (node!=null && node.getDeep()>deep) {           
            
            if (useTabs) { 
                int tabSpaces = (StringUtils.countChar (node.getPath(), '.')-1)*2;
                if (node.getPath().length()>0) { tabSpaces+=2; }
                tab = StringUtils.fillLeftSpaces("", tabSpaces);

            }
            int type = node.getType();

            if (closed==false) {
                XmlNode parent = node.parent;
                if (parent!=null && parent.getType()!=XmlNode.ATTRIBUTE_NODE && type!=XmlNode.ATTRIBUTE_NODE) {
                    res = res + ">";
                    closed = true;
                }
            }

            switch (type) {
                case XmlNode.COMMENT_NODE :     res = res + CR+LF+tab+"<!-- "+node.getValue()+" -->";
                                                closed = true;
                                                break;
                case XmlNode.LABEL_NODE :       if (!descent) {
                                                    if (node.firstChild==null) {
                                                        res = res + CR+LF+tab+"<"+node.getValue()+"/>";
                                                        closed = true;
                                                    }
                                                    else {
                                                        res = res + CR+LF+tab+"<"+node.getValue();
                                                        closed = false;
                                                    }
                                                }
                                                else {
                                                    res = res + CR+LF+tab+"</"+node.getValue()+">";
                                                    closed = true;
                                                }
                                                break;
                case XmlNode.ATTRIBUTE_NODE :   if (!descent) { res = res + " "+node.getValue()+"="; }
                                                break;
                case XmlNode.DATA_NODE :        if (node.getParent()!=null && node.getParent().getType()==XmlNode.ATTRIBUTE_NODE) { res = res + "\""+node.getValue()+"\""; }
                                                else { res = res + CR+LF+tab+node.getValue(); }
                                                break;
            }

            // Busca el següent node.
            if (!descent && node.firstChild!=null) { node = node.firstChild; }
            else if (node.nextBrother!=null) {
                node = node.nextBrother;
                descent = false;
            }
            else {
                node = node.parent;
                descent = true;                
            }
        }
        return res;
    }


    //   _____       _                           _              _          _                                        _
    //  (_   _)     | |                         | |            (_)        / /                                      (_)
    //    | |  _ __ | |_ ___ _ __ _ __  _ __ ___| |_  _ _  ____ _  ___   / /   __ _  ___ _ __   ___ _ __  _ _  ____ _  ___
    //    | | | `_  \ __) _ \ `__) `_ \| `__) _ \ __) _` |/ ___) |/ _ \ / /   / _` |/ _ \ `_  \/ _ \ `__) _` |/ ___) |/ _ \
    //   _| |_| | | | |_  __/ |  | |_) | | |  __/ |_ (_| | (___| | (_) | /   | (_| |  __/ | | |  __/ | | (_| | (___| | (_) |
    //  (_____)_| |_|\__)___|_|  |  __/|_|  \___|\__)__,_|\____)_|\___/_/     \__, |\___|_| |_|\___|_|  \__,_|\____)_|\___/
    //                           | |                                           __| |
    //                           |_|                                          (___/
    //       _                                      _          _  _ __    ____  __ _ _  _
    //      | |                                    | |        | || |\ \  / /  \/  | | || |
    //    __| | ___   ____ _   _ _ _ __   ___ _ __ | |_ ___   |_||_| \ \/ /| \  / | |_||_|
    //   / _` |/ _ \ / ___) | | | ` `  \ / _ \ `_  \ __) __)          )  ( | |\/| | |
    //  | (_| | (_) | (___| |_| | |\ |\ |  __/ | | | |_\__ \         / /\ \| |  | | |___
    //   \__,_|\___/ \____)\__,_|_||_||_|\___|_| |_|\__)___/        /_/  \_\_|  |_|_____)
    //
    //

    /**
     * Interpreta el document XML.
     *
     * <p><b>NOTA</b>: Si està activat el atributs "<b>saveNodesList</b>" per
     * cada node afeguirà una entrada a la llista de nodes que en acabat podem
     * recuperar amb el métode "<b>getNodes</b>".
     *
     * <p><b>NOTA:</b>: Si al instància el objecte amb el métode constructor s'ha
     * indicat una classe "<b>Parseable</b>", per cada node genera un event al
     * métode "<b>onParse()</b> de l'esmentada classe.
     *
     * <p><b>NOTA</b>: Si durant el procés d'interpretació es produeix algun error,
     * o es detecta algun defecte en el document, es crea una instpancia
     * de la classe "<b>ErrorEx</b>" amb la descripció del mateix.
     */
    public synchronized void parse (String xml, ErrorEx error) {

        if (xml!=null) {
            startParseXml (error);
            parseXml (xml, error);
            endParseXml (error);
        }
    }

    /**
     * Permet indicar el document que es vol interpretar.
     *
     * <p>b>NOTA</b>: Si el document està enmagatzemat en un arxiu, cal utilitzar
     * el métode "<b>loadFromFile</b>">
     *
     * @param xmlDocument Nom del document a interpretar.
     */
    public synchronized void parse (InputStream in, ErrorEx error) {

        if (in!=null) {

            Reader reader = getReader (in, error);
            if (error.getError()==null) {

                startParseXml (error);
                StringBuilder text = new StringBuilder ();
                char caract = '\0';

                try {
                    int value = reader.read();
                    while (value!=-1) {

                        caract = (char) value;
                        if (caract==LF) {
                            parseXml (text.toString(), error);
                            text.setLength(0);
                        }
                        else text.append(caract);
                        value = reader.read();
                    }
                    if (text.length()>0) { parseXml (text.toString(), error); }
                }
                catch (IOException ex) { error.setError ("IOEXCEPTION", "ACTION:SET_XML_DOCUMENT", "TRACE:XmlParserEx.setXmlDocument", "DESCRIPTION: No se ha podido leer el fichero XML", "MESSAGE:"+ex.getMessage());  }

                endParseXml (error);
            }
        }
    }

    /**
     * Carrega el document indicat dintre del atribut "<b>xmlDocument</b>.
     *
     * <p><b>NOTA</b>: Si durant el procés es produeix algun error, es crea una instpancia
     * de la classe "<b>ErrorEx</b>" amb la descripció del mateix.
     *
     * @param fileName Nom del fitxer a carregar.
     */
    public synchronized void parseFromFile (String fileName, ErrorEx error) {

        DataFile file = new DataFile();
        file.reset (fileName, error);
        if (error.getError()==null) {
            parse (file.readAsStream(), error);
            file.close (error);
        }
    }

    /**
     * Genera a partir de la llista de nodes, el document Xml i l'enmagatzema en el
     * atribut "<b>xmlDocument</b>" que es pot obtenir mitjançant el métode
     * "<b>getXmlDocument</b>".
     */
    public synchronized String generate (boolean useTabs) { return generateXml (rootNode, useTabs); }
    public synchronized String generate () { return generate (false); }

    public synchronized String generate (XmlNode node, boolean useTabs) { return generateXml (node, useTabs); }
    public synchronized String generate (XmlNode node) { return generate (node, false); }

    /**
     * Grava en un arxiu el document Xml enmagatzemat dintre del atribut "<b>xmlDocument</b>.
     *
     * <p><b>NOTA</b>: Si durant el procés es produeix algun error, es crea una instància
     * de la classe "<b>ErrorEx</b>" amb la descripció del mateix.
     *
     * @param fileName Nom del fitxer a gravar.
     */
    public synchronized void saveToFile (String fileName, ErrorEx error) {

        DataFile file = new DataFile();
        file.rewrite (fileName, error);
        if (error.getError()==null) {
            file.writeLine (generate(), error);
            file.close(error);
        }
    }

    //    _____                           _                         _
    //   / ____)                         | |                       | |
    //  | |     ___ _ __ ____  _ _     __| | ___    _ __   ___   __| | ___ ___
    //  | |    / _ \ `__) ___) _` |   / _` |/ _ \  | `_  \/ _ \ / _` |/ _ \ __)
    //  | |____  __/ | | (___ (_| |  | (_| |  __/  | | | | (_) | (_| |  __/__ \
    //   \_____)___|_|  \____)__,_|   \__,_|\___|  |_| |_|\___/ \__,_|\___|___/
    //
    //

    public XmlNode getRootNode () {
        XmlNode node = rootNode;
        currentNode = node;
        return node;
    }

    public XmlNode getPreviousNode () {
        if (currentNode!=null) { currentNode = currentNode.getPreviousNode(); }
        return currentNode;
    }

    public XmlNode getNextNode () {
        if (currentNode!=null) { currentNode = currentNode.getNextNode(); }
        return currentNode;
    }

    public XmlNode searchFirstNode (String path) {
        currentNode = rootNode;
        
        int pos = StringUtils.charPos (path, '.');
        if (pos==-1) { pos = path.length(); }
        String rootLabel = (StringUtils.copy (path, 0, pos));
        if (!rootLabel.equals(ROOT_NODE_NAME)) { path = StringUtils.concat (ROOT_NODE_NAME, ".", path); }
       
        this.searchPath = StringUtils.lowerCase (path, caseSensitive);
        return searchNextNode ();
    }

    public XmlNode searchNextNode () {

        XmlNode node = currentNode;
        if (node!=null) { node = node.getNextNode(); }
        XmlNode trobat = null;
        
        while (trobat==null && node!=null) {
            //if (searchPath.equals (StringUtils.lowerCase (node.getPath(), caseSensitive))) { trobat = node; }
            if (searchPath.equals (StringUtils.lowerCase (node.getPath()+"."+node.getValue(), caseSensitive))) { trobat = node; }
            node = node.getNextNode();
        }
        currentNode = trobat;
        return currentNode;
    }

    /**
     * Obté el node amb la ruta indicada, i en retorna el valor si existeix o
     * <b>null</b> en cas contrari.
     */
    public String getNodeValue (String path) {

        String res = null;
        XmlNode node = searchFirstNode (path);
        if (node!=null) { res = node.getFirstChildValue(); }

        return res;
    }

    /**
     * Obté el valor del primer fill del node indicat, o <b>null</b> en cas de que
     * el node indicat sigui <b>null</b> o no tingui cap fill.
     */
    public String getFirstChildValue (XmlNode node) {

        String res = null;
        if (node!=null) { res = node.getFirstChildValue(); }

        return res;
    }

    //   __  __      _            _             _  _   _____ ______ _______ _  _
    //  |  \/  |    | |          | |           | || | / ____)  ____)__   __) || |
    //  | \  / | ___| |_ ___   __| | ___ ___   |_||_|| (___ | |__     | |  |_||_|
    //  | |\/| |/ _ \ __) _ \ / _` |/ _ \ __)         \___ \|  __)    | |
    //  | |  | |  __/ |_ (_) | (_| |  __/__ \         ____) | |____   | |
    //  |_|  |_|\___|\__)___/ \__,_|\___|___/        (_____/|______)  |_|
    //
    //

    /**
     * Indica si el interpret XML és sensible a la tipografia o no.
     * @param caseSensitive <b>true</b> indica que el interpret XML és sensible a la
     * tipografia.
     */
    public void setCaseSensitive (boolean caseSensitive) { this.caseSensitive = caseSensitive; }

    /**
     * Permet indicar si es vol enmagatzemar la llista de nodes en un "<b>Vector</b>" o no.
     *
     * @param saveNodesList <b>true</b> si es desitja enmagatzemar la llista de nodes en
     * un "<b>Vector</b>".
     *
     * <p><b>NOTA:</b> Si no s'ha indicat classe que implementi la classe "Parseable",
     * com que ja no es pot interceptar el métode "<b>onParse()</b>", aquest atribut
     * no es pot desabilitar.
     */
    public void setCreateXmlTree (boolean createXmlTree) {
        if (parseableObject==null) { createXmlTree = true; }
        this.createXmlTree = createXmlTree;
    }

    /**
     * Permet indicar la versió del document Xml.
     *
     * <p><b>NOTA</b>: Aquest atribut només té sentit modificar-lo quen es té intenció
     * de generar un document Xml a partir dels nodes enmagatzemats.
     *
     * @param xmlVersion Versió del document Xml.
     */
    public void setXmlVersion (String xmlVersion) { this.xmlVersion = StringUtils.notNull (xmlVersion); }

    /**
     * Permet indicar la caodificació del document Xml.
     *
     * <p><b>NOTA</b>: Aquest atribut només té sentit modificar-lo quen es té intenció
     * de generar un document Xml a partir dels nodes enmagatzemats.
     *
     * @param xmlEncoding Codificació del document Xml.
     */
    public void setXmlEncoding (String xmlEncoding) { this.xmlEncoding = StringUtils.notNull (xmlEncoding); }


    //   __  __      _            _             _  _   _____ ______ _______ _  _
    //  |  \/  |    | |          | |           | || | / ____)  ____)__   __) || |
    //  | \  / | ___| |_ ___   __| | ___ ___   |_||_|| |  __| |__     | |  |_||_|
    //  | |\/| |/ _ \ __) _ \ / _` |/ _ \ __)        | | (_ |  __)    | |
    //  | |  | |  __/ |_ (_) | (_| |  __/__ \        | |__| | |____   | |
    //  |_|  |_|\___|\__)___/ \__,_|\___|___/         \_____|______)  |_|
    //
    //

    /**
     * Obté el estat del atribut "b>caseSensitive</b>".
     *
     * @return <b>true</b> si el interpret és sensible a la tipografia; <b>false</b>
     * en cas contrari.
     */
    public boolean getCaseSensitive () { return caseSensitive; }

    /**
     * Obté el atribut "<b>saveNodesList</b>".
     *
     * @return <b>true</b> si s'ha d'enmagatzemar la llista de nodes; <b>false</b>
     * en cas contrari.
     */
    public boolean getCreateXmlTree () { return createXmlTree; }

    /**
     * Obté la versió del document Xml.
     *
     * @return Retorna la versió del document Xml.
     */
    public String getXmlVersion () { return xmlVersion; }

    /**
     * Obté la codificació del document Xml.
     *
     * @return Retorna la codificació del document Xml.
     */
    public String getXmlEncoding () { return xmlEncoding; }



}
