//    _____ _                       _  _ __    __        _ _              _     _     _  _  _
//   / ____) |                     | || |\ \  / /       | | |            | |   (_)   | || || |
//  | |    | |  _ _ ___ ___  ___   |_||_| \ \/ / _ _ __ | | |_  _ _  __ _| |    _ ___| |__||_|
//  | |    | |/ _` | __) __)/ _ \          )  ( | ` `  \| | __) _` |/ _` | |   | | __) __)
//  | |____| | (_| |__ \__ \  __/         / /\ \| |\ |\ | | |_ (_| | (_| | |___| |__ \ |_
//   \_____)_|\__,_|___/___/\___|        /_/  \_\_||_||_|_|\__)__,_|\__, |_____)_|___/\__)
//                                                                   __| |
//                                                                  (___/


package cat.lib.xml;

import cat.lib.properties.PropertiesEx;
import cat.lib.utils.NumberToString;

/**
 *
 * La classe "XmlTagList" és una eina dissenyada per a facilitar la creació i interpretació de
 * documents Xml.
 *
 * L'objectiu principal intenta evitar que el codi del aplicatiu contingui les cadenes amb els noms
 * dels tags com a cadenes, ja que si es vol canviar una etiqueta xml, caldria canviar-la a tants
 * llocs com aparegui.
 *
 * Per a fer-ho. la classe encapsula un "PropertiesEx" el qual interpreta un fitxer de "properties"
 * i crea un <b>HashMap</b> en el que es vincula una clau alfanumérica amb el nom del tag.
 *
 * Per a definir la llista de tags el més habitual és instanciar el objecte indicant-li el "propertiesEx"
 * que conté la llista de tags. Tanmateix, si no es desitja utilitzar aquest recurs, es pot instanciar
 * el objecte i indicar posteriorment la classe "propertiesEx" a utilitzar mitjançant el métode
 * "<b>setTagList</b>".
 *
 * També, es poden entrar individualment els tags (amb el métode "<b>addTag</b>"), però tenint
 * en compte que la nova entrada s'enmagatzemará dintre del "propertiesEx" previament indicat.
 *
 * La classe aporta diversos métodes per a recuperar el nom del tag o com a cadena Xml incloent els
 * caràcters < / >. Si s'intenta obtenir un tag que no existeix mostra un missatge d'error a la consola.
 *
 * NOTA IMPORTANT: Cal tenir en compte que el HashMap (i les entrades del fitxer properties
 * s'enmagatzemen en un HashMap) NO PERMET CREAR DUES ENTRADES AMB LA MATEIXA CLAU, per tant
 * la última entrada sempre sobreescriu el valor anterior.
 *
 * NOTA IMPORTANT: També cal tenir en compte que aquesta classe cerca les entrades del HashMap
 * així com els valors continguts sense tenir en compte la tipografia. Això pot plantejar un
 * problema en cas de que el fitxer properties CONTINGUI DUES ENTRADES AMB LA MATEIXA CLAU PERÒ DIFERENT
 * TIPOGRAFIA, ja que una d'elles serà inaccessible.
 *
 * NOTRA IMPORTANT: Els noms dels tags, també es tracten sense tenir en compte la tipografia,
 * i si s'intenta afeguir el mateix tagname a més d'una clau, ens mostra un missatge d'error per
 * la consola.
 *
 *
 * @author Miquel Abadal i Cervià
 * @date 25-06-2010
 */
public class XmlTagList {

    //     __   _        _ _           _
    //   / __ \| |      (_) |         | |
    //  | |__| | |_ _ __ _| |_   _   _| |_ ___
    //  |  __  | __) `__) | `_ \| | | | __) __)
    //  | |  | | |_| |  | | |_) | |_| | |_\__ \
    //  |_|  |_|\__)_|  |_|_,__/ \__,_|\__)___/
    //
    //

    private PropertiesEx xmlTags;


    //   __  __      _            _                              _                    _
    //  |  \/  |    | |          | |                            | |                  | |
    //  | \  / | ___| |_ ___   __| | ___     ____ ___  _ __  ___| |_ _ __ _   _  ____| |_ ___  _ __
    //  | |\/| |/ _ \ __) _ \ / _` |/ _ \   / ___) _ \| `_  \ __) __) `__) | | |/ ___) __) _ \| `__)
    //  | |  | |  __/ |_ (_) | (_| |  __/  | (___ (_) | | | |__ \ |_| |  | |_| | (___| |_ (_) | |
    //  |_|  |_|\___|\__)___/ \__,_|\___|   \____)___/|_| |_|___/\__)_|   \__,_|\____)\__)___/|_|
    //
    //

    public XmlTagList () { setTagList (null); }

    public XmlTagList (PropertiesEx xmlTags) { setTagList (xmlTags); }

    //   __  __      _            _             _  _   _____ ______ _______ _  _
    //  |  \/  |    | |          | |           | || | / ____)  ____)__   __) || |
    //  | \  / | ___| |_ ___   __| | ___ ___   |_||_|| (___ | |__     | |  |_||_|
    //  | |\/| |/ _ \ __) _ \ / _` |/ _ \ __)         \___ \|  __)    | |
    //  | |  | |  __/ |_ (_) | (_| |  __/__ \         ____) | |____   | |
    //  |_|  |_|\___|\__)___/ \__,_|\___|___/        (_____/|______)  |_|
    //
    //

    /**
     * Permet indicar la llista de tags enmagatzemada en una instància
     * <b>PropertiesEx</b>.
     *
     * @param xmlTags
     */
    public void setTagList (PropertiesEx xmlTags) { this.xmlTags = xmlTags; }


    //   __  __      _            _             _  _   _____ ______ _______ _  _
    //  |  \/  |    | |          | |           | || | / ____)  ____)__   __) || |
    //  | \  / | ___| |_ ___   __| | ___ ___   |_||_|| |  __| |__     | |  |_||_|
    //  | |\/| |/ _ \ __) _ \ / _` |/ _ \ __)        | | (_ |  __)    | |
    //  | |  | |  __/ |_ (_) | (_| |  __/__ \        | |__| | |____   | |
    //  |_|  |_|\___|\__)___/ \__,_|\___|___/         \_____|______)  |_|
    //
    //

    /**
     * Permet recuperar la instància <b>PropertiesEx</b> que conté la llista de tags.
     * @return
     */
    public PropertiesEx getTagList () { return xmlTags; }

    /**
     * Retorna el tag enmagatzemat amb la clau <b>key</b>.
     *
     * Si la clau o el tag tenen el valor <b>null</b> retorna un <b>null</b> i mostra un missatge d'error
     * per la consola.
     *
     * @param key Clau del tag a obtenir
     * @return <b>tag</b>
     */
    public String getTag (String key) {

        String res = null;

        if (xmlTags!=null) {
            if (key!=null) {
                key = key.trim();
                res = xmlTags.getProperty (key);
            }
        }
        if (res==null) { System.out.println ("[XML_TAG_ERROR] Tag: \""+key+"\" Description: Tag not found. "); }

        return res;
    }

    /**
     * Retorna el tag enmagatzemat amb la clau <b>key</b>.
     *
     * @param key Clau del tag a obtenir
     * @return <b>tag</b>
     */
    public String getOpenTag (String key) { return getTag (key); }

    /**
     * Retorna el tag enmagatzemat amb la clau <b>key</b> afeguint-hi el caràcter "/" al inici.
     *
     * @param key Clau del tag a obtenir
     * @return <b>/tag</b>
     */
    public String getCloseTag (String key) { return "/"+getTag (key); }

    /**
     *
     * * Retorna el tag d'obertura enmagatzemat amb la clau <b>key</b> amb sintaxis XML.
     *
     * @param key Clau del tag a obtenir
     * @return <b>[tag]</b>
     */
    public String getXmlOpenTag (String key) { return "<"+getTag (key)+">"; }

    /**
     * Retorna el tag de tancament enmagatzemat amb la clau <b>key</b> amb sintaxis XML.
     *
     * @param key Clau del tag a obtenir
     * @return <b>[/tag]</b>
     */
    public String getXmlCloseTag (String key) { return "</"+getTag (key)+">"; }

    /**
     * Retorna el tag enmagatzemat amb la clau <b>key</b> encapsulant la cadena <b>value</b> amb
     * sintaxis XML.
     *
     * @param key Clau del tag a obtenir
     * @param value Valor a intercalar entre el tag d'obertura i de tancament.
     * @return <b>[tag]value[/tag]</b>
     */
    public String getXmlTag (String key, String value) {

        String tag = getTag (key);
        return "<"+tag+">"+value+"</"+tag+">";
    }

    /**
     * Retorna el tag enmagatzemat amb la clau <b>key</b> encapsulant el valor <b>value</b> amb
     * sintaxis XML.
     *
     * @param key Clau del tag a obtenir
     * @param value Valor a intercalar entre el tag d'obertura i de tancament.
     * @return <b>[tag]value[/tag]</b>
     */
    public String getXmlTag (String key, byte value) { return getXmlTag (key, NumberToString.toString (value)); }

    /**
     * Retorna el tag enmagatzemat amb la clau <b>key</b> encapsulant el valor <b>value</b> amb
     * sintaxis XML.
     *
     * @param key Clau del tag a obtenir
     * @param value Valor a intercalar entre el tag d'obertura i de tancament.
     * @return <b>[tag]value[/tag]</b>
     */
    public String getXmlTag (String key, short value) { return getXmlTag (key, NumberToString.toString (value)); }

    /**
     * Retorna el tag enmagatzemat amb la clau <b>key</b> encapsulant el valor <b>value</b> amb
     * sintaxis XML.
     *
     * @param key Clau del tag a obtenir
     * @param value Valor a intercalar entre el tag d'obertura i de tancament.
     * @return <b>[tag]value[/tag]</b>
     */
    public String getXmlTag (String key, int value) { return getXmlTag (key, NumberToString.toString (value)); }

    /**
     * Retorna el tag enmagatzemat amb la clau <b>key</b> encapsulant el valor <b>value</b> amb
     * sintaxis XML.
     *
     * @param key Clau del tag a obtenir
     * @param value Valor a intercalar entre el tag d'obertura i de tancament.
     * @return <b>[tag]value[/tag]</b>
     */
    public String getXmlTag (String key, long value) { return getXmlTag (key, NumberToString.toString (value)); }

    /**
     * Retorna el tag enmagatzemat amb la clau <b>key</b> encapsulant el valor <b>value</b> amb
     * sintaxis XML.
     *
     * @param key Clau del tag a obtenir
     * @param value Valor a intercalar entre el tag d'obertura i de tancament.
     * @return <b>[tag]value[/tag]</b>
     */
    public String getXmlTag (String key, float value) { return getXmlTag (key, NumberToString.toString (value)); }

    /**
     * Retorna el tag enmagatzemat amb la clau <b>key</b> encapsulant el valor <b>value</b> amb
     * sintaxis XML.
     *
     * @param key Clau del tag a obtenir
     * @param value Valor a intercalar entre el tag d'obertura i de tancament.
     * @return <b>[tag]value[/tag]</b>
     */
    public String getXmlTag (String key, double value) { return getXmlTag (key, NumberToString.toString (value)); }

    /**
     * Retorna el tag enmagatzemat amb la clau <b>key</b> encapsulant el valor <b>value</b> amb
     * sintaxis XML.
     *
     * @param key Clau del tag a obtenir
     * @param value Valor a intercalar entre el tag d'obertura i de tancament.
     * @return <b>[tag]value[/tag]</b>
     */
    public String getXmlTag (String key, boolean value) { return getXmlTag (key, NumberToString.toStringSN (value)); }


    //   __  __      _            _                         _     _ _
    //  |  \/  |    | |          | |                       | |   | |_)
    //  | \  / | ___| |_ ___   __| | ___ ___    _ __  _   _| |_  | |_  ____ ___
    //  | |\/| |/ _ \ __) _ \ / _` |/ _ \ __)  | `_ \| | | | `_ \| | |/ ___) __)
    //  | |  | |  __/ |_ (_) | (_| |  __/__ \  | |_) | |_| | |_) | | | (___\__ \
    //  |_|  |_|\___|\__)___/ \__,_|\___|___/  |  __/ \__,_|_,__/|_|_|\____)___/
    //                                         | |
    //                                         |_|

    /**
     * Afegueix la clau <b>key</b> i el tags <b>tagName</b> a la llista de tags.
     *
     * @param key
     * @param tagName
     */
    public void addTag (String key, String tagName) {

        if (key!=null && tagName!=null) {
            key = key.trim();
            tagName = tagName.trim();

            if (xmlTags.containsKey (key)) { System.out.println ("[XML_TAG_ERROR] Tag: \""+key+"\" Description: Duplicated key. "); }
            else if (xmlTags.containsValue(key)) { System.out.println ("[XML_TAG_ERROR] Tag: \""+key+"\" Description: Ambiguous tagname; another key define the same tagname. "); }
            else { xmlTags.setProperty (key, tagName); }
        }
    }

}
